#ifndef lint
static char rcsid_st2_cmp_c[] = "\
@(#)  $Header: st2_cmp.c,v 1.10 91/11/26 23:10:16 clynn Rel $  \n";
/*-------------------------------------------------------------------*
 |       Copyright (c) 1991 by BBN Systems and Technologies,         |
 |            A Division of Bolt Beranek and Newman Inc.             |
 |                                                                   |
 |  Permission to use, copy, modify, distribute, and sell this       |
 |  software and its documentation for any purpose is hereby         |
 |  granted without fee, provided that the above copyright notice    |
 |  and this permission appear in all copies and in supporting       |
 |  documentation, and that the name of Bolt Beranek and Newman      |
 |  Inc.  not be used in advertising or publicity pertaining to      |
 |  distribution of the software without specific, written prior     |
 |  permission. BBN makes no representations about the suitability   |
 |  of this software for any purposes.  It is provided "AS IS"       |
 |  without express or implied warranties.                           |
 *-------------------------------------------------------------------*/
#endif lint


/*
M* st2_cmp.c  ST-II Control Message Protocol Routines.
M*
M*	This module contains the routines that implement the ST-II
M*	Control Protocol.  It contains routines to: process incoming
M*	packets (st2_CMPInput, SCMPDecode), dispatch them based on
M*	opcode (HandleXxx), run the state machine
M*	(st2_PCBUpdate), generate SCMP messages (xxx_Message);
M*	encode and send them (SCMPEncode, CtrlMsgSend/ReplyMsgSend,
M*	st2_CMPOutput); and miscellaneous supporting routines.
M*
m* Overview
m*
m*	Packets containing ST-II Control Messages are passed to
m*	st2_CMPInput by st2:st2_intr.  The messages are parsed, by
m*	SCMPDecode, into a data structure (aPrsSCM) and validated.
m*	They are then associated with a virtula link (aVLink) and
m*	hense a stream based on the Virtual Link identifiers; the
m*	stream may not yet be known.  The parsed message and resulting
m*	virtual link information is then dispatched to the appropriate
m*	routine (HandleXxx, where Xxx is the message OpCode).  The
m*	handlers then update state information as is appropriate and
m*	usually cause the stream's state machine (s) to be run
m*	(st2_PCBUpdate).  Either the handlers or the state machine
m*	may call routines (xxx_Message) to build and send other SCMP
m*	messages.
m*
m*	Several databases are maintained.  Those associtated with a
m*	stream, or more accurately a meander of a stream, are:
m*
m*	aST2pcb:	Primary data structure, one per meander.
m*
m*	Target:		Per target information, chained (by nextp)
m*			from the aST2pcb (targetsp).  Contains info
m*			describing the target and a prioritized table
m*			of possible routes toward the target.
m*
m*	aVLink:		Per (ST) neighbor per meander information.
m*			Pointed to (by vlinkp) by those targets
m*			sharing the same next-hop ST Agent.  Entries
m*			are in an array indexed by RVLId, and also
m*			hashed by SVLId and remote ST Agent IP
m*			Address.  (Should be combined with NextHop.)
m*
m*	NextHop:	More information about the next-hop ST Agent,
m*			the path to it, and associated resources.
m*			Pointed to by aVLink (next_hopp; also chained
m*			(by nextp) off of aST2pcb (next_hopsp).
m*			(Should be combined with aVLink.)
m*
m*	aNxtHopList:	Per meander per "next hop" local network
m*	aNxtHop:	address information.  Pointed to from NextHop
m*			(by nxthopp).  Multiple NextHops may be
m*			reached through a single aNxtHop when local
m*			network multicast/broadcast is being used.
m*			This data structure is part of the ST-II Data
m*			Path and contains information required to
m*			quickly forward the data packets; see st2.c.
m*
m*	st2_ConHsh:	This table is part of the ST-II Data Path and
m*			maps incoming Hop-By-Hop Identifiers (HIDs) to
m*			the corresponding aST2pcb data structure.
m*			Each aST2pcb nominally has a single HID which
m*			points to it; there may be multiple HIDs
m*			pointing to the same aST2pcb while HIDs are
m*			being negotiated or changed.
m*
m*	In order to minimize control messages, processing is performed
m*	in several steps.  First, routes to each target are requested
m*	by calling the st2_FindRoutes routine, once per target.
m*	Second, the st2_OptimizeRoutes routine is called to order the
m*	routes for all of the targets.  Third, resources are allocated
m*	and, when allocation fails, alternate routes are tried.
m*	Finally, CONNECT messages are sent for targets with successful
m*	allocations and REFUSE messages are sent for those which
m*	failed.
m*
 */

/*
m* Status:
m*   Features:
m*	Basic CONNECT, ACCEPT, DISCONNECT, REFUSE, delegated HID negotiation.
m*	Data Timestamps, IP Encapsulation & Source Routing.
m*	Failure detection, and notification.
m*   Untested Features:
m*	Asynchronous resource reservation replies, failure modes.
m*   Restrictions/Bugs:
m*	Duplicate neighbor entries may be created when using source routing.
m*	Limited number of next-hops per stream.
m*	Needs to be more robust when presented with errors.
m*	Retransmission timing is slow; packet causing ARP is dropped.
m*   Things to do:
m*	CHANGE (REQUEST) ,ERRORINxxx, HELLO, HID negotiations, NOTIFY,
m*	STATUS (REQUEST).
m*	Data structure cleanup.
m*	Failure recovery; neighbor state & delay.
m*	Network layer multicast support.
m*	State machine is incomplete.
m*
 */

/* Module Revision History
 *
 * $Log:	st2_cmp.c,v $
 * Revision 1.10  91/11/26  23:10:16  clynn
 * Fixed backoff bug in retransmission timer.
 * Fixed bug in NoRecovery option.
 * Fixed bug in inversion of source routes.
 * Fixed bugs related to neighbor VLIds over failures and reboots.
 * More accurate ReasonCode when failure is detected.
 * More meaningful variable names.
 * No LnkRef for REFUSE due to application disconnect or failure.
 * Record more useful information for some traps.
 * Removed some duplicated code.
 * Round average round-trip-time instead of truncating.
 * Updated some documentation.
 * 
 * Revision 1.9  91/11/04  09:21:15  clynn
 * Updated for Public Domain Release.  Major changes: addition of Source
 * Routing, IP Encapsulation, HELLO protocol between neighbors, tracking
 * of neighbors, detection of component or agent failures & notification
 * to applications.  More consistant naming and format, including making
 * all external names begin with "st2_".  Moved some routines and data
 * structures to reduce external references.
 * 
 * Added IP Encapsulation, Source Routing, HELLO, failure detection,
 * notification, and recovery support.
 * 
 * Revision 1.8  91/05/28  16:17:42  clynn
 * New features: Add new targets from Application layer, UserData support,
 * basic bandwidth reservation for point-to-point links, more complete
 * state tables, added pcode parameter to InitaFlowSpec3, extended packet
 * buffer abstraction, added network interface abstraction; ststat utility.
 * Bug fixes: ADDR IN USE problem, data send problem, causes of some
 * crashes, cleanup of protocol control blocks.
 * Work around: DARTNET receive memory leak.
 * Eliminated several small modules to reduce globals; adeded Makefile.
 * 
 * Revision 1.1  91/03/15  18:33:23  clynn
 * Initial revision
 */

#define _FILE_ Strcmp			/* Module bugid 5,89 */


/* External definitions */

#include <stdio.h>
#include <sys/systm.h>
#include <sys/errno.h>

#include <sys/types.h>		/* caddr_t u_xxx for sys/mbuf.h net/
				   if.h netinet/in.h */
#include <sys/socket.h>		/* sockaddr for net/if.h */
#include <sys/socketvar.h>	/* for so_error & so_pcb */
#include <net/if.h>		/* ifnet, ifaddr */
#include <netinet/in.h>		/* before st2.h -> st2_api.h */

typedef unsigned long n_long;	/* Include disease */
#include <netinet/ip.h>		/* IP Header for encapsulation */

#include "st2_api.h"		/* Global Definitions, before st2.h */
#include "st2.h"		/* ST-II Implementation definitions,
				 * includes st2_api.h, before st2_cmp.h */
#include "st2_cmp.h"		/* SCM Protocol definitions */


/* ??? Following should be ABit, DBit, etc, from st2_cmp.h */
#define HC_OPTION_CHANGE	(0)
#define HC_OPTION_ADD		(1 << 7)
#define HC_OPTION_DELETE	(1 << 6)

#define NO_SRP ((struct aSrcRut *) NULL)


/* External routines */

enum ST2Errors		 st2_RsrcAccount (/* gifcp, msec, thrup, costp */);

/* Local routines */

static void		 CompAccept (/* accept_pscmp, reply_pscmp */);
static void		 CompConnect (/* ctrl_pscmp, reply_pscmp */);
static void		 CompDisconnect (/* ctrl_pscmp, reply_pscmp */);
static void		 CompRefuse (/* ctrl_pscmp, reply_pscmp */);
static boolean		 CtrlMsgAddTarg (/* messagep, targetp, dir */);
static int		 CtrlMsgAddTrgSet (/* messagep, cp, vlinkp, set,
								newstate */);
static struct aPrsSCM	*CtrlMsgCreate (/* */);
static void		 CtrlMsgDel (/* pscmp */);
static int		 CtrlMsgDropAcked (/* reply, ref */);
static enum ST2Errors	 CtrlMsgRelay (/* pscmp, vlinkp */);
static enum ST2Errors	 CtrlMsgSend (/* request, vlinkp */);
static int		 DoAcknowledge (/* request, vlinkp, reason */);
static enum ST2Errors	 DoConnect (/* targetp */);
static enum ST2Errors	 DoDisconnect (/* targetp, reason */);
static enum ST2Errors	 DoError (/* request, vlinkp, reason */);
static void		 DoHello (/* np, tvp */);
static int		 DoHidReply (/* messagep, vlinkp, opcode, hop_id */);
static enum ST2Errors	 DoRefuse (/* targetp, comp */);
static boolean		 FlowSpecEncode (/* messagep, hflowp */);
static boolean		 FlowSpecFromParm (/* flowparmp, hflowp */);
static enum ST2Errors	 HandleAccept (/* messagep, vlinkp */);
static enum ST2Errors	 HandleAck (/* messagep, vlinkp */);
static enum ST2Errors	 HandleBadOpCode (/* messagep, vlinkp */);
static enum ST2Errors	 HandleChange (/* messagep, vlinkp */);
static enum ST2Errors	 HandleChangeRequest (/* messagep, vlinkp */);
static enum ST2Errors	 HandleConnect (/* messagep, vlinkp */);
static enum ST2Errors	 HandleDisconnect (/* messagep, vlinkp */);
static enum ST2Errors	 HandleErrorInRequest (/* messagep, vlinkp */);
static enum ST2Errors	 HandleErrorInResponse (/* messagep, vlinkp */);
static enum ST2Errors	 HandleHello (/* messagep, vlinkp */);
static enum ST2Errors	 HandleHidApprove (/* messagep, vlinkp */);
static enum ST2Errors	 HandleHidChange (/* messagep, vlinkp */);
static enum ST2Errors	 HandleHidChangeRequest (/* messagep, vlinkp */);
static enum ST2Errors	 HandleHidReject (/* messagep, vlinkp */);
static enum ST2Errors	 HandleNotify (/* messagep, vlinkp */);
static enum ST2Errors	 HandleRefuse (/* messagep, vlinkp */);
static enum ST2Errors	 HandleStatus (/* messagep, vlinkp */);
static enum ST2Errors	 HandleStatusResponse (/* messagep, vlinkp */);
static boolean		 HIDActivate (/* cp, hop_id */);
#ifdef NOTYET
enum ST2Errors		 HIDAdd (/*cp, hop_id*/);
#endif NOTYET
static enum ST2Errors	 HIDChange (/* subop, hid, vlinkp */);
static void		 HIDClear (/* hop_id */);
static HID		 HIDNew (/* cp */);
static boolean		 HIDSet (/* cp, hop_id */);
static enum ST2Errors	 IPEncapFwd (/* link_type, lclIPAdr, nxtIPAdr,
							ipsrcrutp, anhp */);
static enum ST2Errors	 IPEncapRev (/* link_type, ipp, anhp */);
#ifdef NOTYET
static void		 log_FlowSpec (/* hflowp, subr */);
#endif NOTYET
static void		 log_Forwarder (/* cp, subr */);
static void		 log_NextHop (/* nexthop, subr */);
static void		 log_Parms (/* pscmp */);
static void		 log_Route (/* routep, subrp */);
static void		 log_SAP (/* stsap, stringp */);
static void		 log_Stream (/* cp, subrp */);
static void		 log_Target (/* targetp, subr */);
static void		 log_VLink (/* vlinkp, subr */);
static void		 MTUAdjust (/* cp, nxthopp */);
static struct aNeighbor	*NbrCreate (/* nxtstipadr, flags, pscmp, targetp */);
static void		 NbrFailed (/* np, reason */);
#ifdef NOTYET
static struct aNeighbor *NbrFind (/* ipadr */);
#endif NOTYET
static int		 NbrInsert (/* vlinkp, RecoveryTimeout */);
static void		 NbrRemove (/* vlinkp */);
static void		 NbrRTTUpdate (/* np, refnum, tvp */);
#ifdef NOTYET
static void		 NbrSort (/* vlinkp, rcvryto */);
#endif NOTYET
static struct NextHop	*NextHopAdd (/* ipsrcrutp, arcp, anhp, anhlpp,
								pktdescpp */);
static enum ST2Errors	 NextHopDel (/* nexthopp */);
static struct aST2pcb	*PCBByName (/* Namep */);
static enum ST2Errors	 ReplyMsgSend (/* request, reply */);
static boolean		 RouteCheck (/* struct Route * */);
static enum ST2Errors	 SCMPDecode (/* scmhdrp, scmlen, pscmpp, pscmlenp */);
static enum ST2Errors	 SCMPEncode (/* mtu, xtrahdrlen, pscmp, pdupp,
					 pdulenp */);
int			 st2_ApplEntityMatch (/* entp, originp */);
#ifdef NOTYET
|int			 st2_ClkFast (/**/);
#endif NOTYET
int			 st2_ClkSlow (/**/);
enum ST2Errors		 st2_CMPInput (/* gifcp, cp, msgp, st2p, datap,
					   encaplen */);
enum ST2Errors		 st2_CMPOutput (/* pscmp, vlinkp */);
#ifdef NOTYET
|int			 st2_ctrlinput (/* m, pcb, ptype, ifp, lcid */);
#endif NOTYET
enum ST2Errors		 st2_DoAccept (/* targetp, flowspecp, opts,
								userdatap */);
int/*enum ST2Errors*/	 st2_DoConnect (/* cp, sockadrp, maxsockbytes,
						flowspecp, userdatap */);
int/*enum ST2Errors*/	 st2_DoDisconnect (/* cp, reason, targetp, userdatap */);
#ifdef NOTYET
|int			 st2_Ioctl (/* sop, mp, namp, rightsp */);
#endif NOTYET
void			 st2_PCBDel (/* cp */);
struct aST2pcb		*st2_PCBGet (/* */);
void			 st2_PCBUpdate (/* cp */);
extern void		 st2_VLinkReset (/* vlinkp, suggest_vlid */);
static enum ST2Errors	 TargAdd (/* cp, aTargetp, reference */);
static struct Target	*TargByParm (/* cp, dir, aTargetp */);
static enum ST2Errors	 TargDel (/* targetp */);
static enum ST2Errors	 TargDisc (/* targetp, reference, reason,
					       userdatap */);
static struct aVLink	*VLinkByVLId (/* vlid */);
static void		 VLinkDel (/* vlinkp */);
struct aVLink		*VLinkNew (/* cp, direction, type */);
static boolean		 VLinkSetRemote (/* vlinkp, remote_vlid, rmtIPAdr */);


/* Local Data Structures */


/* Any of this Agent's IPAdr's, for stream Names */
static INETADDR		local_inetaddr = INADDR_ANY;


/* Define bit vectors specifying mandatory and user optional parameters
 * for each scmp message type (aka OpCode).
 */

static enum ST2Errors	(*ops[])() = {
#define anOpCode(value,name,bits,mparm,oparm,vlids) Ident (Handle)name,
	ST2OpCodes
#undef  anOpCode
	};

#define NUM_ST_OPCODES (DimensionOf (ops))


	octet4		 st2_parmManByOp[ NUM_ST_OPCODES ] = {
#define anOpCode(value,name,bits,mparm,oparm,vlid) mparm 0,
#define QM(pn) (1 << (31 - (pn))) |
    ST2OpCodes
#undef QM
#undef anOpCode
};


static octet4		 st2_parmOptByOp[ NUM_ST_OPCODES ] = {
#define anOpCode(value,name,bits,mparm,oparm,vlid) oparm 0,
#define QO(pn) (1 << (31 - (pn))) |
    ST2OpCodes
#undef QO
#undef anOpCode
};


/* Define minimum and maximum parameter lengths */

static struct ParmLen {
    octet1		 min,
			 max;
}			 parmlens[ NUM_ST_PARMS ] = {
#define aPCode(v,s,p,n,mn,mx,x) { mn, mx },
#define sPCode(v,s,p,n,mn,mx,x) { mn, mx },
#define tPCode(v,s,p,n,mn,mx,x) { mn, mx },
#define xPCode(v,s,p,n,mn,mx,x) { mn, mx },
    ST2PCodes
#undef xPCode
#undef tPCode
#undef sPCode
#undef aPCode
};


static struct aNeighborList st2_neighbors = { InitaExpAry (DEF_NUMNBRS,0xf6) };


STATIC struct aRXQue st2_retry_q = {	/* splnet */
 (struct aPrsSCM *) &( st2_retry_q ), (struct aPrsSCM *) &( st2_retry_q ) };


/* Inline Routines to read/write State variables, and translation tables. */


#define activate_aNxtHop( nxthopp, hop_id ) /* from HandleHidApprove st2_PCBUpdate */ \
 {  nxthopp->HID = hop_id; BitSet32 (nxthopp->flags, NxtHopActive); }



#define deactivate_aNxtHop( nxthopp )	/* from NextHopDel TargDisc st2_PCBDel */ \
 {  BitClr32 (nxthopp->flags, NxtHopActive);				\
    /* ??? racy: nxthopp->gifcp = NULL; */ }


	/* ??? this doesn't propagate G-Bit */
#define disconnect_Forwarder( cp, reason, userdatap ) /* from HandleDisconnect st2_DoDisconnect */ \
 {  struct Target	*targetp;					\
    SetState (FWD,cp,fwdr_state,FW_DISCONNECT);				\
    cp->fwdrerror = reason;						\
    ForAllTargets (cp, targetp)						\
    {									\
	if ( IsTargState (targetp,TS_DEAD) )				\
	    continue;							\
									\
	if ( NOT IsTargStateIn (targetp,TS_RELEASED,TS (TS_CLOSED)	\
			   TS (TS_CLOSING) TS (TS_WANTCLOSED)) )	\
	    SetTargState (targetp,TS_WANTCLOSED);			\
									\
	/* if not permanent (i.e., may repair )				\
	{								\
	    if ( NOT IsTargStateIn (targetp,TS_DISCONNECTED,TS (TS_ONHOLD)) ) \
		SetTargState (targetp,TS_ONHOLD);			\
	}								\
	else permanent error (from application) */			\
									\
	SetTargState (targetp,TS_DISCONNECTED);				\
									\
	targetp->discon_userdatap = userdatap;				\
	targetp->error = reason;					\
	cp->needupdate++;						\
    }									\
}


#define ForAllRxQ(pp,pscmp)		/* can scan if current dequeued */ \
    for ( pp = (struct aPrsSCM *) &( st2_retry_q ), pscmp = pp->nextp	\
	 ; pscmp NE (struct aPrsSCM *) &( st2_retry_q ) ; pscmp = pp->nextp )



#define Forwarder_for_HID( hop_id )	/* from HandleConnect HandleStatus */ \
 ((    (hop_id NE (HID) 0)						\
   AND (st2_ConHsh[ ConfigParm (hid_mask) & hop_id ].hid EQ hop_id))	\
  ? st2_ConHsh[ ConfigParm (hid_mask) & hop_id ].cp : NO_PCBP)


#define MsecBetween(nowp,thenp) \
  ( (((nowp)->tv_sec  - (thenp)->tv_sec)  * 1000) \
  + (((nowp)->tv_usec - (thenp)->tv_usec) / 1000) )


#define PackTState(tp,x) ( ((int) (tp)->targstate[ (int) Overall ] << 24) \
	| ((int) (tp)->targstate[ (int) Prevhop ] << 16)		\
	| ((int) (tp)->targstate[ (int) Nexthop ] << 8) | (int) (x) )


/*??? Need to re-verify that the target is still active and has not been
   disconnected or refused since we queued the accept.  This was the reason
   for originally queueing the accepts with the target, so that they would
   be discarded if the target was torn down.  For speed, could simply
   note if any targets have been deleted and if not, just forward all
   the accepts back... */


#define relay_accepts( cp )		/* from HandleConnect HIDSet st2_DoConnect */ \
 {  int			 oldpri;					\
    struct aPrsSCM	*pscmp,						\
			*listp;						\
    oldpri = splnet ();							\
    {	listp = cp->accepts;						\
	cp->accepts = NO_PSCMP;						\
    }									\
    (void) splx ( oldpri );						\
									\
    if (    (listp NE NO_PSCMP)						\
	AND (cp->upstream_vlinkp->np NE NO_NBRP) )			\
    {									\
	/* ??? flow_spec is  struct aFlowSpec3 */			\
	(void) NbrInsert( cp->upstream_vlinkp,				\
			ntoh2 ((&( cp->flow_spec))->RecoveryTimeout) );	\
	/* sets cp->upstream_vlinkp->rcvryto */				\
    }									\
									\
    while ( (pscmp = listp) NE NO_PSCMP )				\
    {									\
	listp = pscmp->nextp;						\
	pscmp->nextp = NO_PSCMP;					\
	(void) CtrlMsgRelay( pscmp, cp->upstream_vlinkp );		\
	CtrlMsgDel( pscmp );						\
    }									\
}


#ifdef  STIIDEBUG


#define share_aPrsSCM( pscmp, bgs )	/* from HandleAccept DoConnect DoDisconnect CtrlMsgRelay */ \
 {  if ( (pscmp)->msgcnt <= 0 )						\
    {									\
	(void) printf ( "attempt to share unreferenced parsed SCM\n" );	\
	bgs;								\
    }									\
    else								\
	(pscmp)->msgcnt++;						\
}

#else

#define share_aPrsSCM( pscmp, bgs )	/* from HandleAccept DoConnect DoDisconnect CtrlMsgRelay */ \
 {  if ( (pscmp)->msgcnt <= 0 )						\
    {									\
	bgs;								\
    }									\
    else								\
	(pscmp)->msgcnt++;						\
}

#endif  STIIDEBUG


#define NxtRefNum(x) (Count (scmp_refnum,1))


/*
S* RemRing ( variable, actions ) and cp
s*
s*	MACRO for nxtNamep, fdxp, nxtGroupp
s*	Remove cp from ring named by "variable" and then perform actions
s*
 */


#define RemRing( variable, actions )					\
  if ( cp->variable NE NO_PCBP )					\
    {									\
	struct aST2pcb *wcp;						\
									\
	oldpri = splnet ();						\
									\
	if ( (wcp = cp->variable) NE NO_PCBP )				\
	{								\
									\
	    while (    (wcp->variable NE cp) /* NULL is paranoia */	\
		   AND (wcp->variable NE NO_PCBP) )			\
		wcp = wcp->variable;					\
									\
	    if ( wcp EQ cp->variable )					\
		wcp->variable = NO_PCBP; /* ring gone */		\
	    else							\
		wcp->variable = cp->variable;				\
	    cp->variable = NO_PCBP;					\
									\
	    actions							\
	}								\
									\
	(void) splx ( oldpri );						\
    }


/*
S* VLinkHash ( sndrIPAdr, sVLId )
s*
s*	Macro to find the virtual link structure corresponding to the
s*	given remote IPAdr and Sender Virtual Link Id.
s*
 *	There is no magic in any of this.
 *	??? this discards most informatative bits of SVLId & sndrIPAdr
 *
 *	hash = sndrIPAdr + (sVLId << 6);
 *	hash += hash >> 11;
 *	hash &= ((1<<11)-1);
 *	hash %= ConfigParm (vlnk_hash)
 */

#define VLinkHash( sndrIPAdr, sVLId )	/* from VLinkSetRemote st2_CMPInput VLinkDel */ \
    ((int) ( (  ( (   (sndrIPAdr + (sVLId << 6))			\
		   + ((sndrIPAdr + (sVLId << 6)) >> 11)))		\
	      & ((1<<11)-1) )						\
	    % ConfigParm (vlnk_hash)) )


/*
S* CompAccept ( accept_pscmp, reply_pscmp )
s*
s*	Routine called when an ACCEPT is acknowledged or timesout.
s*
 */

  static void				/* Next bugid 0x50104 */
CompAccept( accept_pscmp, reply_pscmp ) /* from CtrlMsgDropAcked st2_CMPOutput */
					/* via st2_DoAccept */
  struct aPrsSCM	*accept_pscmp,	/* We CtrlMsgDel it */
			*reply_pscmp;	/* Caller retains, NULL on timeout */
{
    int			 nt = accept_pscmp->targlst.nxtfree;
    Reference		 ref = accept_pscmp->hoscmphdr.Ref;
    struct aST2pcb	*cp = accept_pscmp->cp;
    struct Target	*tp;


    tp = accept_pscmp->targlst.TargetDescs[0].targetp; /* Only 1 */
    CtrlMsgDel( accept_pscmp );		/* Before fail_upstream_target */

    if ( tp EQ NO_TRGP )
    {
	BUGSTOP (-,cp,0x50101);	/* what wiped it out */
	return;
    }

    if ( IsTargState (tp, TS_DEAD) )
	return;


    if ( tp->pending_reference NE ref )
    {
#ifdef  STIIDEBUG
	/* what snuck in */
	BUGSTOP (-,(ref<<16) | tp->pending_reference,0x50102);
#endif  STIIDEBUG
    }
    else
    {
	tp->pending_reference = 0;	/* let machine proceed */
	cp->needupdate++;
    }

    if ( nt NE 1 )
    {
	BUGSTOP (-,nt,0x50103);	/* Protocol Error */
	return;				/* ??? what to do */
    }


    /* ==================================================================== */
    /* Send REFUSE in response to retransmission timeout */

    if (    (reply_pscmp EQ NO_PSCMP)
	AND (NOT IsTargState (tp,TS_DEAD)) )
    {
	if ( tp->error EQ NoError )
	    tp->error = RetransTimeout;

	if ( NOT IsTargStateIn (tp,TS_RELEASED,TS (TS_CLOSED)
				TS (TS_CLOSING) TS (TS_WANTCLOSED)) )
	    SetTargState (tp,TS_WANTCLOSED);

	switch ( tp->targstate[ (int) Prevhop ] )
	{
	case TS_CONNECTED:
	    tp->cp->needupdate++;
	    SetTargState (tp,TS_WANTHOLD);
	    break;

	case TS_WANTHOLD:
	    /* ??? don't wait yet */
	    tp->cp->needupdate++;
	    SetTargState (tp,TS_FAILED);
	    break;

	case TS_PENDHOLD:
	    tp->cp->needupdate++;
	    SetTargState (tp,TS_ONHOLD);
	    break;

	case TS_ONHOLD:
	    break;

	case TS_FAILED:
	    tp->cp->needupdate++;
	    break;

	default:
	case TS_REFUSING:
	    tp->cp->needupdate++;
	    SetTargState (tp,TS_DISCONNECTED);
	    break;

	case TS_DISCONNECTED:
	    break;
	} /* end of previous hop state switch */
    }
/*  else No state change on ack of ACCEPT */

    if ( cp->needupdate > 0 )
	st2_PCBUpdate( cp );
	/* cp may be invalid */

    return;
}

/*
S* CompConnect( ctrl_pscmp, reply_pscmp )
s*
s*	Downstream CONNECT has been acked or timedout.
s*
 */

  static void				/* Next bugid 0x50207 */
CompConnect( ctrl_pscmp, reply_pscmp )	/* from CompAccept CtrlMsgDropAcked st2_CMPOutput */
					/* via DoConnect */
  struct aPrsSCM	*ctrl_pscmp,	/* We CtrlMsgDel it */
			*reply_pscmp;	/* Caller retains, NULL on timeout */
{
    enum ST2Errors	 error;
    int			 n = ctrl_pscmp->targlst.nxtfree;
    Reference		 ref = ctrl_pscmp->hoscmphdr.Ref;
    struct aST2pcb	*cp;
    struct aTargetDesc	*atdp = &( ctrl_pscmp->targlst.TargetDescs[0] );


    if ( reply_pscmp NE NO_PSCMP )
    {
	switch ( reply_pscmp->hoscmphdr.OpCode )
	{
	case OpAccept:			/* ??? ISSUE: have trouble here as */
	    BUGSTOP (-,n,0x50201);	/* ACCEPT probably doesn't specify */
	    /* Fall through */		/* all targets */

	case OpAck:
	case OpHidApprove:
	case OpHidReject:
	    error = NoError;
	    break;

	default:			/* Unexpected acknowledgement */
	    BUGSTOP (-,reply_pscmp->hoscmphdr.OpCode,0x50202);
	    /* Fall through */		/* will get extraneous message ... */

	case OpRefuse:			/* ??? ISSUE: have trouble here as 
	    BUGSTOP (-,n,0x50203);	/* REFUSE probably doesn't specify */
	    /* Fall through */		/* all targets */

	case OpErrorInRequest:
	    error = ErrorUnknown;
	    if ( reply_pscmp->reasoncodep NE (octet2 *) NULL )
		error = (enum ST2Errors) *reply_pscmp->reasoncodep;
	    break;
	} /* end of reply message type switch */
    }
    else
    {
	error = RetransTimeout;		/* timeout */

	if ( atdp->targetp->curchoicep NE NO_RUTCHOICEP )
	{
	    /* ??? function of reason */
	 /* atdp->targetp->curchoicep->flags |= RouteFailed | RouteNoRsrc; */
	    atdp->targetp->curchoicep->flags |= RouteFailed | RouteNoRoute;
	}
	else
	{
	    BUGSTOP (-,1,0x50204); /* paranoia */
	}
    }

    cp = NO_PCBP;
    for ( ; n > 0 ; n--, atdp++ )
    {
	if ( atdp->targetp EQ NO_TRGP )
	    continue;			/* ??? what wiped out the entry */

	if ( IsTargState (atdp->targetp, TS_DEAD) )
	    continue;

	if ( atdp->targetp->pending_reference NE ref )
	{
	    /* what snuck in */
	    BUGSTOP (-,(ref<<16)|atdp->targetp->pending_reference,0x50205);
	}
	else
	{
	    atdp->targetp->pending_reference = 0; /* let machine proceed */
	}

	if ( NOT IsTargState (atdp->targetp, TS_OPENING) )
	{
	    BUGSTOP (-,PackTState (atdp->targetp,0x50206), 0x50806);
	    cp = atdp->targetp->cp;
	    cp->needupdate++;		/* maybe waiting for us */
	    continue;			/* probably shutting down */
	}

	if ( error EQ NoError )
	{
	    SetTargState (atdp->targetp,TS_OPEN);
	}
	else
	{
	    atdp->targetp->error = error;
					/* ??? NoRecovery */
	    if ( NOT IsTargStateIn (atdp->targetp,TS_DISCONNECTED,
				   TS (TS_REFUSING) TS (TS_FAILED)) )
	    {
		SetTargState (atdp->targetp,TS_FAILED);
	    }

	    SetTargState (atdp->targetp,TS_WANTCLOSED);
	    cp = atdp->targetp->cp;
	    cp->needupdate++;
	}

    } /* end of for all reported targets loop */

    CtrlMsgDel( ctrl_pscmp );

    if (    (cp NE NO_PCBP)
	AND (cp->needupdate > 0) )
    {
	st2_PCBUpdate( cp );
	/* cp may be invalid */
    }

    return;
}

/*
S* CompDisconnect ( ctrl_pscmp, reply_pscmp )
s*
s*	Downstream DISCONNECT has been acked or timedout.
s*
 */

  static void				/* Next bugid 0x50303 */
CompDisconnect( ctrl_pscmp, reply_pscmp )	/* from CtrlMsgDropAcked st2_CMPOutput */
					/* via DoDisconnect */
  struct aPrsSCM	*ctrl_pscmp,	/* We CtrlMsgDel it */
			*reply_pscmp;	/* Caller retains, NULL on timeout */
{
    int			 n = ctrl_pscmp->targlst.nxtfree;
    struct aTargetDesc	*atdp = &( ctrl_pscmp->targlst.TargetDescs[0] );
    struct aST2pcb	*cp = NO_PCBP;
    enum ST2Errors	 reason;
    Reference		 ref = ctrl_pscmp->hoscmphdr.Ref;


    if (   (reply_pscmp EQ NO_PSCMP)
	OR ((int) reply_pscmp->hoscmphdr.OpCode NE (int) OpAck) )

	reason = RetransTimeout;
    else
	reason = NoError;


    /* ??? G-Bit */
    for ( ; n > 0 ; n--, atdp++ )
    {
	if ( atdp->targetp EQ NO_TRGP )
	    continue;			/* what nulled it out */

	if ( IsTargState (atdp->targetp, TS_DEAD) )
	    continue;

	if ( atdp->targetp->pending_reference NE ref )
	{
	    if ( atdp->targetp->pending_reference NE 0 )
		BUGSTOP (-,(ref << 16) | atdp->targetp->pending_reference,0x50301);	/* what snuck in */
	}
	else
	{
	    atdp->targetp->pending_reference = 0; /* let machine proceed */
	}

	if ( atdp->targetp->error EQ NoError )
	    atdp->targetp->error = reason;

	switch (  atdp->targetp->targstate[(int)Nexthop] )
	{
	case TS_CLOSING:
	    SetTargState (atdp->targetp,TS_CLOSED);
	    cp = atdp->targetp->cp;
	    cp->needupdate++;
	    break;

	case TS_PENDSUSPEND:
	    SetTargState (atdp->targetp,TS_SUSPENDED);
	    /* ??? start timer */	/* nothing to do */
	    break;

	default:
	    BUGSTOP (-,atdp->targetp->targstate[(int)Nexthop],0x50302);
	    cp = atdp->targetp->cp;
	    cp->needupdate++;
	    break;
	} /* end of state switch */

    } /* end of for all reported targets loop */

    CtrlMsgDel( ctrl_pscmp );

    if (    (cp NE NO_PCBP)
	AND (cp->needupdate > 0) )
    {
	st2_PCBUpdate( cp );
	/* cp may be invalid */
    }

    return;
}

/*
S* CompRefuse ( ctrl_pscmp, reply_pscmp )
s*
s*	Upstream REFUSE has been acked or timedout.
s*
 */

  static void				/* Next bugid 0x50404 */
CompRefuse( ctrl_pscmp, reply_pscmp )	/* from CtrlMsgDropAcked st2_CMPOutput */
					/* via st2_PCBUpdate*2 */
  struct aPrsSCM	*ctrl_pscmp,
			*reply_pscmp;
{
    int			 n = ctrl_pscmp->targlst.nxtfree;
    struct aTargetDesc	*atdp = &( ctrl_pscmp->targlst.TargetDescs[0] );
    struct aST2pcb	*cp = NO_PCBP;
    Reference		 ref = ctrl_pscmp->hoscmphdr.Ref;


#ifdef STIIDEBUG
    if (    (reply_pscmp NE NO_PSCMP)
	AND ((int) reply_pscmp->hoscmphdr.OpCode NE (int) OpAck) )
	BUGSTOP (-,reply_pscmp->hoscmphdr.OpCode,0x50401);
#endif STIIDEBUG

    /* ??? G-Bit */
    for ( ; n > 0 ; n--, atdp++ )
    {
	if ( atdp->targetp EQ NO_TRGP )
	    continue;

	if ( IsTargState (atdp->targetp, TS_DEAD) )
	    continue;

	if ( atdp->targetp->pending_reference NE ref )
	{				/* what snuck in */
#ifdef STIIDEBUG
	    if ( atdp->targetp->pending_reference NE 0 )
	    {
		/* ??? used change_reference, not pending */
		BUGSTOP (-,(ref << 16) | atdp->targetp->pending_reference,0x50402);
	    }
#endif STIIDEBUG
	}
	else
	{
	    atdp->targetp->pending_reference = 0; /* let machine proceed */
	}

	switch (  atdp->targetp->targstate[(int)Prevhop] )
	{
	case TS_REFUSING:
	    SetTargState (atdp->targetp,TS_DISCONNECTED);
	    cp = atdp->targetp->cp;
	    cp->needupdate++;
	    break;

	case TS_PENDHOLD:
	    SetTargState (atdp->targetp,TS_ONHOLD);
	    cp = atdp->targetp->cp;
	    cp->needupdate++;
	    break;

	default:			/* Some concurrent processing */
	    BUGSTOP (,atdp->targetp->targstate[(int)Prevhop],0x50403);
	    break;
	} /* end of state switch */

    } /* end of for all reported targets loop */

    CtrlMsgDel( ctrl_pscmp );

    if (    (cp NE NO_PCBP)
	AND (cp->needupdate > 0) )
    {
	st2_PCBUpdate( cp );		/* Run the state machine */
	/* cp may be invalid */
    }

    return;
}

/*
S* CtrlMsgAddTarg ( messagep, targetp, dir )
s*
s*	Routine to append the specified target to the given SCMP message.
s*
 */

  static boolean			/* Next bugid 0x50502 */
CtrlMsgAddTarg( messagep, targetp, dir ) /* from CtrlMsgAddTrgSet DoRefuse st2_DoAccept */
					/* MACRO/REPLICATE */
  struct aPrsSCM	*messagep;
  struct Target		*targetp;
  enum link_direction	 dir;		/* UPSTREAM_VLINK | DOWNSTREAM_VLINK */
{
    struct aTargetDesc	*tdp;


    IfLog (DBG_UPDTARG)
	log_Target( targetp, "CtrlMsgAddTarg" );

    if ( messagep->targlst.nxtfree >= messagep->targlst.allocated )
    {
	BUGSTOP (-,messagep->targlst.nxtfree,0x50501);
	return ( FALSE );		/* ??? expand expary */
    }

    tdp = &( messagep->targlst.TargetDescs[ messagep->targlst.nxtfree ] );
    messagep->targlst.nxtfree++;
					/* Note CtrlMsgCreate bzero'd it */
    tdp->targetp = targetp;
    tdp->Targetp = (dir EQ UPSTREAM_VLINK)
		   ? targetp->tparminp : targetp->tparmoutp;

    return ( TRUE );
}

/*
S* CtrlMsgAddTrgSet ( messagep, cp, vlinkp, set, newstate )
s*
s*	Routine to gather all appropriate targets of a stream for the
s*	specified state transisition and virtual link and append them
s*	to the given SCMP message.  Caller makes the state transisition.
s*
 */

  static int				/* Next bugid 0x50600 */
CtrlMsgAddTrgSet( messagep, cp, vlinkp, set, newstate ) /* from DoConnect DoDisconnect */
  struct aPrsSCM	*messagep;
  struct aST2pcb	*cp;
  struct aVLink		*vlinkp;
  unsigned long		 set;		/* TS_WANTOPEN	TS_WANTCLOSED */
  enum targ_state	 newstate;	/* TS_OPENING	TS_CLOSING */
{
    struct Target	*targetp;
    int			 targets_selected = 0;
    struct aUserData	*udp;


    IfLog (DBG_UPDTARG)
	log_Forwarder( cp, "adding targets to aPrsSCM message" );

    ForAllTargets (cp, targetp)
    {
	if ( NOT IsTargState (targetp, TS_BUSY) )
	    continue;

	IfLog (DBG_UPDTARG)
	{
	    (void) printf ( "Target # %d:\n", targetp->target_index );
	    log_Target( targetp, "CtrlMsgAddTrgSet" );
	}


	switch ( newstate )
	{
	case TS_OPENING:	udp = targetp->cp->userdatap;		break;
	case TS_CLOSING:	udp = targetp->discon_userdatap;	break;
	case TS_PENDSUSPEND:
	default:		udp = (struct aUserData *) NULL;	break;
	} /* end of switch newstate */


	if (    (targetp->vlinkp EQ vlinkp)
	    AND ((set & TM (targetp->targstate[
			  (int)st2_targ_states[(int)newstate].machine])) NE 0)
	    AND (targetp->pending_reference EQ (Reference) 0)
				/* ??? obsolete */
	    /* ??? Bind if not in use & come back later */
	    AND (udp EQ messagep->UserDatap) ) /* ??? bcmp data */
	{
	    IfLog (DBG_UPDTARG)
		(void) printf ( "...selected for message\n" );

	    if ( targets_selected >= ConfigParm (targ_per_list) )
	    {
		cp->needupdate++;	/* Come back for rest */
		break;			/* out of for all targets loop */
	    }

	    if ( CtrlMsgAddTarg( messagep, targetp, vlinkp->direction ) )
	    {
		targetp->pending_reference = messagep->hoscmphdr.Ref;
					/* ??? obsolete */
		/* ??? shouldn't caller do this */
		if ( newstate EQ TS_OPENING)
		{
		    targetp->connect_reference = messagep->hoscmphdr.Ref;
		    targetp->accref_reference = (Reference) 0;
		}			/* No Accept/Refuse */

		targets_selected++;
	    }
	}
	else IfLog (DBG_UPDTARG)
	{
	    if ( (set & TM (targetp->targstate[
			(int)st2_targ_states[(int)newstate].machine])) EQ 0 )
		(void) printf ( "...Looking for states %x found %x (%d)\n",
				set, targetp->targstate[
				(int)st2_targ_states[(int)newstate].machine],
				targetp->targstate[
				(int)st2_targ_states[(int)newstate].machine] );

	    if ( targetp->vlinkp NE vlinkp )
		(void) printf ( "...Looking for vlinkp 0x%x found 0x%x\n",
			        vlinkp, targetp->vlinkp );

	    if ( targetp->pending_reference NE (Reference) 0 )
		(void) printf ( "...Reference %d still pending\n",
			        targetp->pending_reference );

	    if (    (targetp->change_reference NE (Reference) 0)
		AND (targetp->change_reference NE messagep->hoscmphdr.Ref) )
		(void) printf ( "...Required reference %d found %d \n",
			        messagep->hoscmphdr.Ref,
			        targetp->change_reference );

	    if ( targetp->discon_userdatap NE messagep->UserDatap )
		(void) printf ( "...Required UserDatap %x found %x\n",
			        messagep->UserDatap, udp );
	}
    } /* end of for all targets loop */

    return ( targets_selected );
}

/*
S* CtrlMsgCreate ()
s*
s*	Routine to get space for an ST-II Control Message.
s*
 */

  static struct aPrsSCM *		/* Next bugid 0x50702 */
CtrlMsgCreate()				/* from DoAcknowledge DoConnect DoDisconnect DoError DoHello DoHidReply DoRefuse HandleStatus st2_DoAccept */
{
    struct aPktDesc	*pp = NO_PKTDSCP;
    struct aPrsSCM	*ctrlmsg;


    ctrlmsg = GetSpace (struct aPrsSCM *, &( pp ),
			PKT_PARSEDSCM, PKT_MD_NEW, PKT_XA_PS,
			0,0, sizeof (struct aPrsSCM), 0,0);
    if ( ctrlmsg EQ NO_PSCMP )
    {
#ifdef  STIIDEBUG
	(void) printf ( "*** ST CtrlMsgCreate failed for %d bytes\n",
			sizeof (struct aPrsSCM) );
#endif  STIIDEBUG
	BUGSTOP (-,1,0x50701);

	return ( NO_PSCMP );
    }

    Bzero (ctrlmsg,sizeof (*ctrlmsg));

    ctrlmsg->msgcnt = 1;
    ctrlmsg->selfp = pp;
    ctrlmsg->targlst.allocated = TARG_PER_SCMP_DEF;

    return ( ctrlmsg );
}

/*
S* CtrlMsgDel ( pscmp )
 */

  static void				/* Next bugid 0x50802 */
CtrlMsgDel( pscmp )			/* from TargByApplIfc CompAccept CompConnect CompDisconnect CompRefuse CtrlMsgDropAcked CtrlMsgRelay CtrlMsgSend*2 DoConnect*2 DoDisconnect*2 drop_pending_CtrlMsg HandleStatus ReplyMsgSend*3 st2_CMPInput st2_CMPOutput*5 st2_PCBDel */
  struct aPrsSCM	*pscmp;
{
    struct aPktDesc	*pktp   = pscmp->pktp,
			*trashp = pscmp->trashp,
			*selfp  = pscmp->selfp;
    struct aST2pcb	*cp     = pscmp->cp;


    if ( pscmp->msgcnt <= 0 )
    {
	IfLog (DBG_DTA_STR)
	    (void) printf ( "ST2: delete of unreferenced parsed SCM (%x)!\n",
			   pscmp );
	BUGSTOP (-,pscmp,0x50801);
	return;
    }

    /*
     * When the reference count drops to zero, it's time to free
     * the space allocated for this message.  You knew that.  And
     * yes, we're paranoid: even though the space associated with
     * the parse block and message are likely to be freed when we
     * free "selfp," we careful to zero the pointers before doing
     * any calls to free memory because... because too many years
     * of programming have taught us to be careful.
     */

    if ( --pscmp->msgcnt > 0  )
	return;

    pscmp->pktp   = NO_PKTDSCP;
    pscmp->selfp  = NO_PKTDSCP;
    pscmp->trashp = NO_PKTDSCP;

    if ( cp NE NO_PCBP )
    {
	pscmp->cp = NO_PCBP;
	cp->refcnt--;
	IfLog (DBG_CNT_PCB)
	    printf ( "refcnt: cp = %8x  refcnt = %2d  CtrlMsgDel       pscmp %x\n",
		    cp, cp->refcnt, pscmp );
	if ( cp->refcnt EQ 0 )
	    st2_PCBDel( cp );		/* We're last user, cp now invalid */
    }

    if ( pktp NE NO_PKTDSCP )
	FreePkts (pktp);		/* ??? list */

    if ( trashp NE NO_PKTDSCP )
	FreePkts (trashp);

    if ( selfp NE NO_PKTDSCP )
	FreePkts (selfp);

    return;
}


/*
S* CtrlMsgDropAcked ( reply, ref )
s*
s*	Routine to scan and remove from the retransmission queue a message
s*	which has just been acknowledged.
s*
 */

  static int				/* Next bugid 0x50900 */
CtrlMsgDropAcked( reply, ref )		/* from st2_CMPInput */
  struct aPrsSCM	*reply;		/* Caller retains */
  octet2		 ref;
{
    struct aPrsSCM	*listp = NO_PSCMP,
			*prvp,
			*pscmp;
    int			 n = 0,
			 oldpri;


    /* Collect all acknowledged messages */

    oldpri = splnet ();
    {
	ForAllRxQ (prvp,pscmp)
	{
	    if (   (pscmp->hoscmphdr.SVLId NE reply->hoscmphdr.RVLId)
		OR (pscmp->hoscmphdr.Ref   NE ref) )
	    {
		prvp = prvp->nextp;
		continue;
	    }

	    pscmp->nextp->prevp = pscmp->prevp;
	    pscmp->prevp->nextp = pscmp->nextp;
	    pscmp->prevp = NO_PSCMP;

	    pscmp->nextp = listp;
	    listp = pscmp;
	    n++;

	} /* end of for all rx q */
    }
    (void) splx ( oldpri );

    /* -------------------------------------------------------------------- */
    /* Call completion routine (if any) & release message */

    while ( listp NE NO_PSCMP )
    {
	pscmp = listp;
	listp = listp->nextp;

	IfLog (DBG_CTL_MSG)
	    (void) printf ( "Reply received to %x scm op %d, ref %d after %d tries\n",
			    pscmp, pscmp->hoscmphdr.OpCode,
			    pscmp->hoscmphdr.Ref, pscmp->tries);

	if ( pscmp->comp NE (void (*)()) NULL )
	{
	    IfLog (DBG_MSG_CMP)
		printf ( "Calling message %8x (op %u) completion routine %8x after receiving message %8x (op %u)\n",
			 pscmp, pscmp->hoscmphdr.OpCode, pscmp->comp, reply,
			 reply->hoscmphdr.OpCode );

		/* CompAccept() CompConnect() CompDisconnect() CompRefuse()  */
	    pscmp->comp ( pscmp, reply /*,NoError*/ );
	    /* It CtrlMsgDels pscmp, we keep reply */
	}
	else
	    CtrlMsgDel( pscmp );	/* decrement pcb's use count */

    } /* end of while */

    return ( n );
}

/*
S* CtrlMsgRelay ( pscmp, vlinkp )
s*
s*	Routine to pass pending mesages, such as ACCEPTs, on to the
s*	previous-hop.
s*
 */

  static enum ST2Errors			/* Next bugid 0x51003 */
CtrlMsgRelay( pscmp, vlinkp )		/* from TargByApplIfc HandleAccept */
					/* only upstream */
  struct aPrsSCM	*pscmp;		/* Caller keeps (a ref) */
  struct aVLink		*vlinkp;
{
    enum ST2Errors	 reason;
    int			 mtu;
    long		 rcvrytoeff,	/* effective recovery timeout */
			 rcvrytonbr;	/* recovery timeout from neighbor */
    struct aFlowSpec3	*fsp;
    struct aST2pcb	*cp;


    kassert (pscmp->pktp EQ NO_PKTDSCP);

    cp = vlinkp->cp;
    if ( cp EQ NO_PCBP )
    {
	IfLog (DBG_INI_DS_MSG)
	    printf ( "CtrlMsgRelay fails - no vlinkp->cp\n" );

	BUGRETURN (-,InvalidPointer,0x51001,enum ST2Errors);
    }

    share_aPrsSCM (pscmp, BUGSTOP (-,(pscmp),0x51002));

    pscmp->hoscmphdr.Ref = NxtRefNum ();
    pscmp->hoscmphdr.RVLId = vlinkp->remote_vlid;
    pscmp->hoscmphdr.SVLId = vlinkp->local_vlid;
    pscmp->hoscmphdr.SndrIPAdr = vlinkp->lclIPAdr;

    pscmp->nxthopp = vlinkp->route.nxthopp;


    /* Begin RecoveryTimeout on upstream vlink */

    if (    (vlinkp->direction EQ UPSTREAM_VLINK)
	AND (pscmp->hoscmphdr.OpCode EQ (octet1) OpAccept)
	AND (vlinkp->np NE NO_NBRP)
	AND (pscmp->FlowSpecp->FlowVer EQ STFSVer3) )
    {
	fsp = (struct aFlowSpec3 *) pscmp->FlowSpecp;
	rcvrytonbr = ntoh2 (fsp->RecoveryTimeout);
	rcvrytoeff = NbrInsert( vlinkp, (unsigned int) rcvrytonbr );
	/* sets vlinkp->rcvryto */
	if ( rcvrytoeff > rcvrytonbr )
	    fsp->RecoveryTimeout = hton2 (rcvrytoeff);
    }


    mtu = (vlinkp->direction EQ UPSTREAM_VLINK)
	  ? vlinkp->route.nxthopp->gifcp->ifc_mtu : cp->mtu;

    reason = SCMPEncode( mtu, 0, pscmp, &( pscmp->pktp ), &( pscmp->pdulen ) );
    if ( reason NE NoError )
    {
	IfLog (DBG_INI_DS_MSG)
	    printf ( "CtrlMsgRelay fails - encode error %u\n", reason );

	CtrlMsgDel( pscmp );
	return ( reason );
    }

    reason = st2_CMPOutput( pscmp, vlinkp );

    IfLog (DBG_INI_DS_MSG)
	printf ( "CtrlMsgRelay - st2_CMPOutput returns %u\n", reason );

    return ( reason );
}

/*
S* CtrlMsgSend ( request, vlinkp )
s*
s*	Routine to sned an SCMP Control message over a specified virtual link.
s*	??? need to return error; arg for number of targets that would fit.
s*	??? caller may have bumped msgcnt to use target count to change state.
s*
 */

  static enum ST2Errors			/* Next bugid 0x51102 */
CtrlMsgSend( request, vlinkp )		/* from DoAcknowledge DoConnect DoDisconnect DoError DoHello DoHidReply DoRefuse st2_DoAccept */
  struct aPrsSCM	*request;	/* unencoded message, we free it */
  struct aVLink		*vlinkp;
{
    enum ST2Errors	 result;
    int			 opcode,
			 refnum;
    octet4		 mandatory[ 1 ];


    if ( vlinkp EQ NO_VLINKP )
    {
	/* ??? should be able to reply to some with vlinkp EQ 0,
	       based on received message */
	CtrlMsgDel( request );

	BUGRETURN (-,InvalidPointer,0x51101,enum ST2Errors);
    }

    /* ??? if no vlinkp, invert from request */
    request->hoscmphdr.RVLId = vlinkp->remote_vlid;
    request->hoscmphdr.SVLId = vlinkp->local_vlid;
    request->hoscmphdr.SndrIPAdr = vlinkp->lclIPAdr;

    /* us, as addressed in msg/ifnetp/ifaddr */
    if (    (request->Namep EQ (struct aName *) NULL)
	AND (vlinkp->cp NE NO_PCBP) ) /* skip for HELLO, etc */
    {
	request->Namep = &( request->Name );

	/* ??? if no vlinkp, if cp, copy ptr from request */
	Bcopy (&( vlinkp->cp->Name ),request->Namep,sizeof (struct aName));
    }

    mandatory[0] = st2_parmManByOp[ (int) request->hoscmphdr.OpCode ];
    if (    (IsSet32 (mandatory, (int) STDetectorIPAdr))
	AND (request->detectoripadrp EQ (INETADDR) 0) )
    {
	request->DetectorIPAdr = htonIP (request->hoscmphdr.SndrIPAdr);
	request->detectoripadrp = &( request->DetectorIPAdr );
     }

    opcode = (int) request->hoscmphdr.OpCode;
    refnum = request->hoscmphdr.Ref;

    request->nxthopp = vlinkp->route.nxthopp;
    request->pktp = NO_PKTDSCP;
    request->pdulen = 0;

    /* ??? need to get back more to do flag */
    result = SCMPEncode( vlinkp->route.nxthopp->gifcp->ifc_mtu, 0,
			  request, &( request->pktp ),
	/* SCMPEncode () */ &( request->pdulen ) );

    if ( result NE NoError )
    {
#ifdef  STIIDEBUG			/* ??? move to encoder */
	(void) printf ( "Encode failed: reason code %d\n", result );
#endif  STIIDEBUG

	CtrlMsgDel( request );
    }
    else
    {
	result = st2_CMPOutput( request, vlinkp );
    }					/* It unlinks request */

    IfLog (DBG_CTL_MSG)
    {
	log_VLink( vlinkp, "CtrlMsgSend" );
	(void) printf ( "     message (%x) op %d, ref %d error %d\n",
		       request, opcode, refnum, result );
    }

    return ( result );			/* ??? # targets sent, if no error */
}

/*
S* DoAcknowledge ( request, vlinkp, reason )
s*
s*	Routine to send an ACK in response to a received SCMP message.
s*
 */

  static int				/* Next bugid 0x51200 */
DoAcknowledge( request, vlinkp, reason ) /* from HandleAccept HandleConnect HandleDisconnect HandleHello HandleRefuse */
  struct aPrsSCM	*request;	/* caller keeps */
  struct aVLink		*vlinkp;
  enum ST2Errors	 reason;
{
    struct aNeighbor	*np;		/* for diagnostic */
    struct aST2pcb	*cp;		/* for diagnostic */
    struct aPrsSCM	*ackMsgp;


    IfLog (DBG_INI_DS_MSG)
    {
	if ( vlinkp NE NO_VLINKP )
	{
	    cp = vlinkp->cp;
	    np = vlinkp->np;
	}
	else
	{
	    cp = (struct aST2pcb *) 0xFFFFFFFF;
	    np = (struct aNeighbor *) 0xFFFFFFFF;
	}
	(void) printf ( "DoAcknowledge cp = %x  np = %x  vlinkp = %x  op = %u  ref = %u  reason = %u\n",
		        cp, np, vlinkp, request->hoscmphdr.OpCode,
		        request->hoscmphdr.Ref, reason );
    }

    ackMsgp = CtrlMsgCreate();
    if ( ackMsgp EQ NO_PSCMP )
    {
#ifdef  STIIDEBUG
	(void) printf ( "DoAcknowledge: out of memory for reply\n" );
#endif  STIIDEBUG
	return ( 0 );			/* maybe have space when rx'd */
    }

    ackMsgp->hoscmphdr.OpCode = (octet1) OpAck;
    ackMsgp->hoscmphdr.Ref = request->hoscmphdr.Ref;
    ackMsgp->hoscmphdr.LnkRef = (Reference) 0;
    ackMsgp->ReasonCode = (octet2) FormalReasonFor (reason);
    ackMsgp->reasoncodep = &( ackMsgp->ReasonCode );

    if (    (vlinkp NE NO_VLINKP)
	AND (vlinkp->vlink_cnt > 0) )
	/* ??? reason = */ (void) CtrlMsgSend( ackMsgp, vlinkp );
    else
	/* ??? reason = */ (void) ReplyMsgSend( request, ackMsgp );

    return ( 0 );
}

/*
S* DoConnect ( targetp )
s*
s*	Called in the TS_WANTOPEN state to build and send an SCMP CONNECT
s*	message.
s*
 */

  static enum ST2Errors			/* Next bugid 0x51307 */
DoConnect( targetp )			/* from st2_PCBUpdate */
  struct Target		*targetp;
{
    enum ST2Errors	 error;
    struct aTargetDesc	*tdescp;
    int			 n;
    enum targ_state	 nxtstate;
    struct aST2pcb	*cp = targetp->cp;
    struct aVLink	*vlinkp = targetp->vlinkp;
    struct aPrsSCM	*connmsgp;


    IfLog (DBG_INI_DS_MSG)
	log_Target( targetp, "DoConnect" );

    kassert (vlinkp NE NO_VLINKP);
    kassert (vlinkp->next_hopp NE NO_NEXTHOPP);

    connmsgp = CtrlMsgCreate();
    if ( connmsgp EQ NO_PSCMP )
    {
#ifdef  STIIDEBUG
	(void) printf ( "DoConnect: out of memory for target %x\n",
		       targetp );
#endif  STIIDEBUG
	return ( CantGetResrc );	/* ??? how does it get retried later */
    }

    connmsgp->comp = CompConnect;	/* Completion routine */

    connmsgp->hoscmphdr.OpCode = (octet1) OpConnect;
    connmsgp->hoscmphdr.Options = cp->optapi;
    connmsgp->hoscmphdr.Ref = NxtRefNum ();
    connmsgp->hoscmphdr.LnkRef = (Reference) 0;

    /* ??? Multicast handling will be needed here. */

    if ( cp->detectorIPAdr NE INADDR_ANY)
    {
	connmsgp->DetectorIPAdr = (octet4) cp->detectorIPAdr;
	connmsgp->detectoripadrp = &( connmsgp->DetectorIPAdr );
    }
    else
	BUGSTOP (-,cp,0x51301);		/* ??? ought to be bound by now! */

    connmsgp->Originp = &( cp->Origin );

    /* ??? vector by version */
    (void) FlowSpecEncode( connmsgp, (struct aFlowSpec *) &( cp->flow_spec ) );

    /* ???RFlowSpec */

    connmsgp->UserDatap = cp->userdatap;

    /* ??? what if not hid_info.hid_set */
    connmsgp->HID = (octet2) vlinkp->next_hopp->hid_info.hop_id;
    connmsgp->hdrhidp = &( connmsgp->HID );

    IfLog (DBG_INI_DS_MSG)
	log_Parms( connmsgp );

    /* Add target names for others on same next hop */
    if ( CtrlMsgAddTrgSet( connmsgp, cp, vlinkp,
		       STATESET (TS (TS_WANTOPEN)), TS_OPENING ) <= 0 )
    {					/* No matching targets */
	BUGSTOP (-,vlinkp,0x51302);	/* ??? probably a bug */
	CtrlMsgDel( connmsgp );	
	return ( NoError );		/* ??? return error message */
    }

    /* Only make state change for those targets successfully sent (no
     * error from CtrlMsgSend); get back count of targets that would fit;
     * needupdate if some didn't.
     */

    /* Don't let it get freed */
    share_aPrsSCM (connmsgp, BUGSTOP (-,(connmsgp),0x51303));

    error = CtrlMsgSend( connmsgp, vlinkp ); /* It frees connmsgp */


    tdescp = &( connmsgp->targlst.TargetDescs[0] );
    n = connmsgp->targlst.nxtfree;

    for ( ; n > 0 ; tdescp++, n-- )
    {
	struct Target *wtp = tdescp->targetp;


	if ( wtp EQ NO_TRGP )
	{
	    BUGSTOP (-,targetp,0x51304);
	    continue;
	}

	if ( NOT IsTargState (wtp,TS_WANTOPEN) )
	{			/* Something else happened */
	    BUGSTOP (-,wtp,0x51305);
	    continue;
	}

	if ( wtp->curchoicep EQ NO_RUTCHOICEP )
	{
	    BUGSTOP (-,wtp,0x51306);
	    continue;
	}

	if ( error EQ NoError )		/* ??? vs targetp->error */
	{
	    if ( wtp->curchoicep->type EQ TO_APPL )
		nxtstate = TS_OPEN;
	    else
		nxtstate = TS_OPENING;
	}
	else
	{
	    /* ??? function of reason */
	 /* wtp->curchoicep->flags |= RouteFailed | RouteNoRsrc; */
	    wtp->curchoicep->flags |= RouteFailed | RouteNoRoute;

	    wtp->error = error;

	    /* ??? retry or force to TS_FAILED */
					/* ??? NoRecovery */
	    if ( NOT IsTargStateIn (wtp,TS_DISCONNECTED,
				   TS (TS_REFUSING) TS (TS_FAILED)) )
	    {
		SetTargState (wtp,TS_FAILED);
	    }

	    wtp->cp->needupdate++;	/* Need to run state machine */
	    nxtstate = TS_CLOSED;	/* assume nothing was sent */
	}

	SetTargState (wtp, nxtstate);

    } /* end of all targets loop */

    CtrlMsgDel( connmsgp );

    return ( NoError );			/* I.e., state update succeeded */
}

/*
S* DoDisconnect ( targetp, reason )
s*
s*	Routine to build and send an SCMP DISCONNECT message.
s*	??? individual target or G-Bit
s*
 */

  static enum ST2Errors			/* Next bugid 0x51406 */
DoDisconnect( targetp, reason )		/* from st2_PCBUpdate*2 */
  struct Target		*targetp;
  enum ST2Errors	 reason;
{
    enum ST2Errors	 error;
    struct aTargetDesc	*tdescp;
    int			 n;
    enum targ_state	 nxtstate;
    struct aST2pcb	*cp = targetp->cp;
    struct aPrsSCM	*disMsgp;


    IfLog (DBG_INI_DS_MSG)
	log_Target( targetp, "DoDisconnect" );

    targetp->error = reason;

    disMsgp = CtrlMsgCreate();
    if ( disMsgp EQ NO_PSCMP )
    {
#ifdef  STIIDEBUG
	(void) printf ( "DoDisconnect: out of memory for target %n\n",
		       targetp );
#endif  STIIDEBUG
	return ( CantGetResrc );	/* ??? how does it get retried later */
    }

    disMsgp->comp = CompDisconnect;	/* Completion routine */

    disMsgp->hoscmphdr.OpCode = (octet1) OpDisconnect;
    disMsgp->hoscmphdr.Options = 0;
    disMsgp->hoscmphdr.Ref = NxtRefNum ();
    disMsgp->hoscmphdr.LnkRef = (Reference) 0;
    disMsgp->ReasonCode = (octet2) FormalReasonFor (reason);
    disMsgp->reasoncodep = &( disMsgp->ReasonCode );
    disMsgp->DetectorIPAdr = (octet4) cp->detectorIPAdr; /* ??? */
    disMsgp->detectoripadrp = &( disMsgp->DetectorIPAdr );

    disMsgp->UserDatap = targetp->discon_userdatap;


    /* Add target names for others on same next hop */
    /* ??? of TS_WANTSUSPEND -> TS_PENDSUSPEND */
    if ( CtrlMsgAddTrgSet( disMsgp, cp, targetp->vlinkp,
		       STATESET (TS (TS_WANTCLOSED)),TS_CLOSING) <= 0 )
    {
	BUGSTOP (-,targetp->vlinkp,0x51401);	/* ??? probably a bug */
	CtrlMsgDel( disMsgp );
	targetp->cp->needupdate++;	/* Need to run state machine */
	SetTargState (targetp,TS_CLOSED); /* ??? upstream side */
	return ( NoError );
    }

    /* Don't let it get freed */
    share_aPrsSCM (disMsgp, BUGSTOP (-,(disMsgp),0x51402));

    /* Send message on to next hop */
    error = CtrlMsgSend( disMsgp, targetp->vlinkp ); /* It keeps disMsgp */

    /* for all targets in message, but only if sent */

    tdescp = &( disMsgp->targlst.TargetDescs[0] );
    n = disMsgp->targlst.nxtfree;

    for ( ; n > 0 ; tdescp++, n-- )
    {
	struct Target *wtp = tdescp->targetp;


	if ( wtp EQ NO_TRGP )
	{
	    BUGSTOP (-,targetp,0x51403);
	    continue;
	}

	if ( wtp->curchoicep EQ NO_RUTCHOICEP )
	{
	    BUGSTOP (-,wtp,0x51404);	/* ??? what to do */
	    continue;
	}

	if ( IsTargState (wtp,TS_WANTCLOSED) )
	{
	    if ( error EQ NoError )	/* ??? vs targetp->error */
	    {
		if ( wtp->curchoicep->type EQ TO_APPL )
		{
		    nxtstate = TS_CLOSED; /* No ACK from socket */
		    if ( NOT IsTargStateIn (wtp,TS_DISCONNECTED,
					   TS (TS_REFUSING) TS (TS_FAILED)) )
		    {
			SetTargState (wtp,TS_FAILED);
		    }
		    wtp->cp->needupdate++;
		}
		else
		    nxtstate = TS_CLOSING; /* Wait for ACK */
	    }
	    else
	    {
		nxtstate = TS_CLOSED;
		wtp->cp->needupdate++;
	    }
	}
	else if ( IsTargState (wtp,TS_WANTSUSPENDED) )
	{
	    if ( error EQ NoError )	/* ??? vs targetp->error */
	    {
		if ( wtp->curchoicep->type EQ TO_APPL )
		{
		    nxtstate = TS_CLOSED;
		    if ( NOT IsTargStateIn (wtp,TS_DISCONNECTED,
					   TS (TS_REFUSING) TS (TS_FAILED)) )
		    {
			SetTargState (wtp,TS_FAILED);
		    }
		    wtp->cp->needupdate++;
		}
		else
		    nxtstate = TS_PENDSUSPEND;
	    }
	    else
		nxtstate = TS_SUSPENDED;
	}
	else
	{
	    BUGSTOP (-,wtp,0x51405);
	    nxtstate = TS_CLOSED;
	}

	SetTargState (wtp, nxtstate);

    } /* end of all targets loop */

    CtrlMsgDel( disMsgp );    

    return ( NoError );			/* I.e., state update succeeded */
}

/*
S* DoError ( request, vlinkp, reason )
s*
s*	Routine to send an SCMP ERROR-IN-[REQUEST,RESPONSE] message, or
s*	drop it, as is appropriate for the message containing the error.
s*
 */

  static enum ST2Errors			/* Next bugid 0x51502 */
DoError( request, vlinkp, reason )	/* from HandleHidApprove HandleHidReject*2 st2_CMPInput */
  struct aPrsSCM	*request;	/* caller keeps */
  struct aVLink		*vlinkp;	/* May be NULL */
  enum ST2Errors	 reason;
{
    enum OpCodes	 opcode;
    struct aPrsSCM	*nack;
    struct sockaddr_in	*sip = (struct sockaddr_in *) &( request->rcvingsa );


    switch ( request->hoscmphdr.OpCode )
    {
    /* Commands take error in request. */
    case OpConnect:
    case OpDisconnect:
    case OpAccept:
    case OpRefuse:
    case OpHidChange:
    case OpHidChangeRequest:
    case OpChange:
    case OpChangeRequest:
    case OpStatus:
    case OpHello:
    case OpNotify:
	opcode = OpErrorInRequest;
	break;

    /* Replies take error in response. */
    case OpHidApprove:
    case OpHidReject:
    case OpAck:
	/* ??? StatusResponse */
	opcode = OpErrorInResponse;
	break;

    /* Don't ack error in request or error in response. */
    case OpErrorInRequest:
    case OpErrorInResponse:
    default:
	IfLog (DBG_INI_DS_MSG)
	    printf ( "DoError: request %x, opcode %u, reason %u ignored\n",
		    request, request->hoscmphdr.OpCode, reason, vlinkp );

	BUGSTOP (-,(request->hoscmphdr.OpCode << 16) | (int) reason,0x51501);
	return ( NoError );
    } /* end of opcode switch */


    IfLog (DBG_INI_DS_MSG)
	printf ( "error_in_%s_message: request %x, opcode %u, reason %u via vlink %x\n",
		(opcode EQ OpErrorInResponse) ? "Response" : "Request",
		request, request->hoscmphdr.OpCode, reason, vlinkp );


    nack = CtrlMsgCreate();
    if ( nack EQ NO_PSCMP )
    {
#ifdef  STIIDEBUG
	(void) printf ( "DoError: out of memory for error in %n\n",
		        request->hoscmphdr.OpCode );
#endif  STIIDEBUG
	return ( CantGetResrc );	/* ??? how does it get retried later */
    }

    nack->hoscmphdr.OpCode = (octet1) opcode;
    nack->hoscmphdr.Ref = request->hoscmphdr.Ref;
    nack->hoscmphdr.LnkRef = request->hoscmphdr.LnkRef;
    nack->ReasonCode = (octet2) FormalReasonFor (reason);
    nack->reasoncodep = &( nack->ReasonCode );
    nack->DetectorIPAdr = sip->sin_addr.s_addr;
    nack->detectoripadrp = &( nack->DetectorIPAdr );

#ifdef NOTYET
|	    /*?? Include errored PDU if there's room? */
|	    /* Problem is sticking pcode info in from of pdu. */
|	    nack->ErrdPDUp = xxx;
#endif NOTYET

    if (    (vlinkp NE NO_VLINKP)
	AND (vlinkp->vlink_cnt > 0) )
	/* ??? reason = */ (void) CtrlMsgSend( nack, vlinkp );
    else
	/* ??? reason = */ (void) ReplyMsgSend( request, nack );

    return ( NoError );
}

/*
S* DoHello ( np, tvp )
s*
s*	Send a HELLO message to a neighbor.
s*
 */

  static void				/* Next bugid 0x51602 */
DoHello( np, tvp )			/* from st2_ClkSlow */
  struct aNeighbor	*np;
  struct timeval	*tvp;
{
    long		 lu;
    Reference		 refnum;
    struct aPrsSCM	*hellop;


    if ( (hellop = CtrlMsgCreate()) EQ NO_PSCMP )
    {
#ifdef  STIIDEBUG
	(void) printf ( "DoHello: out of memory for reply\n" );
#endif  STIIDEBUG
	BUGSTOP (-,CantGetResrc,0x51601);
	return;
    }

    /* "Basic" hello parameters */
    hellop->hoscmphdr.OpCode = (octet1) OpHello;
    hellop->hellotimerp = &( st2_stats.hellomsec );
    if ( ConfigParm (hellotimerholddown) > st2_stats.hellomsec )
	hellop->hoscmphdr.Options = RBit;

    /* "once in a while" measure RTT to neighbor by asking for ACK of HELLO. */
    lu = MsecBetween (tvp,&(np->rtthello_sent)); /* our time delta */
    if (   (np->rtthello_sent.tv_sec EQ 0)
	OR (lu > np->rtthello_ms) )
    {
	refnum = NxtRefNum ();
	hellop->hoscmphdr.Ref = refnum;	 /* ask for ack */
	np->rtthello_refnum = refnum;	/* indicates waiting for ack, too */
	np->rtthello_sent = *tvp;
    }

    if ( CtrlMsgSend( hellop, np->vlinkp ) EQ NoError )
    {
	np->totsent++;
	Count (hellosent,1);
    }

    return;
}

/*
S* DoHidReply ( messagep, vlinkp, opcode, hop_id )
s*
s*	Routine to send a HID-APPROVE or HID-REJECT to the previous-hop.
s*
 */

  static int				/* Next bugid 0x51700 */
DoHidReply( messagep, vlinkp, opcode, hop_id ) /* from HandleConnect */
					/* ??? INLINE */
  struct aPrsSCM	*messagep;
  struct aVLink		*vlinkp;
  enum OpCodes		 opcode;
  HID			 hop_id;
{
    Reference		 reference = (Reference)messagep->hoscmphdr.Ref;
    struct aPrsSCM	*hidreply;
    octet2		 hdrhid = (octet2) hop_id;
    octet4		 mask;
    octet4		*maskp;
    int			 shift;
    int			 maskcnt;


    hidreply = CtrlMsgCreate();
    if ( hidreply EQ NO_PSCMP )
    {
#ifdef  STIIDEBUG
	(void) printf ( "DoHidReply: out of memory for reply\n" );
#endif  STIIDEBUG
	return ( -1 );
    }

    hidreply->hoscmphdr.OpCode = (octet1) opcode;
    hidreply->hoscmphdr.Options = 0;
    hidreply->hoscmphdr.Ref = reference;
    hidreply->hoscmphdr.LnkRef = (Reference) 0;
    hidreply->hdrhidp = &hdrhid;
    hidreply->hoscmphdr.SndrIPAdr = vlinkp->lclIPAdr;


    /*
     * In what follows, remember that hid value is stored in uniform
     *  SCMP header so it is in host byte order.  The bits going into
     * the mask for the FreeHIDs parameter need to be converted to
     * network byte order.
     */
    maskcnt = ConfigParm (fhid_masks);
    hidreply->FreeHIDsp = (struct aFreeHIDs *) &( hidreply->freehids );
    hidreply->freehids.pcode = hton1 ((octet1) STpFreeHIDs);
    hidreply->freehids.plen = hton1 (sizeof (hidreply->freehids));
    hidreply->freehids.basehid = hton2 ((octet2) (hop_id & ~ 0x1F));

    /*
     * Bit 0 of first mask corresponds to hid with value equal
     *  to BaseHID with the 5 least significant bits set to zero.
     */
    hop_id &= ~0x1f;
    maskp = &( hidreply->freehids.freehidbitmask[0] );
    while ( --maskcnt >= 0 )
    {
	mask = 0;
	for ( shift = 31 ; shift >= 0 ; shift-- )
	{
	    if ( st2_ConHsh[ ConfigParm (hid_mask) & hop_id ].cp EQ NO_PCBP )
	    {
		mask |= (1 << shift);
	    }
	    hop_id++;
	} /* end of for loop */
	*maskp++ = hton4 (mask);
    } /* end of while loop */

    /* ??? reason = */ (void) CtrlMsgSend( hidreply, vlinkp );

    return ( 0 );
}

/*
S* DoRefuse ( targetp, comp )
s*
s*	Send an SCMP REFUSE message to the previous-hop.
s*	Caller responsible for state transisitions.
s*
 */
					/* ??? propagating or initiating */
					/*     keep detector  use our IPAdr */

  static enum ST2Errors			/* Next bugid 0x51802 */
DoRefuse( targetp, comp )		/* from st2_PCBUpdate*2 */
  struct Target		*targetp;
  void			(*comp)(/* ctlmsg, replymsg */);
{
    struct aST2pcb	*cp = targetp->cp;
    struct aVLink	*vlinkp = cp->upstream_vlinkp; /* ??? might be gone */
    struct aPrsSCM	*refuse;
    enum ST2Errors	 reason;


    IfLog (DBG_INI_DS_MSG)
	(void) printf ( "DoRefuse\n" );
    IfLog (DBG_UPDTARG)
	log_Target( targetp, "DoRefuse" );

    if ( (refuse = CtrlMsgCreate()) EQ NO_PSCMP )
    {
#ifdef  STIIDEBUG
	(void) printf ( "DoRefuse: out of memory\n" );
#endif  STIIDEBUG
	IfLog (DBG_UPDTARG)
	    log_Target( targetp, "DoRefuse - no buffers" );
	return ( CantGetResrc );
    }

    kassert (targetp->error NE NoError);

    refuse->comp = comp;		/* Completion routine */

    refuse->DetectorIPAdr = vlinkp->lclIPAdr;
    refuse->detectoripadrp = &( refuse->DetectorIPAdr );
    refuse->hoscmphdr.OpCode = (octet1) OpRefuse;
    refuse->hoscmphdr.Ref = NxtRefNum ();
    refuse->hoscmphdr.LnkRef = (targetp->lnkrefflg)
				? targetp->linked_reference : 0;
    refuse->ReasonCode = (octet2) FormalReasonFor (targetp->error);
					/* ??? demanded NoError above */
    refuse->reasoncodep = &( refuse->ReasonCode );

    refuse->UserDatap = targetp->refuse_userdatap;

    if ( CtrlMsgAddTarg( refuse, targetp, vlinkp->direction ) )
    {
	reason = CtrlMsgSend( refuse, vlinkp );
					/* If CtrlMsgDels refuse */
	targetp->lnkrefflg = 0;
    }
    else
    {
#ifdef  STIIDEBUG
	(void) printf ( "Failed to add target to refuse\n" );
#endif  STIIDEBUG
	BUGSTOP (-,targetp,0x51801);
	reason = ErrorUnknown;
	targetp->error = reason;
    }

    return ( reason );
}

/*
S* drop_hid_Forwarder ( cp, hop_id )
s*
s*	NOTYET
s*
 *   Returns: true (non-zero) if hid dropped; false (zero) if
 *	hid is not actually associated with this forwarder
 */
#ifdef NOTYET
|	
|	  void
|	_drop_hid_Forwarder ( cp, hop_id ) /* from <nowhere> */
|	  struct aST2pcb	*cp;
|	  HID			 hop_id;
|	{
|	    int			 lstidx;
|	
|	    kassert (hop_id NE (HID) 0);
|	
|	    PTRCHECK (cp, struct aST2pcb *, return);
|	    kassert (Forwarder_for_HID (hop_id) EQ cp);
|	
|	    HIDClear ( hop_id );
|	    for ( lstidx = 0 ; lstidx < DimensionOf (cp->hop_id) ; lstidx++ )
|	    {
|		if ( cp->hop_id[ lstidx ] EQ hop_id )
|		{
|		    cp->hop_id[ lstidx ] = (HID) 0;
|		    return;
|		}
|	    }
|	    return;
|	}
#endif NOTYET

/*
S* drop_pending_CtrlMsg ( cp )
s*
s*	Routine to remove all SCMP messages related to a specified connection
s*	from the retransmission queue.
s*
 */

  static int				/* Next bugid 0x51900 */
drop_pending_CtrlMsg( cp )		/* from st2_PCBDel */
					/* on flush of cp */
  struct aST2pcb 	*cp;
{
    struct aPrsSCM	*listp = NO_PSCMP,
			*prvp,
			*pscmp;
    int			 n = 0,
			 oldpri;

    oldpri = splnet ();
    {
	ForAllRxQ (prvp,pscmp)
	{
	    if ( pscmp->cp NE cp )
	    {
		prvp = prvp->nextp;
		continue;
	    }

	    pscmp->nextp->prevp = pscmp->prevp;
	    pscmp->prevp->nextp = pscmp->nextp;
	    pscmp->prevp = NO_PSCMP;

	    pscmp->nextp = listp;
	    listp = pscmp;
	    n++;

	} /* end of for all rx q */
    }
    (void) splx ( oldpri );


    while ( listp NE NO_PSCMP )
    {
	pscmp = listp;
	listp = listp->nextp;

	IfLog (DBG_CTL_MSG)
	    (void) printf ( "Dropped scmp request %d, ref %d after %d tries\n",
			    pscmp->hoscmphdr.OpCode, pscmp->hoscmphdr.Ref,
			    pscmp->tries );

	CtrlMsgDel( pscmp );

    } /* end of while */

    return ( n );
}

/*
S* FlowSpecEncode ( messagep, hflowp )
s*
s*	Routine to export FlowSpec information into an SCMP Parameter.
s*
 */

  static boolean			/* Next bugid 0x52000 */
FlowSpecEncode( messagep, hflowp )	/* from DoConnect st2_DoAccept */
  struct aPrsSCM	*messagep;
  struct aFlowSpec	*hflowp;
{
    struct aFlowSpec3	*flowparmp = &( messagep->flowspec ); /* ??? ptr */
    struct aFlowSpec3	*hfp = (struct aFlowSpec3 *) hflowp;


    flowparmp->pcode            = hton1 ((octet1) STpFlowSpec);
    flowparmp->plen             = hton1 (sizeof (struct aFlowSpec3));
    flowparmp->FlowVer          = hton1 (STFSVer3);
    flowparmp->fill1            = 0;
    flowparmp->DutyFactor       = hton1 (hfp->DutyFactor);
    flowparmp->ErrorRate        = hton1 (hfp->ErrorRate);
    flowparmp->Precedence       = hton1 (hfp->Precedence);
    flowparmp->Reliability      = hton1 (hfp->Reliability);
    flowparmp->Tradeoffs        = hton2 (hfp->Tradeoffs);
    flowparmp->RecoveryTimeout  = hton2 (hfp->RecoveryTimeout);
    flowparmp->LimitOnCosts     = hton2 (hfp->LimitOnCosts);
    flowparmp->LimitOnDelay     = hton2 (hfp->LimitOnDelay);
    flowparmp->LimitOnPDUBytes  = hton2 (hfp->LimitOnPDUBytes);
    flowparmp->LimitOnPDURate   = hton2 (hfp->LimitOnPDURate);
    flowparmp->MinBytesXRate    = hton4 (hfp->MinBytesXRate);
    flowparmp->AccdMeanDelay    = hton4 (hfp->AccdMeanDelay);
    flowparmp->AddDelayVariance = hton4 (hfp->AddDelayVariance);
    flowparmp->DesPDUBytes      = hton2 (hfp->DesPDUBytes);
    flowparmp->DesPDURate       = hton2 (hfp->DesPDURate);

    messagep->FlowSpecp = (struct aFlowSpec *) flowparmp;

    return ( TRUE );
}

/*
S* FlowSpecFromParm ( flowparmp, hflowp )
s*
s*	Routine to import a FlowSpec from an SCMP Parameter.
s*
 */

  static boolean			/* Next bugid 0x52100 */
FlowSpecFromParm( flowparmp, hflowp )	/* from HandleConnect*2 */
  struct aFlowSpec	*flowparmp;
  struct aFlowSpec3	*hflowp;
{
    struct aFlowSpec3	*fsp;


    switch ( flowparmp->FlowVer )
    {
    case STFSVer3:
	fsp = (struct aFlowSpec3 *) flowparmp;

	hflowp->pcode            = ntoh1 (flowparmp->pcode);
	hflowp->plen             = ntoh1 (flowparmp->plen);
	hflowp->FlowVer          = ntoh1 (flowparmp->FlowVer);
	hflowp->fill1            = 0;
	hflowp->DutyFactor       = ntoh1 (fsp->DutyFactor);
	hflowp->ErrorRate        = ntoh1 (fsp->ErrorRate);
	hflowp->Precedence       = ntoh1 (fsp->Precedence);
	hflowp->Reliability      = ntoh1 (fsp->Reliability);
	hflowp->Tradeoffs        = ntoh2 (fsp->Tradeoffs);
	hflowp->RecoveryTimeout  = ntoh2 (fsp->RecoveryTimeout);
	hflowp->LimitOnCosts     = ntoh2 (fsp->LimitOnCosts);
	hflowp->LimitOnDelay     = ntoh2 (fsp->LimitOnDelay);
	hflowp->LimitOnPDUBytes  = ntoh2 (fsp->LimitOnPDUBytes);
	hflowp->LimitOnPDURate   = ntoh2 (fsp->LimitOnPDURate);
	hflowp->MinBytesXRate    = ntoh4 (fsp->MinBytesXRate);
	hflowp->AccdMeanDelay    = ntoh4 (fsp->AccdMeanDelay);
	hflowp->AddDelayVariance = ntoh4 (fsp->AddDelayVariance);
	hflowp->DesPDUBytes      = ntoh2 (fsp->DesPDUBytes);
	hflowp->DesPDURate       = ntoh2 (fsp->DesPDURate);

	return ( TRUE );

    default:
	return ( FALSE );

    } /* end of switch */
}

/*
S* HandleAccept ( messagep, vlinkp )
s*
s*	SCMP handler for a received ACCEPT message.
s*
 */

  static enum ST2Errors			/* Next bugid 0x52205 */
HandleAccept( messagep, vlinkp )	/* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps */
  struct aVLink		*vlinkp;	/* valididated */
{
    enum ST2Errors	 reason;
    int			 oldpri;
    long		 rcvrytoeff,	/* effective recovery timeout */
			 rcvrytonbr;	/* recovery timeout from neighbor */
    struct aST2pcb	*cp = vlinkp->cp;
    struct aFlowSpec3	*fsp;
    struct Target	*targetp;


    if ( vlinkp EQ NO_VLINKP )
    {
	BUGSTOP (-,messagep,0x52201);
	return ( NoError );
    }

    if ( messagep->targlst.nxtfree NE 1 )	/* ??? more robust */
    {
	BUGRETURN (-,ProtocolError,0x52202,enum ST2Errors);
    }	/* To send an ErrorInRxxx */

    targetp = TargByParm( cp, DOWNSTREAM_VLINK,
				messagep->targlst.TargetDescs[0].Targetp );
    if (   (targetp EQ NO_TRGP)
	OR (targetp->connect_reference NE messagep->hoscmphdr.LnkRef) )
    {
	BUGRETURN (-,InvalidTarget,0x52203,enum ST2Errors);
    }	/* To send an ErrorInRxxx */


    if (   (targetp->accref_reference EQ messagep->hoscmphdr.Ref) /* rx */
	OR (    (targetp->accref_reference NE 0) /* stale */
	    AND (((  messagep->hoscmphdr.Ref
		   - targetp->accref_reference) & 0x8000) NE 0)) )
    {
	reason = DuplicateIgn;
    }
    else
    {
	targetp->accref_reference = messagep->hoscmphdr.Ref;
	reason = NoError;
    }

    (void) DoAcknowledge( messagep, vlinkp, reason );


    if ( reason NE NoError )		/* Ignore Duplicate */
	return ( NoError );		/* Don't send an ErrorInRxxx */


    if ( IsTargState (targetp,TS_OPENING) )
	SetTargState (targetp,TS_OPEN);


    /* Set recovery timeout to max of downstream value and permitted local */

    if (    (vlinkp NE NO_VLINKP)
	AND (vlinkp->np NE NO_NBRP)
	AND (messagep->FlowSpecp->FlowVer EQ STFSVer3) )
    {
	fsp = (struct aFlowSpec3 *) messagep->FlowSpecp;
	rcvrytonbr = ntoh2 (fsp->RecoveryTimeout);
	rcvrytoeff = NbrInsert( vlinkp, (unsigned int) rcvrytonbr );
	/* sets vlinkp->rcvryto */
	if ( rcvrytoeff > rcvrytonbr )
	    fsp->RecoveryTimeout = hton2 (rcvrytoeff);
    }

    /* Switch from "downstream" target format to "upstream" & prepare to fwd */

    messagep->targlst.TargetDescs[0].targetp = targetp;
    messagep->targlst.TargetDescs[0].Targetp = targetp->tparminp;
    messagep->hoscmphdr.LnkRef = targetp->linked_reference;
    messagep->hoscmphdr.Ref    = NxtRefNum ();

    targetp->pending_reference = (Reference) 0;	/* ??? obsolete */
    targetp->lnkrefflg = 0;

    /* Must hold ACCEPTs until sending HID is valid */
    if ( cp->hid_ready )
    {
	(void) CtrlMsgRelay( messagep, cp->upstream_vlinkp );
    }
    else
    {
	share_aPrsSCM (messagep, BUGSTOP (-,(messagep),0x52204));

	oldpri = splnet ();
	{
	    messagep->nextp = cp->accepts;
	    cp->accepts = messagep;
	}
	(void) splx ( oldpri );
    }

    return ( NoError );
}

/*
S* HandleAck ( messagep, vlinkp )
s*
s*	SCMP handler for a received ACK message.
s*	Followup on return.
s*
 */

  static enum ST2Errors			/* Next bugid 0x52303 */
HandleAck( messagep, vlinkp )		/* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps ARGSUSED */
  struct aVLink		*vlinkp;	/* May be NULL */
{
    IfLog (DBG_INI_DS_MSG)
	(void) printf ( "Acknowledgement\n" );

    if ( vlinkp EQ NO_VLINKP )
    {
#ifdef WRONGWAY
	BUGRETURN (-,ProtocolError,0x52301,enum ST2Errors);
					/* To send an ErrorInRxxx */
#else
	BUGSTOP (,((  messagep->hoscmphdr.RVLId << 16)
		    | messagep->hoscmphdr.SVLId),0x52302);
	return ( NoError );
#endif WRONGWAY
    }

    /* If Ack for HELLO and waiting for ACK, do RTT processing */
    if (    (vlinkp->type EQ NEIGHBOR_VLINK)
	AND (vlinkp->np->rtthello_refnum NE 0) )
    {
	NbrRTTUpdate( vlinkp->np, (Reference) messagep->hoscmphdr.Ref,
		       &( messagep->timestamp ) );
    }

    return ( NoError );			/* to notify_pending_Forwarder */
}

/*
S* HandleBadOpCode ( messagep, vlinkp )
s*
s*	Dummy routine called when a bad SCMP OpCode is encountered.
s*
 */

  static enum ST2Errors			/* Next bugid 0x52402 */
HandleBadOpCode( messagep, vlinkp )	/* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps */
  struct aVLink		*vlinkp;	/* ARGSUSED */
{
    BUGSTOP (,messagep->hoscmphdr.OpCode,0x52401);
    return ( OpCodeUnknown );		/* To send an ErrorInRxxx */
}


/*
S* HandleChange ( messagep, vlinkp )
s*
s*	SCMP handler for a received CHANGE message.
s*
 */

  static enum ST2Errors			/* Next bugid 0x52500 */
HandleChange( messagep, vlinkp )	/* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps */
  struct aVLink		*vlinkp;	/* Exists ARGSUSED*/
{
#ifdef NOTYET
|		int difference;
|	
|		for ( targparm in messagep.targetsp )
|		{
|			targetp = TargByParm ( cp, UPSTREAM_VLINK, targparm );
|			if ( targetp EQ NO_TRGP )
|				continue;
|			difference = targetp->change_reference - messagep->hoscmphdr.Reference;
|	
|			if (    ((difference & NBitMask (REF_BITS)) NE 0)
|			    AND ((difference & (1 << (REF_BITS-1))) NE 0) )
|			{
|				adjust_target_Forwarder ( cp, targetp,
|							&messagep->flow_spec );
|				/*?? Need some way to queue follow up--when
|					change request succeeds of fails we
|					want to send ack/nack toward origin */
|			}
|		}
#endif NOTYET
	return ( NoSupport );		/* To send an ErrorInRxxx */
}

/*
S* HandleChangeRequest ( messagep, vlinkp )
s*
s*	SCMP handler for a received CHANGE-REQUEST message.
s*
 */

  static enum ST2Errors			/* Next bugid 0x52600 */
HandleChangeRequest( messagep, vlinkp ) /* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps */
  struct aVLink		*vlinkp;	/* Exists ARGSUSED */
{
#ifdef NOTYET
|		struct aVLink *upstream_link;
|	
|	/*?? need to check if we are the origin, if we are, notify
|	     application actually, at origin the CtrlMsgRelay call
|	     will pass the ChangeRequest message up to the api.
|	  ?? need to handle appropriate transformations
|	     in api and check api compliance???
|	
|		(void) DoAcknowledge ( messagep, vlinkp, NoError );
|		upstream_link =
|			VLinkByVLId ( vlinkp->cp->upstream_vlid );
|		/* ??? reason = */ (void) CtrlMsgRelay ( messagep, upstream_link );
|	 */
#endif NOTYET
	return ( NoSupport );		/* To send an ErrorInRxxx */
}

/*
S* HandleConnect ( messagep, vlinkp )
s*
s*	SCMP handler for a received CONNECT message.
s*
 */

  static enum ST2Errors			/* Next bugid 0x52712 */
HandleConnect( messagep, vlinkp )	/* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps */
  struct aVLink		*vlinkp;	/* May be NULL when no RVLId */
{
    enum OpCodes	 opcode;
    enum ST2Errors	 error;
    HID			 hop_id;
    int			 tn;
    struct aRouteChoiceList arcl;
    struct aST2pcb	*cp;
    struct aFlowSpec3	*fsp;
    struct sockaddr_in	*sip;


#ifndef NOTYET
    if ( messagep->RFlowSpecp NE (struct aFlowSpec *) NULL )
    {
	/* ??? not yet support FullDuplex */
	BUGRETURN (,NoSupport,0x52701,enum ST2Errors);
    }
#endif NOTYET

    /* ==================================================================== */
    /* If not have a Virtual Link, create a new Stream, ST2pcb, vlink/prvhop */

    if ( vlinkp EQ NO_VLINKP )
    {
	/* Keep local host address of interface where message arrived. */
	if ( messagep->gifcp EQ NO_NETIFP )
	    BUGRETURN (-,InvalidPointer,0x52702,enum ST2Errors);
					/* To send an ErrorInRxxx */

	/* ---------------------------------------------------------------- */
	/* Find stream table entry by Name, or if none, create a new entry. */
	/* ??? not for LocalOrigin/cp->gifcp */
	/* ??? rank: gwy vs listening/accepted connections */
	/* ??? rework for distributed stream name */

	if ( (cp = st2_PCBGet()) EQ NO_PCBP )
	{
	    BUGRETURN (-,CantGetResrc,0x52703,enum ST2Errors);
	}				/* To send an ErrorInRxxx */

	/* ---------------------------------------------------------------- */
	/* Record steam information */

	Bcopy (messagep->Originp,&( cp->Origin ),sizeof (cp->Origin));

	Bcopy (messagep->Originp,&( cp->soname.applent ),
						sizeof (cp->soname.applent));
	cp->soname.applent.pcode = (octet1) STRmtAppEnt;
	cp->soname.sa_family = ConfigParm (coip_family);

	Bcopy (messagep->Namep,&( cp->Name ),sizeofaName);

	cp->optapi = messagep->hoscmphdr.Options & (PBit | SBit | TSRBits);

	/* ??? ought to be earlier, & compared with min acceptable level */
	fsp = (struct aFlowSpec3 *) messagep->FlowSpecp;
	if ( fsp->FlowVer EQ STFSVer3 )
	    cp->precedence = fsp->Precedence;

	cp->detectorIPAdr = ntohIP (*messagep->detectoripadrp);

	(void) FlowSpecFromParm( messagep->FlowSpecp,
				       &( cp->flow_spec ) );
	/* ??? boolean return */

	if ( messagep->RFlowSpecp )	/* ??? reverse cp */
	{
	    (void) FlowSpecFromParm( messagep->RFlowSpecp,
					   &( cp->rflow_spec ) );
	    /* ??? boolean return */
	}

	/* ---------------------------------------------------------------- */
	/* Setup upstream virtual link, first the aNxtHop */

	cp->prvhop.gifcp = messagep->gifcp;
	cp->prvhop.lclhdrlen = messagep->gifcp->ifc_lclhdrlen;
	cp->prvhop.fwdtoIPAdr = messagep->hoscmphdr.SndrIPAdr;

	sip = (struct sockaddr_in *) &( cp->prvhop.lclsockaddr );
	sip->sin_family = ConfigParm (coip_family);
	sip->sin_addr.s_addr = messagep->hoscmphdr.SndrIPAdr;

	/* Record any encapsulation info for upstream link */
	if ( messagep->encaplen > 0 )
	{
	    error = IPEncapRev( GATEWAY_VLINK, messagep->encaphdrp,
			        &( cp->prvhop ) );
	    if ( error NE NoError )
	    {
		st2_PCBDel( cp );	/* WARNING: cp is no longer valid */
		BUGRETURN (-,error,0x52704,enum ST2Errors);
	    }
	}

	Bzero (&( arcl ),sizeof (struct aRouteChoiceList));
	st2_FindRoutes( sip->sin_addr.s_addr, messagep->hoscmphdr.SndrIPAdr,
		        &( arcl ), FRF_DIRECT | FRF_NOSKT |
		        (cp->prvhop.encaplen > 0) ? FRF_USEIP : 0 );
	if ( (arcl.nxtfree EQ 0)  OR  (arcl.RouteChoices[0].error NE NoError) )
	{   /* ??? needs work when route table has srcruts */
	    st2_PCBDel( cp );		/* WARNING: cp is no longer valid */
	    BUGRETURN (-,arcl.RouteChoices[0].error,0x52705,enum ST2Errors);
	}
	sip->sin_addr.s_addr = arcl.RouteChoices[0].fwdtoIPAdr;
	cp->prvhop.fwdtoIPAdr = arcl.RouteChoices[0].fwdtoIPAdr;

	/* ................................................................ */
	/* Create upstream control link */

	vlinkp = VLinkNew( cp, UPSTREAM_VLINK, GATEWAY_VLINK );
	if ( vlinkp EQ NO_VLINKP )
	{
	    st2_PCBDel( cp );		/* WARNING: cp is no longer valid */
	    BUGRETURN (-,CantGetResrc,0x52706,enum ST2Errors);
	}				/* To send an ErrorInRxxx */

	cp->upstream_vlinkp = vlinkp;

	vlinkp->lclIPAdr = st2_LocalIPAdrOfGifc( messagep->gifcp );

	/* Use previous hop info from connection table. */
	vlinkp->route.nxthopp = &( cp->prvhop );

	/* ??? implications of new incarnation */
	(void) VLinkSetRemote( vlinkp, messagep->hoscmphdr.SVLId,
			       messagep->hoscmphdr.SndrIPAdr );
	/* ??? boolean error */
	/* sets vlinkp->rmtIPAdr */

	/* ??? not correct test w/ IP encapsulation */
	vlinkp->route.type = (messagep->hoscmphdr.SndrIPAdr
			      EQ messagep->hoscmphdr.peropcode4)
			      ? TO_HOST : TO_GATEWAY;

	/* ---------------------------------------------------------------- */
	/* add upstream neighbor, just in case */

	vlinkp->np = NbrCreate( vlinkp->rmtIPAdr, FRF_NOSKT,
			        messagep, NO_TRGP );

	cp->dspin = st2_Forward;
	cp->flags |= ConnActive;

	SetState (PCB,cp,pcb_state,CINS_OPEN);

    } /* end of new virtual link */

    /* -------------------------------------------------------------------- */
    /* Have a vlinkp */

    /*
     * Make sure stream parameters in the connect are compatible with
     * previously supplied stream parameters.
     */
    cp = vlinkp->cp;
    if ( cp EQ NO_PCBP )
    {
	BUGSTOP (-,vlinkp,0x52707);
	BUGSTOP (,vlinkp->np,0x52708);
	return ( InconsistVLID );
    }

    if (   cp->optapi
	NE (messagep->hoscmphdr.Options & (PBit | SBit | TSRBits)) )
    {
	BUGRETURN (-,ProtocolError,0x52709,enum ST2Errors);
    }					/* To send an ErrorInRxxx */

    /* ==================================================================== */
    /* Initiate HID Negotiation */

    /*
     * Check for a hop identifier.  If one is given in the connect message,
     * we have to verify it and potentially reject it.  Either way, we will
     * continue with connect processing because hid negotiation proceeds in
     * parallel.
     */
    if ( messagep->hdrhidp NE (octet2 *) NULL )
    {
	hop_id = (HID) (*messagep->hdrhidp);
	opcode = OpHidApprove;

	IfLog (DBG_HID_NEG)
	{
	    (void) printf ( "negotiate_hid (0x%x, hid %d, ref %d, vlinkp 0x%x)\n",
		   cp, hop_id, (Reference) messagep->hoscmphdr.Ref, vlinkp );
	}

	if ( hop_id EQ (HID) 0 )	/* We (next-hop) assign the hid */
	{
	    if ( cp->hid_ready )
		hop_id = cp->hop_id[0]; /* Current HID */
	    else
		hop_id = HIDNew( cp ); /* ??? 0 on error */
	}

	/* Previous hop is first to propose a HID */
	else if ( NOT cp->hid_ready )
	{
	    /* Will fail if proposed HID is in use by another */
	    if ( NOT HIDSet( cp, hop_id ) )
	    {
		IfLog (DBG_HID_NEG)
		    (void) printf ( "negotiate_hid -- HIDSet failed\n" );
		opcode = OpHidReject;
	    }
	}

	/* Previous-hop is proposing, but we've already agreed to a HID */
	else
	{
	    if ( cp NE Forwarder_for_HID (hop_id) )
	    {
		IfLog (DBG_HID_NEG)
		    (void) printf ( "negotiate_hid -- bound_hid_Forwarder\n" );
		opcode = OpHidReject;
	    }
	}

	IfLog (DBG_HID_NEG)
	{
	    if ( opcode EQ OpHidApprove )
	    {
		/* Approve the HID we finally chose for the forwarder. */
		(void) printf ( "negotiate_hid approving hid %d (nbo %d)\n",
			        hop_id, hton2 (hop_id) );
	    (void) printf ( "hid_approve_Message\n" );
	    }
	    else
	    {
		(void) printf ( "hid_reject_Message\n" );
	    }
	}

	(void) DoHidReply( messagep, vlinkp, opcode, hop_id );

	/* ??? need to worry about out of order delivery */
	if ( opcode EQ OpHidApprove )
	{
	    /* Relay any accept's held until HID is set. */
	    relay_accepts(cp);
	}
    }
    else
	(void) DoAcknowledge( messagep, vlinkp, NoError );

    /* ==================================================================== */
    /* Add new targets to target table.  Mark old targets for reconnect. */

    for ( tn = 0; tn < messagep->targlst.nxtfree; tn++ )
    {
	struct aTarget *aTargetp = messagep->targlst.TargetDescs[ tn ].Targetp;

	if ( aTargetp NE (struct aTarget *) NULL )
	    (void) TargAdd( cp, aTargetp,
			       (Reference) messagep->hoscmphdr.Ref );

	/* ??? error CantGetResrc */
	/* ??? mark for batched Refuse */
	/* ??? where to send Refuse */
    }

    /* ==================================================================== */
    /* Cache User data away, assuming at least one target */

    if (    (messagep->UserDatap NE (struct aUserData *) NULL)
	AND (cp->fwdrfreespace NE NO_PKTDSCP) )
    {
	if ( cp->userdatap NE (struct aUserData *) NULL )
	{
	    BUGSTOP (-,DupParm,0x52710); /* ??? be more forgiving */
	}
	else
	{
	    char *udp = AddSpace (char *, &( cp->fwdrfreespace ), 
				  PKT_MD_PROVHD,
				  (int) messagep->UserDatap->plen);

	    if ( udp NE (char *) NULL )
	    {
		Bcopy (messagep->UserDatap,udp,messagep->UserDatap->plen);
		cp->userdatap = Mkp (struct aUserData *, udp, 0);
	    }
	    else
	    {
		BUGSTOP (,UserDataSize,0x52711);
		/* ??? return an error message */
	    }
	}
    }

    /* ==================================================================== */

    if ( cp->needupdate NE 0 )
	st2_PCBUpdate( cp );		/* Run state machines */
	/* cp may be invalid */

    return ( NoError );
}

/*
S* HandleDisconnect ( messagep, vlinkp )
s*
s*	SCMP handler for a received DISCONNECT message.
s*
 */

  static enum ST2Errors			/* Next bugid 0x52803 */
HandleDisconnect( messagep, vlinkp )	/* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps */
  struct aVLink		*vlinkp;	/* Exists */
{
    enum ST2Errors	 reason = (enum ST2Errors) *messagep->reasoncodep;
    int			 targidx;
    Reference		 reference = (Reference) messagep->hoscmphdr.Ref;
    struct aST2pcb	*cp;
    struct aTarget	*atp;
    struct Target	*targetp;
    struct aUserData	*userdatap = (struct aUserData *) NULL;


    if (   (vlinkp EQ NO_VLINKP)
	OR ((cp = vlinkp->cp) EQ NO_PCBP) )
    {
	BUGRETURN (-,NameUnknown/*InvalidPointer*/,0x52801,enum ST2Errors);
    }					/* To send an ErrorInRxxx */

    (void) DoAcknowledge( messagep, vlinkp, NoError );

    if ( reason EQ NoError )
    {
	BUGSTOP (,reason,0x52802);	/* Should have an error */
	reason = ErrorUnknown;		/* ??? ApplDisconnect */
    }

    /* Cache User data away, assuming at least one target */

    if (    (messagep->UserDatap NE (struct aUserData *) NULL)
	AND (cp->fwdrfreespace NE NO_PKTDSCP) )
    {
	userdatap = AddSpace (struct aUserData *, &( cp->fwdrfreespace ),
			      PKT_MD_PROVHD, (int) messagep->UserDatap->plen);

	if ( userdatap NE (struct aUserData *) NULL )
	{
	    Bcopy (messagep->UserDatap,userdatap,messagep->UserDatap->plen);
	}
	/* else just ignore it */
    }


    /* G-Bit says disconnect all targets. */
    if ( (messagep->hoscmphdr.Options & GBit) NE 0 )
    {
	/* ??? this doesn't propagate G-Bit */
	disconnect_Forwarder(cp, reason,userdatap);
    }
    else /*??? Do we tear down the whole thing if there are no targets? */
    {
	for ( targidx = 0; targidx < messagep->targlst.nxtfree; targidx++ )
	{
	    atp = messagep->targlst.TargetDescs[targidx].Targetp;

	    targetp = TargByParm( cp, UPSTREAM_VLINK, atp );
	    if ( targetp NE NO_TRGP )
	    {
		(void) TargDisc( targetp, reference, reason,
					  userdatap );
		/* ??? error */
	    }
	}
    }

    cp->needupdate++;
    st2_PCBUpdate( cp );
    /* cp may be invalid */

    return ( NoError );
}

/*
S* HandleErrorInRequest ( messagep, vlinkp )
s*
s*	SCMP handler for a received ERROR-IN-REQUEST message.
s*	Followup on return.
s*
 */

  static enum ST2Errors			/* Next bugid 0x52903 */
HandleErrorInRequest( messagep, vlinkp ) /* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps */
  struct aVLink		*vlinkp;	/* May be NULL */
{
    if ( vlinkp EQ NO_VLINKP )
    {
	BUGSTOP (-,(messagep->scmhdrp->Ref << 16) | (* messagep->reasoncodep),0x52901);
#if 0
	BUGRETURN (-,ProtocolError,0x52902,enum ST2Errors);
	/* ??? really error */
#endif
	return ( NoError );
    }


    return ( (enum ST2Errors) *messagep->reasoncodep );
    /* to notify_pending_Forwarder */
}

/*
S* HandleErrorInResponse ( messagep, vlinkp )
s*
s*	SCMP handler for a received ERROR-IN-RESPONSE message.
s*	Followup on return.
s*
 */

  static enum ST2Errors			/* Next bugid 0x53003 */
HandleErrorInResponse( messagep, vlinkp ) /* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps */
  struct aVLink		*vlinkp;	/* May be NULL */
{
    if ( vlinkp EQ NO_VLINKP )
    {
	BUGSTOP (-,(messagep->scmhdrp->Ref << 16) | (* messagep->reasoncodep),0x53001);
#if 0
	BUGRETURN (-,ProtocolError,0x53002,enum ST2Errors);
#endif
	return ( NoError );
    }

    return ( (enum ST2Errors) *messagep->reasoncodep );
    /* to notify_pending_Forwarder */
}

/*
S* HandleHello ( messagep, vlinkp )
s*
s*	SCMP handler for a received HELLO message.
s*
 */

  static enum ST2Errors			/* Next bugid 0x53104 */
HandleHello( messagep, vlinkp )		/* from st2_CMPInput */
					/* via ops[] */
  struct aPrsSCM	*messagep;	/* Caller keeps */
  struct aVLink		*vlinkp;	/* Should be NO_VLINKP */
{
    long		 lcl_ms,
			 nbr_ms;
    struct aNeighbor	*np;
    struct timeval	*tvp = &( messagep->timestamp );


    /* Make sure we know about this neighbor */
    if (   (messagep->hoscmphdr.RVLId EQ VLID_NONE) /* nbr doen't know it */
					/* nbr has stale VLId for us */
	OR ((vlinkp = VLinkByVLId(messagep->hoscmphdr.RVLId)) EQ NO_VLINKP)
	OR (vlinkp->type NE NEIGHBOR_VLINK)
	OR (vlinkp->rmtIPAdr NE messagep->hoscmphdr.SndrIPAdr)
	/* ??? check encap len too? canocicalize first */
					/* we failed to create nbr last time */
	OR ((np = vlinkp->np) EQ NO_NBRP) )
    {					/* don't have a vlink to this nbr */
	np = NbrCreate( messagep->hoscmphdr.SndrIPAdr, FRF_NOSKT,
		        messagep, NO_TRGP );
	/* NbrCreate set np->vlinkp if successful */

	if (   (np EQ NO_NBRP)
	    OR ((vlinkp = np->vlinkp) EQ NO_VLINKP) )
	{
	    BUGSTOP (-,CantGetResrc,0x53101);
	    return ( NoError );
	}
    }
    else if ( messagep->hoscmphdr.SVLId NE vlinkp->remote_vlid )
    {
	/* Update remote VLId */

	BUGSTOP (,  (vlinkp->remote_vlid << 16)
		  | messagep->hoscmphdr.SVLId,0x53103);
	vlinkp->remote_vlid = messagep->hoscmphdr.SVLId;
    }


    /* Acknowledge message if requested to do so */
    if ( messagep->hoscmphdr.Ref NE 0 )
	(void) DoAcknowledge( messagep, vlinkp, NoError );

    /* Receipt of a HELLO proves that the neighbor implements HELLOs */
    np->totheard++;
    Count (hellorecv,1);


    /* If neighbor restarted, record & notify streams. */
    if ( (messagep->hoscmphdr.Options & RBit) NE 0 )
    {
	if ( (np->flags & ~NbrFlgRestarted) NE 0 ) /* if flag previously zero */
	{
	    BUGSTOP (,np->rmtIPAdr,0x53102); /* Neighbor restarting */
	    NbrFailed( np, RemoteRestart );
	}
	np->flags |=   NbrFlgRestarted;	 /* set flag */
    }
    else
	np->flags &= ~ NbrFlgRestarted;	 /* clear flag */


    /* How long has it been since last HELLO and do we even care? */
    if ( (np->flags & NbrFlgActive) NE 0 )
    {					/* yes, we care */
	if (    (np->hellorcvd.tv_sec NE 0) /* becomes non-0 when NbrFlgActive
					       becomes true, not when heard
					       first HELLO */
	    AND (np->hellotimer_ms NE 0) )
	{				/* and this isn't the first HELLO */
	    /* calculcate the time deltas in milliseconds */

	    lcl_ms = MsecBetween (tvp,&(np->hellorcvd)); /* ours */
	    nbr_ms = *messagep->hellotimerp - np->hellotimer_ms; /* remote's */

	    IfLog (DBG_NBR_RTT)
		printf ( "lcl_ms = %u, nbr_ms = %u", lcl_ms, nbr_ms);

	    /* it's important to assign the time value to variables in order
	       to prevent the compiler from preserving the high-bit carry */

	    nbr_ms -= lcl_ms;		/* difference here versus difference there */
	    if ( nbr_ms < 0 )   nbr_ms = -nbr_ms;

	    IfLog (DBG_NBR_RTT)
		printf ( " difference = %u, var = %u\n",
			 nbr_ms, (np->rtt_var_ms << 1) );

	    if ( nbr_ms > (np->rtt_var_ms << 1) )  /* 2 * rtt_var_ms */
		return ( NoError );	/* ignore message */

	    /* ??? use nbr_ms to calculate variance */
	}

	/* Update time last heard to now */
	np->hellorcvd = *tvp;
	np->hellotimer_ms = *messagep->hellotimerp;
    }

    return ( NoError );
}

/*
S* HandleHidApprove ( messagep, vlinkp )
s*
s*	SCMP handler for a received HID-APPROVE message.
s*	Followup on return.
s*
 *  This is called by a previous-hop to process a hid accept control
 *  message received from the next-hop.  If the accepted HID matches
 *  the one we were trying to negotiate, then send an ack back to
 *  the next-hop and mark that our negotiation is complete.  If
 *  they don't match, send an appropriate error-in-reply
 *  (error-in-request? check the specification) on to the next-hop
 *  and follow up with a HID-CHANGE request to the next-hop to try
 *  to settle on our desired HID.
 */

  static enum ST2Errors			/* Next bugid 0x53203 */
HandleHidApprove( messagep, vlinkp )	/* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps */
  struct aVLink		*vlinkp;	/* Exists */
{
    enum ST2Errors	 reason = NoError;
    HID			 hop_id;
    struct aNxtHop	*nxthopp;
    struct NextHop	*next_hopp;


    if ( vlinkp EQ NO_VLINKP )
    {
	BUGSTOP (-,messagep,0x53201);
	return ( NoError );
    }

    /*
     *? We need to add some stuff to distinguish approve of change/add/delete
     *?  Also, response to an add may have a pending delete in the drop table
     *?  so we'll want to follow up on that.
     */
   
    if ( (next_hopp = vlinkp->next_hopp) EQ NO_NEXTHOPP )
    {					/* Must be stale HandleHidApprove */
	BUGRETURN (,InvalidRVLId,0x53202,enum ST2Errors);
    }

    hop_id = *messagep->hdrhidp;
					/* ??? DIRECT_HOP vs MULTICAST_HOP */
    if ( next_hopp->hid_info.hop_id EQ (HID) 0 )
    {
	nxthopp = next_hopp->route.nxthopp;
	next_hopp->hid_info.hid_set = TRUE;
	next_hopp->hid_info.hop_id = hop_id;
	if ( nxthopp NE NO_NXTHOPP )
	{
	    activate_aNxtHop(nxthopp, hop_id);
	}

    }
    else if ( hop_id EQ next_hopp->hid_info.hop_id )
    {
	/* HID agrees with our choice */
	next_hopp->hid_info.hid_set = TRUE;
    }
    else
    {
	/* There's a conflict between approved hid and our choice of hid. */

	reason = ProtocolError /*HID approved does not match proposed HID*/;

	/* ??? multiple outstanding changes due to multicast negotiation */

	(void) DoError( messagep, vlinkp, reason );

	/* ??? if DoError, do we really want to notify forwarder */

	(void) HIDChange( CHANGE_HID, next_hopp->hid_info.hop_id, vlinkp );

    }

    return ( reason );			/* to notify_pending_Forwarder */
}

/*
S* HandleHidChange ( messagep, vlinkp )
s*
s*	NOTYET SCMP handler for a received HID-CHANGE message.
s*
 */

  static enum ST2Errors			/* Next bugid 0x53302 */
HandleHidChange( messagep, vlinkp )	/* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps */
  struct aVLink		*vlinkp;	/* Exists ARGSUSED */
{

#ifdef NOTYET
|	#define MessageOption (message,type) 0
|	
|	   struct NextHop *next_hopp = vlinkp->next_hopp;
|	   HID new_hid = MessageOption ( messagep, MESSAGE_HID );
|	   
|	   switch ( messagep->options )
|	   {
|	   case HC_OPTION_CHANGE:
|	      if ( hid_check ( vlinkp, messagep->hoscmphdr.Reference,
|			      new_hid, next_hopp ) )
|	      {
|		 hid_approve ( vlinkp, messagep->hoscmphdr.Reference,
|			      new_hid, next_hopp );
|		 next_hopp->hid_info.hid_set = TRUE;
|	      }
|	      return;
|		  
|	    case HC_OPTION_ADD:
|	      if (    (next_hopp->hid_info.hop_id NE (HID) 0)
|		  AND (next_hopp->alt_hop_id NE (HID) 0) )
|	      {
|		 (void) DoError ( messagep, vlinkp, ErrorUnknown );
|			 /*"Too many add hid requests.  Must delete some."*/
|	      }
|	      else if ( hid_check ( vlinkp, messagep->hoscmphdr.Reference,
|				  new_hid, next_hopp ) )
|	      {
|		 if ( next_hopp->hid_info.hop_id EQ (HID) 0 )
|		   next_hopp->hid_info.hop_id = new_hid;
|		 else
|		   next_hopp->alt_hop_id = new_hid;
|		 hid_approve ( vlinkp, messagep->hoscmphdr.Reference,
|			      new_hid, next_hopp );
|	      }
|	      return;
|	
|	    case HC_OPTION_DELETE:
|	      if ( new_hid EQ next_hopp->alt_hop_id )
|	      {
|		 next_hopp.alt_hop_id = (HID) 0;
|	      }
|	      else if ( new_hid EQ next_hopp->hid_info.hop_id )
|	      {
|		 if ( next_hopp->alt_hop_id NE (HID) 0 )
|		 {
|		    next_hopp->hid_info.hop_id = next_hopp->alt_hop_id;
|		    next_hopp->alt_hop_id = (HID) 0;
|		 }
|		 else
|		 {
|		    (void) DoError ( messagep, vlinkp, );
|			 /*"Can't delete last remaining hop identifier."*/
|		 }
|	      }
|	      else
|		(void) DoError ( messagep, vlinkp,
|			ErrorUnknown /*"Unknown HID passed to delete."*/ );
|	      return;
|	    default:
|	      (void) DoError ( messagep, vlinkp,
|		        ErrorUnknown /*"Unknown HIDCHANGE subcommand."*/ );
|	      return;
|	   }
#endif

    BUGRETURN (-,NoSupport,0x53301,enum ST2Errors);
}					/* To send an ErrorInRxxx */

/*
S* HandleHidChangeRequest ( messagep, vlinkp )
s*
s*	NOTYET SCMP handler for a received HID-CHANGE-REQUEST message.
s*
 *  The next hop has asked us to change the HID we use to communicate
 *  with it.  The original HID will remain in force unless we send
 *  a HID-CHANGE request.  However, we should ACK or NACK the
 *  HIDChangeRequest to squelch retransmission.
 */

  static enum ST2Errors			/* Next bugid 0x53402 */
HandleHidChangeRequest( messagep, vlinkp ) /* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps */
  struct aVLink		*vlinkp;	/* Exists ARGSUSED */
{

#ifdef NOTYET
|	   struct NextHop *next_hopp = vlinkp->next_hopp;
|	   HID hop_id = messagep->hid_value;
|	   
|	   /* If we haven't agreed to a HID yet, then go with what the
|	    *  next hop would like us to use. */
|	   if ( NOT next_hopp->hid_info.hid_set )
|	   {
|	      next_hopp->hid_info.hop_id = hop_id;
|	      (void) HIDChange ( ST_HID_CHANGE, next_hopp->hid_info.hop_id,
|				  vlinkp );
|	   }
|	   
|	   /* If we have agreed to the HID, but we haven't opened the
|	    *  data pathway to some other next hop, we can still process
|	    *  the change request. */
|	   else if ( vlinkp->cp->st2_fwdr_state NE ALL_READY )
|	   {
|	      next_hopp->hid_info.hid_set = FALSE;
|	      next_hopp->hid_info.hop_id = hop_id;
|	      (void) HIDChange ( ST_HID_CHANGE, next_hopp->hid_info.hop_id,
|				  vlinkp );
|	   }
|	   else /*we agreed to a hid and data may be flowing so we add/delete */

|		/* Look for an open slot in the drop table */
|		slot = -1;
|		for (i = 0; i < DimensionOf (next_hop.modify) AND slot < 0; i++)
|		{
|		    if ( next_hop.modify[ i ].add_hid EQ (HID) 0 )
|		    slot = i;
|		}
|		if ( slot >= 0 )
|		{
|		  next_hop.drop[slot].add_hid = add_hid;
|		  next_hop.drop[slot].drop_hid = drop_hid;
|		  (void) HIDChange ( ADD_HID, add_hid, link );
|		  /*?? Need to follow up on this in HIDApprove/Reject code */
|		}
|	   
|		/* We didn't find an empty slot, so reject the request */
|		else
|		  (void) DoError ( messagep, vlinkp, ErrorUnknown);
|				   /*"Too many add hid requests required"*/
|	      }
#endif NOTYET

    BUGRETURN (-,NoSupport,0x53401,enum ST2Errors);
}					/* To send an ErrorInRxxx */

/*
S* HandleHidReject ( messagep, vlinkp )
s*
s*	SCMP handler for a received HID-REJECT message.
s*	Followup on return.
s*
 */

  static enum ST2Errors			/* Next bugid 0x53503 */
HandleHidReject( messagep, vlinkp )	/* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps */
  struct aVLink		*vlinkp;	/* Exists */
{
    struct NextHop *next_hopp;
    enum ST2Errors reason = NoError;


    if ( vlinkp EQ NO_VLINKP )
    {
	BUGSTOP (-,messagep,0x53501);
	return ( NoError );
    }

    next_hopp = vlinkp->next_hopp;
					/* ??? vs EQ MULTICAST_HOP */
    if ( next_hopp->hid_info.type NE DIRECT_HOP )
    {
					/* MulticastHIDReject */
	BUGRETURN (-,NoSupport,0x53502,enum ST2Errors);
#ifdef NOTYET
|		find stream entry for messagep$StreamName
|		  find entry for this next hop
|			if not multicast
|			  then use returned suggestion HID
|				keep it in the entry for next hop
|			 send HIDCHANGE[newtid, HID] to next hop
|			 end
|		else if second rejection from next hop
|		    then send CONNECT[targetp] without multicast option
|				adding this targetp set
|		else
|		   build HID Guess Mask based on returned
|			suggestions and masks--combine masks
|			for highest suggested HIDs and pick
|			first one HID available to all
|		   while all have replied, pick HID acceptable
|			to most of the targets and do
|			send HIDCHANGE[newtid, HIDguess] to next hop
|		end
#endif NOTYET

    }
    else if ( next_hopp->hid_info.hid_set )
    {
	/*"HIDReject received after negotiation complete"*/
	reason = ProtocolError;
	(void) DoError( messagep, vlinkp, reason );
	/* ??? if DoError, do we really want to notify fwdr */
	/* ??? flush HID and start over? */
    }
    else
    {
	HID basehid;
	octet4 *maskp;
	int count;

	/*??? If FreeHIDsp is not given, Refuse and Disconnect. */

	/* ??? use table */
	kassert (messagep->FreeHIDsp NE (struct aFreeHIDs *) NULL);
	basehid = ntoh2 (messagep->FreeHIDsp->basehid) & ~0x1f;
	maskp = &( messagep->FreeHIDsp->freehidbitmask[ 0 ] );
	count = (ntoh1 (messagep->FreeHIDsp->plen) - 4) / 4;
	next_hopp->hid_info.hop_id = (HID) 0;

	while ( --count >= 0 )
	{
	    unsigned int bitnum = 0;
	    unsigned long hidmask = ntoh4 (*maskp);
	    while ( hidmask NE 0 )
	    {
		if ( (hidmask & (1<<31)) NE 0 )
		{
		    next_hopp->hid_info.hop_id = basehid + bitnum;
		    hidmask = 0;
		    count = -1;
		}
		bitnum++;
		hidmask <<= 1;
	    } /* end of while hidmask */

	    basehid += 32;
	    maskp++;
	} /* end of while count */

	if ( next_hopp->hid_info.hop_id EQ (HID) 0 )
	{
	    reason = HIDNegFails;
	    (void) DoError( messagep, vlinkp, reason );
	    /* ??? if DoError, do we really want to notify fwdr */
	}
	else
	    (void) HIDChange( CHANGE_HID, next_hopp->hid_info.hop_id,
			       vlinkp );
    }

    return ( reason );			/* to notify_pending_Forwarder */
}

/*
S* HandleNotify ( messagep, vlinkp )
s*
s*	NOTYET SCMP handler for a received NOTIFY message.
s*
 */

  static enum ST2Errors			/* Next bugid 0x53600 */
HandleNotify( messagep, vlinkp )	/* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps */
  struct aVLink		*vlinkp;	/* Exists ARGSUSED */
{

  /* ??? deliver to appl or pass it on */

    return ( NoError );
}

/*
S* HandleRefuse ( messagep, vlinkp )
s*
s*	SCMP handler for a received REJECT message.
s*	a) response to connect		has LnkRef
s*	b) application "disconnect"	check Reason
s*	c) need repair			check Reason
s*
 */

  static enum ST2Errors			/* Next bugid 0x53703 */
HandleRefuse( messagep, vlinkp )	/* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps */
  struct aVLink		*vlinkp;	/* Exists */
{
    enum ST2Errors	 reason = (enum ST2Errors)* messagep->reasoncodep;
    int			 targidx;
    struct aST2pcb	*cp = vlinkp->cp;
    struct aTarget	*aTargetp;
    struct Target	*targetp;


    if ( vlinkp EQ NO_VLINKP )
    {
	BUGSTOP (-,messagep,0x53701);
	return ( NoError );
    }

    (void) DoAcknowledge( messagep, vlinkp, NoError );

    if ( cp EQ NO_PCBP )
    {
	BUGSTOP (-,InvalidPointer,0x53702);
	return ( NoError );		/* Don't send ErrorInRxxx */
    }

    /* Update state of REFUSEd target (s) */
    for ( targidx = 0; targidx < messagep->targlst.nxtfree; targidx++ )
    {
	aTargetp = messagep->targlst.TargetDescs[targidx].Targetp;

	targetp = TargByParm( cp, DOWNSTREAM_VLINK, aTargetp );
	if ( targetp EQ NO_TRGP )
	    continue;			/* Inconsistency */

	/* ??? ought to drop anything on accepts queue for same target
	   if don't find target, drop from target list on refuse */

	targetp->pending_reference = (Reference) 0;
					/* ??? obsolete */
	switch ( targetp->targstate[ (int) Nexthop ] )
	{	
	    /* These states indicate we have not opened a connection
	       to the next hop, so we need only tear down our data
	       structures. */
	case TS_CLOSED:
	    break;

	    /* We really don't expect refuse while in these states
		   since they imply no connect has been sent. */

	case TS_WANTOPEN:
	case TS_WANTSUSPENDED:
	case TS_NEW:
#ifdef  STIIDEBUG
	    (void) printf ( "Refuse received for unconnected target?\n" );
#endif  STIIDEBUG
	    SetTargState (targetp,TS_CLOSED);
	    /* Fall through */

	    /* These states indicate we think there might be a
	       connection to the next hop, but refuse means it
	       has been torn down so we move to the closed state. */
	case TS_OPEN:
	case TS_OPENING:
	case TS_SUSPENDED:
	case TS_ROUTING:
	case TS_CLOSING:
	default:
#ifdef NOTYET
|			if ( CanRepair ( *messagep->reasoncodep ) )
|			    SetTargState (targetp,TS_SUSPENDED);
|			else if ( CanReroute ( *messagep->reasoncodep ) )
|			    SetTargState (targetp,TS_ROUTING);
|			else
#endif NOTYET
	    SetTargState (targetp,TS_CLOSED);
	    cp->needupdate++;
	    break;

	} /* end of targetp->targstate switch */


	/* Decide whether to reroute or pass th buck, based on reason. */
					/* ??? NoRecovery */
	if ( NOT IsTargStateIn (targetp,TS_DISCONNECTED,
			       TS (TS_REFUSING) TS (TS_FAILED)) )
	{
	    targetp->error = reason;
	    SetTargState (targetp,TS_FAILED);
	    cp->needupdate++;
	}

    } /* end of for all targets in REFUSE loop */


    /* Run the state machine */
    if ( cp->needupdate > 0 )
    {
	st2_PCBUpdate( cp );
    }


    return ( NoError );			/* Don't send ErrorInRxxx */
}

/*
S* HandleStatus ( messagep, vlinkp )
s*
s*	NOTYET SCMP handler for a received STATUS message.
s*
 */

  static enum ST2Errors			/* Next bugid 0x53804 */
HandleStatus( messagep, vlinkp )	/* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps */
  struct aVLink		*vlinkp;	/* May be NULL */
{
    struct aST2pcb	*cp = NO_PCBP;
    struct aPrsSCM	*rspMsgp;


    if ( (messagep->hoscmphdr.Options & HBit) NE 0 )
    {
	messagep->hdrhidp = &messagep->hoscmphdr.peropcode2;

	cp = Forwarder_for_HID ((HID) (*messagep->hdrhidp));

	if (    (vlinkp NE NO_VLINKP)
	    AND (vlinkp->cp NE cp) )
	{
	    BUGRETURN (-,ProtocolError,0x53801,enum ST2Errors);
	}				/* To send an ErrorInRxxx */
    }
    else if ( vlinkp NE NO_VLINKP )
	cp = vlinkp->cp;

    if ( messagep->Namep NE (struct aName *) NULL )
    {
	struct aST2pcb *wcp;

	wcp = PCBByName( messagep->Namep );
					/* ??? ring of wcp's */
	if ( cp NE NO_PCBP )
	{
	    if ( cp NE wcp )
	    {
		BUGRETURN (-,ProtocolError,0x53802,enum ST2Errors);
	    }				/* To send an ErrorInRxxx */
	}

	/*
	 * If multiple forwarders link to stream, it's ambiguious
	 * and we can't know which one the caller wants.  We need
	 * a vlid or hid to resolve the ambiguity.
	 */
	else if ( wcp->nxtNamep NE NO_PCBP )
	{
	    return ( StreamExists );
	}
	else
	{
	    kassert (wcp NE NO_PCBP);
	    cp = wcp;
	}
    }

    /* ==================================================================== */
    /* Send Status Response Message */

    rspMsgp = CtrlMsgCreate();
    if ( rspMsgp EQ NO_PSCMP )
    {
#ifdef  STIIDEBUG
	(void) printf ( "HandleStatus: out of memory for reply\n" );
#endif  STIIDEBUG
	BUGRETURN (-,CantGetResrc,0x53803,enum ST2Errors);
    }					/* To send an ErrorInRxxx */

    rspMsgp->hoscmphdr.OpCode = (octet1) OpStatusResponse;
    rspMsgp->hoscmphdr.Options = messagep->hoscmphdr.Options & ~ HBit;

#ifdef NOTYET
    if ( cp->hid_ready )
    {
	for ( i = 0 ; i < MAX_HIDS_PER_STREAM ; i++ )
	{
	    /* Stuff them into the message. */
	}
    }
#endif NOTYET

    CtrlMsgDel( rspMsgp );		/* don't loose rspMsg memory */

    return ( NoError );			/* Don't send ErrorInRxxx */
}

/*
S* HandleStatusResponse ( messagep, vlinkp )
s*
s*	NOTYET SCMP handler for a received STATUS-RESPONSE message.
s*
 */

  static enum ST2Errors			/* Next bugid 0x53900 */
HandleStatusResponse( messagep, vlinkp ) /* from st2_CMPInput */
  struct aPrsSCM	*messagep;	/* caller keeps */
  struct aVLink		*vlinkp;	/* May be NULL ARGSUSED */
{
    return ( NoError );			/* Don't send ErrorInRxxx */
}

/*
S* HIDActivate ( cp, hop_id )
s*
s*	Routine to verify that a selected HID is still free, and to claim it
s*	by inserting it into the connection hash table for subsequent use in
s*	ST-II forwarding.
s*
 */

/*  ??? INLINE
    Since hids map injective to hid table slots, we really need check only
    for a busy slot.  The only time when the bit vector would be of use is
    when we want to find a free hid, but the spec says we should choose
    the hid value at random.  Any strategy using the bit vector that improves
    performance over simply testing the slot would lead to correlated hid
    values, and would be bad.  The only other benefit is that the bit
    vector prevents us from using values for which no slot is reserved, but
    that's just the reserved hid values which we could simply test for after
    we confirm a slot is available.
*/

  static boolean			/* Next bugid 0x54000 */
HIDActivate( cp, hop_id )		/* from HIDSet */
  struct aST2pcb	*cp;		/* 		ok */
  HID			 hop_id;	/* 		ok */
{
    struct aConHshEntry *conHashp;
    int			 oldpri;
    boolean		 result;


    kassert (cp NE NO_PCBP);
    kassert (hop_id NE (HID) 0);

    conHashp = &( st2_ConHsh[ ConfigParm (hid_mask) & hop_id ] );
    oldpri = splnet ();

    /* Each HID maps to one slot, so if the slot is open the HID is not
     * in use.  Although, the slot might be in use for a different stream
     * since many HIDs map to each slot.
     */

    if ( conHashp->cp EQ NO_PCBP )
    {
	conHashp->hid = hop_id;
	conHashp->cp = cp;
	result = TRUE;
    }

    /* If the slot is in use, but for the same hid and connection,
     * that's ok.  Not likely ( rx ), but ok.
     */
    else if ( (conHashp->cp EQ cp) AND (conHashp->hid EQ hop_id) )
	result = TRUE;

    /* Otherwise the slot is in use for another connection or hid
     * we we have to reject the request.
     */
    else
	result = FALSE;

    (void) splx ( oldpri );
    return ( result );
}

/*
S* HIDAdd ( cp, hop_id )
s*
s*	NOTYET
s*
 *   Returns: true (non-zero) if hid could not be added;
 *	false (zero) if hid was already in use or too many.
 * We should have it return OK or an error code to indicate
 * whether too many or busy hid causes error.
 */
#ifdef NOTYET
|	
|	  enum ST2Errors
|	_HIDAdd ( cp, hop_id ) /* from <nowhere> */
|	  struct aST2pcb *cp;
|	  HID hop_id;
|	{
|		int lstidx;
|		kassert (hop_id NE (HID) 0);
|		PTRCHECK (cp, struct aST2pcb *, return ( NoError ));
|	
|		for (lstidx = 0; lstidx < DimensionOf (cp->hop_id); lstidx++)
|		{
|			if ( cp->hop_id[ lstidx ] EQ (HID) 0 )
|			{
|				if ( HIDActivate ( cp, hop_id ) )
|				{
|					cp->hid_ready = TRUE;
|					cp->hop_id[lstidx] = hop_id;
|					if ( cp->accepts NE NO_PSCMP )
|						relay_accepts(cp);
|					return ( NoError );
|				}
|				return ( NoConHidVlid );
|			}
|		}
|	
|		return ( TooManyHIDs );
|	}
#endif NOTYET

/*
S* hid_change_Message ( vlinkp, type, hop_id )
s*
 */
#ifdef NOTYET
|	
|	  int
|	_hid_change_Message ( vlinkp, type, hop_id ) /* from <nowhere> */
|	  struct aVLink		*vlinkp;	/* ARGSUSED */
|	  int			 type;
|	  HID			 hop_id;
|	{
|	    return ( -1 );
|	}
#endif NOTYET

/*
S* HIDChange ( subop, hid, vlinkp )
s*
s*	NOTYET Routine to initiate a HID-CHANGE for a next-hop.
s*
 */

  static enum ST2Errors			/* Next bugid 0x54102 */
HIDChange( subop, hid, vlinkp )		/* from HandleHidApprove HandleHidReject */
  enum hid_modification	 subop;		/* ARGSUSED */
  HID			 hid;
  struct aVLink		*vlinkp;
{

#ifdef NOTYET
|		Message change = new_Message  (ST_HID_CHANGE);
|		change->subop = subop;
|		change->hid = next_hopp->hid;
|		/* send_control_Message (messagep, vlinkp); */
#endif NOTYET

    BUGRETURN (-,NoSupport,0x54101,enum ST2Errors);
}

/*
S* HIDClear ( hop_id )
s*
s*	Remove ConHsh table forwarding entry for specified HID.
s*
 *	(??? ought to release HID)
 */

  static void				/* Next bugid 0x54202 */
HIDClear( hop_id )			/* from HIDSet st2_PCBDel*2 */
  HID			 hop_id;
{
    struct aConHshEntry	*conHashp;
    int			 oldpri;


    conHashp = &( st2_ConHsh[ ConfigParm (hid_mask) & hop_id ] );
    oldpri = splnet ();
    {
	if ( conHashp->hid EQ hop_id )
	{
	    conHashp->cp = NO_PCBP;
	    conHashp->hid = (HID) 0;
	    conHashp->old = hop_id;
	}
	else if ( hop_id NE HID_RESV3 )
	    BUGSTOP (,hop_id,0x54201); /* ???then what */
    }
    (void) splx ( oldpri );

    return;
}

/*
S* HIDNew ( cp )
s*
s*	Select a new, free HID and assigns it to the specified stream.
s*	The HID is NOT activated for forwarding operations.
s*
 */

  static HID				/* Next bugid 0x54303 */
HIDNew( cp )				/* from HandleConnect */
  struct aST2pcb	*cp;
{
    int			 try_limit = 10;
    int			 tries;
    static HID		 new_hid = HID_DYNAMIC;
    static HID		 increment = 3333;


    PTRCHECK (cp, struct aST2pcb *, return ( (HID) 0 ));
    kassert (cp->hop_id[0] EQ (HID) 0);

    while ( --try_limit >= 0 )
    {
 	/* This routine looks a little weird because we increment HID
	 * values in Raw format.  So we have to check first to make
	 * sure the corresponding table entry is empty and then make
	 * sure we haven't picked one of the reserved hids.  The check
	 * for reserved hid has to be done in host byte order. */

	/* Want to pick a raw hids that are equally likely to have a
	 * non-zero low byte as a high byte.  Picking a large
	 * increment, relatively prime to 2^16 ought to work ok. */

	for ( tries = 0 ; tries <= ConfigParm (hid_mask) ; tries++ )
	{
	    struct aConHshEntry *conHashp;

	    new_hid += increment;
	    conHashp = &( st2_ConHsh[ ConfigParm (hid_mask) & new_hid ] );
	    if ( conHashp->cp EQ NO_PCBP )
	    {
		/* Must compare in host order */
		if ( ntoh2 (new_hid) >= ntoh2 (HID_DYNAMIC) )
		{
		    if ( HIDSet( cp, new_hid ) )
			return ( new_hid );

		    BUGSTOP (-,1,0x54301);
		    return ( 0 );	/* ??? error */
		}
	    }
	}
    }
    BUGSTOP (-,1,0x54302);
    return ( 0 );			/* ??? error */
}

/*
S* HIDSet ( cp, hop_id )
s*
s*	Activate the specified HID for the given connection, making it the
s*	primary HID if there is more than one.  Relay any ACCEPTs which were
s*	being held until the HID negotiation completed.
s*	Return TRUE if the HID was activated.
s*
 */

  static boolean			/* Next bugid 0x54400 */
HIDSet( cp, hop_id )			/* from HandleConnect HIDNew */
  struct aST2pcb	*cp;		/* 	ok */
  HID			 hop_id;	/* 	ok */
{
    PTRCHECK (cp, struct aST2pcb *, return ( FALSE ));
    kassert (cp->hop_id[0] EQ (HID) 0); /* ??? seems bogus - multiple */

    if ( ! HIDActivate( cp, hop_id ) )	/* Check/setup hash table */
	return ( FALSE );

    if ( cp->hop_id[0] NE (HID) 0 )	/* This HID becomes primary */
	HIDClear( cp->hop_id[0] );	/* ??? clear st2_ConHsh table */

    cp->hop_id[0] = hop_id;
    cp->hid_ready = TRUE;

    if ( cp->accepts NE NO_PSCMP )
	relay_accepts(cp);

    return ( TRUE );
}

/*
S* IPEncapFwd ( linktype, lclIPAdr, nxtIPAdr, ipsrcrutp, anhp )
s*
s*	Routine to build IP encapsulation header.
s*
 */

  static enum ST2Errors			/* Next bugid 0x54502 */
IPEncapFwd( linktype, lclIPAdr, nxtIPAdr, ipsrcrutp, anhp ) /* from NbrCreate NextHopAdd */
  enum link_type	 linktype;	/* NEIGHBOR_VLINK or not */
  INETADDR		 lclIPAdr,	/* Local IPAdr for IP Source */
			 nxtIPAdr;	/* IP Dst, assuming no srcrut */
  struct aSrcRut	*ipsrcrutp;	/* Ptr to IP SrcRut, if any */
  struct aNxtHop	*anhp;		/* Hop to be filled in */
{
    unsigned char	*ucp;
    int			 lclhdrlen = anhp->lclhdrlen, /* Room for local net */
			 iplen = sizeof (struct ip), /* w/o srcrut */
			 n,
			 oplen;
    unsigned long	 cksum;
    unsigned short	*cksmp;
    struct ip		*ipp;
    struct sockaddr_in	*sip;


    ipp = Mkp (struct ip *, &( anhp->lclsockaddr.sa_data[ lclhdrlen ] ), 0);
    if ( ((int) ipp & ALIGNMENT) NE 0 )
	BUGSTOP (-,ipp,0x54501);	/* ??? Fix lower layers! */

    sip = Mkp (struct sockaddr_in *,&( anhp->lclsockaddr ),0);
    sip->sin_family = AF_INET;		/* To use IP type in network */

    ipp->ip_v   = IPVERSION;
    ipp->ip_tos = (linktype EQ NEIGHBOR_VLINK)
			? (ConfigParm (ip_tos) >> 16) : ConfigParm (ip_tos);
    ipp->ip_len = 0;			/* for partial cksum computation */
    ipp->ip_id  = anhp->HID; /*???Set*/	/* Is net-order */
    ipp->ip_off = hton2 (IP_DF);
    ipp->ip_ttl = (linktype EQ NEIGHBOR_VLINK)
			? (ConfigParm (ip_ttl) >> 16) : ConfigParm (ip_ttl);
    ipp->ip_p   = IPPROTO_ST;
    ipp->ip_sum = 0;			/* for partial cksum computation */
    ipp->ip_src.s_addr = (u_long) htonIP (lclIPAdr);
    ipp->ip_dst.s_addr = (u_long) htonIP (nxtIPAdr);

    /* Add IP Source Route Option, if required */
    if ( ipsrcrutp NE NO_SRP )
    {
					/* First IPAdr becomes IP dst */
					/* IP dst is already last IPAdr */
	ipp->ip_dst.s_addr = (u_long) htonIP (ipsrcrutp->IPAdr[0]);

	oplen = ntoh1 (ipsrcrutp->plen);
	if ( oplen > sizeof (struct aSrcRut) )
	{
	    ucp = (unsigned char *) (ipp + 1);

	    ucp[IPOPT_OPTVAL] = hton1 (IPOPT_NOP); /* Nice alignment */
	    ucp++;

	    oplen -= 1 + 4;		/* 1 + sizeof (ipsrcrutp->IPAdr[0]) */
	    ucp[IPOPT_OPTVAL] = (   (enum PCodes) ipsrcrutp->pcode
				 EQ hton1 (STpIPSSrcRut))
				? hton1 (IPOPT_SSRR) : hton1 (IPOPT_LSRR);
	    ucp[IPOPT_OLEN]   = hton1 (oplen);
	    ucp[IPOPT_OFFSET] = hton1 (IPOPT_MINOFF);
	    Bcopy ((char *) ipsrcrutp + 8,&( ucp[IPOPT_OFFSET + 1] ),
						oplen - IPOPT_OFFSET - 1);
	    iplen += 1 + oplen;		/* NOP + SrcRut */
	}
    }

    ipp->ip_hl = iplen >> 2;		/* Total IP header size */

    /* Precompute IP checksum w/o IP length */

    cksmp = (unsigned short *) ipp;
    cksum = 0;
    for ( n = iplen / 2 ; n > 0 ; n-- )	cksum += *cksmp++;
    cksum = (cksum >> 16) + (cksum & 0xFFFF);
    cksum = (cksum >> 16) + (cksum & 0xFFFF);

    ipp->ip_sum = ONES_NEGATE (cksum);

    ipp->ip_len = iplen;		/* Leave in host order for later use */
    anhp->encaplen = iplen;		/* Encapsulation header length */

    return ( NoError );
}

/*
S* IPEncapRev ( link_type, ipp, anhp )
s*
s*	Routine to invert received IP encapsulation.
s*
 */

  static enum ST2Errors			/* Next bugid 0x54602 */
IPEncapRev( linktype, ipp, anhp )	/* from HandleConnect NbrCreate */
  enum link_type	 linktype;	/* NEIGHBOR_VLINK or not */
  struct ip		*ipp;		/* Received IP Header */
  struct aNxtHop	*anhp;		/* Hop to be filled in */
{
    unsigned char	*endp,
			*ip,
			*op,
			*ucp;
    int			 i,
			 len,
			 n;
    unsigned long	 cksum;
    unsigned short	*cksmp;
    struct ip		*ripp;
    struct sockaddr_in	*sip;


    ripp = Mkp (struct ip *, &( anhp->lclsockaddr.sa_data
			       [ anhp->gifcp->ifc_lclhdrlen ] ), 0);
    endp = (unsigned char *) (ripp + 1);

    ripp->ip_v   = IPVERSION;
    ripp->ip_tos = (linktype EQ NEIGHBOR_VLINK)
			? (ConfigParm (ip_tos) >> 16) : 0; /* normal */
    ripp->ip_len = 0;			/* for partial cksum computation */
    ripp->ip_id  = hton2 (2);
    ripp->ip_off = hton2 (IP_DF);
    ripp->ip_ttl = (linktype EQ NEIGHBOR_VLINK)
			? (ConfigParm (ip_ttl) >> 16) : ConfigParm (ip_ttl);
    ripp->ip_p   = IPPROTO_ST;
    ripp->ip_sum = 0;			/* for partial cksum computation */
    ripp->ip_src.s_addr = ipp->ip_dst.s_addr;
    ripp->ip_dst.s_addr = ipp->ip_src.s_addr;

    /* Invert IP Source Route Option, if presnt */
    len  = ipp->ip_hl << 2;
    len -= sizeof (struct ip);		/* Length of IP options received */

    /* Where to get IP options */
    ucp = Mkp (unsigned char *, ipp, sizeof (struct ip));

    while ( len > 0 )
    {
	i = ntoh1 (ucp[IPOPT_OLEN]);	/* IP Option length, ex. EOL, NOP */

	switch ( ntoh1 (ucp[IPOPT_OPTVAL]) )
	{
	case IPOPT_EOL:	len = 0;	/* stop processing */		break;
	case IPOPT_NOP:	len -= 1;	ucp += 1;	continue;	break;
	default:	if ( i EQ 0 )
			{
			    BUGSTOP (-,ucp[IPOPT_OPTVAL],0x54601);
			    len = 0;
			}
			len -= i;	ucp += i;	continue;	break;

	case IPOPT_LSRR:		/* by octets to avoid alignment prob */
	case IPOPT_SSRR:
	    /* maintain IPAdr alignment */
	    endp[0+IPOPT_OPTVAL] = hton1 (IPOPT_NOP);

	    endp[1+IPOPT_OPTVAL] = ucp[ IPOPT_OPTVAL ];
	    endp[1+IPOPT_OLEN]   = hton1 (i);
	    endp[1+IPOPT_OFFSET] = hton1 (IPOPT_MINOFF);
	    endp += 1+IPOPT_OFFSET+1;	/* For IPAdrs */

	    /* Last srcrut IPAdr becomes ip dest */
	    ip = ucp + i - 4;
	    op = (unsigned char *) &( ripp->ip_dst.s_addr );
	    *op++ = *ip++;	*op++ = *ip++;
	    *op++ = *ip++;	*op++ = *ip++;

	    /* Interior srcrut IPAdrs are inverted */
	    ip -= 8;			/* backspace IPAdr */
	    for ( i -= 7 ; i > 0 ; i -= 4 )
	    {
		*endp++ = *ip++;	*endp++ = *ip++;
		*endp++ = *ip++;	*endp++ = *ip++;
		ip -= 8;		/* backspace IPADr */
	    }

	    /* IP src becomes last option IPAdr */
	    ip = (unsigned char *) &( ipp->ip_src.s_addr );
	    *endp++ = *ip++;	*endp++ = *ip++;
	    *endp++ = *ip++;	*endp++ = *ip++;

	    len = 0;			/* Only one IP Source Route opt! */
	    break;

	  } /* end of ip option type switch  */
    } /* end of while scanning IP options */

    len = (int) endp - (int) ripp;	/* Total IP Header length */
    ripp->ip_hl = len / sizeof (octet4); /* ... into IP Header */

    /* Precompute IP chacksum w/o IP length */

    cksmp = (unsigned short *) ripp;
    cksum = 0;
    for ( n = len / 2 ; n > 0 ; n-- )	cksum += *cksmp++;
    cksum = (cksum >> 16) + (cksum & 0xFFFF);
    cksum = (cksum >> 16) + (cksum & 0xFFFF);

    ripp->ip_sum = ONES_NEGATE (cksum);
    ripp->ip_len = len;			/* Leave in host order for later use */

    anhp->encaplen = len;		/* Encapsulation header length */
    sip = (struct sockaddr_in *) &( anhp->lclsockaddr );
    sip->sin_family = AF_INET;		/* To use IP type in network */ 

    sip->sin_addr.s_addr = ripp->ip_dst.s_addr;

    return ( NoError );
}

/*
S* log_FlowSpec ( hflowp, subr )
s*
s*	NOTYET
s*
 */
#ifdef NOTYET
|	
|	  void
|	_log_FlowSpec( hflowp, subr )	/* from <nowhere> */
|	  struct aFlowSpec3 *hflowp;
|	  char *subr;
|	{
|	    (void) printf ( "%s (flowspec 0x%x)\n", subr, hflowp );
|	    (void) printf ( "    Version: %d\n", hflowp->FlowVer );
|	    (void) printf ( "    DutyFactor: %d/256\n", hflowp->DutyFactor );
|	    (void) printf ( "    ErrorRate: %d\n", hflowp->ErrorRate );
|	    (void) printf ( "    Precedence: %d\n", hflowp->Precedence );
|	    (void) printf ( "    Reliability: %d/256\n", hflowp->Reliability );
|	    (void) printf ( "    Tradeoffs: %d\n", hflowp->Tradeoffs );
|	    (void) printf ( "    RecoveryTimeout: %d ms\n", hflowp->RecoveryTimeout );
|	    (void) printf ( "    LimitOnCosts: %d\n", hflowp->LimitOnCosts );
|	    (void) printf ( "    LimitOnDelay: %d ms\n", hflowp->LimitOnDelay );
|	    (void) printf ( "    LimitOnPDUBytes: %d\n", hflowp->LimitOnPDUBytes );
|	    (void) printf ( "    LimitOnPDURate: %d\n", hflowp->LimitOnPDURate );
|	    (void) printf ( "    MinBytesXRate: %d\n", hflowp->MinBytesXRate );
|	    (void) printf ( "    AccdMeanDelay: %d ms\n", hflowp->AccdMeanDelay );
|	    (void) printf ( "    AddDelayVariance: %d\n", hflowp->AddDelayVariance );
|	    (void) printf ( "    DesPDUBytes: %d\n", hflowp->DesPDUBytes );
|	    (void) printf ( "    DesPDURate: %d\n", hflowp->DesPDURate );
|	
|	    return;
|	}
#endif NOTYET

/*
S* log_Forwarder ( cp, subr )
s*
s*	Routine to log information about a connection.
s*
 */

  static void				/* Next bugid 0x54700 */
log_Forwarder( cp, subr )		/* from CtrlMsgAddTrgSet st2_PCBDel */
  struct aST2pcb	*cp;
  char			*subr;
{
    int i;
    struct Target *targp;

    PTRCHECK (cp, struct aST2pcb *, return;);

    (void) printf ( "%s (forw 0x%x) HIDs =", subr, cp );

    for ( i = 0 ; i < MAX_HIDS_PER_STREAM ; i++ )
    {
	if ( cp->hop_id[ i ] NE (HID) 0 )
	    (void) printf ( " %d", cp->hop_id[ i ] );
    }

    (void) printf ( "\n" );

    ForAllTargets (cp, targp)
    {
	log_Target( targp, "   with target " );
    }
    return;
}

/*
S* log_NextHop ( nexthop, subr )
s*
s*	Routine to log information from a next-hop structure.
s*
 */

  static void				/* Next bugid 0x54800 */
log_NextHop( nexthop, subr )		/* from st2_PCBUpdate*2 */
  struct NextHop	*nexthop;
  char			*subr;
{
    PTRCHECK (nexthop, struct NextHop *, return;);

    if ( nexthop NE NO_NEXTHOPP )
	(void) printf ( "%s (nexthop 0x%x) nhop_state %d %s\n",
		      subr, nexthop, nexthop->nhop_state,
		      (st2_nnhop_states > (int) nexthop->nhop_state)
		      ? st2_nhop_states[ (int) nexthop->nhop_state ].namep
		      : "?" );		/*?? next agent inetaddress */
    else
	(void) printf ( "%s (nexthop 0x%x)\n", subr, nexthop );

    return;
}

/*
S* log_Parms ( pscmp )
s*
s*	Routine to identify/log those parameters that wer present in an
s*	SCMP Message.
s*
 */

  static void				/* Next bugid 0x54900 */
log_Parms( pscmp )			/* from DoConnect */
  struct aPrsSCM	*pscmp;
{
    if ( pscmp->FreeHIDsp )
	(void) printf ( "FreeHIDsp: %d %d\n",
		      pscmp->FreeHIDsp->pcode, pscmp->FreeHIDsp->plen );
    if ( pscmp->FlowSpecp )
	(void) printf ( "FlowSpecp: %d %d\n",
		      pscmp->FlowSpecp->pcode, pscmp->FlowSpecp->plen );
    if ( pscmp->Groupp )
	(void) printf ( "Groupp: %d %d\n",
		      pscmp->Groupp->pcode, pscmp->Groupp->plen );
    if ( pscmp->HIDp )
	(void) printf ( "HIDp,: %d %d\n",
		      pscmp->HIDp->pcode, pscmp->HIDp->plen );
    if ( pscmp->McastAdrp )
	(void) printf ( "McastAdrp: %d %d\n",
		      pscmp->McastAdrp->pcode, pscmp->McastAdrp->plen );
    if ( pscmp->Namep )
	(void) printf ( "Namep: %d %d\n",
		      pscmp->Namep->pcode, pscmp->Namep->plen );
    if ( pscmp->NxtHopIPAdrp )
	(void) printf ( "NxtHopIPAdrp: %d %d\n",
		      pscmp->NxtHopIPAdrp->pcode, pscmp->NxtHopIPAdrp->plen );
    if ( pscmp->Originp )
	(void) printf ( "Originp: %d %d\n",
		      pscmp->Originp->pcode, pscmp->Originp->plen );
    if ( pscmp->OrgTsmpp )
	(void) printf ( "OrgTsmpp: %d %d\n",
		      pscmp->OrgTsmpp->pcode, pscmp->OrgTsmpp->plen );
    if ( pscmp->RecordRoutep )
	(void) printf ( "RecordRoutep: %d %d\n",
		      pscmp->RecordRoutep->pcode, pscmp->RecordRoutep->plen );
    if ( pscmp->RFlowSpecp )
	(void) printf ( "RFlowSpecp: %d %d\n",
		      pscmp->RFlowSpecp->pcode, pscmp->RFlowSpecp->plen );
    if ( pscmp->RGroupp )
	(void) printf ( "RGroupp: %d %d\n",
		      pscmp->RGroupp->pcode, pscmp->RGroupp->plen );
    if ( pscmp->RHIDp )
	(void) printf ( "RHIDp: %d %d\n",
		      pscmp->RHIDp->pcode, pscmp->RHIDp->plen );
    if ( pscmp->RNamep )
	(void) printf ( "RNamep: %d %d\n",
		      pscmp->RNamep->pcode, pscmp->RNamep->plen );
    if ( pscmp->UserDatap )
	(void) printf ( "UserDatap: %d %d\n",
		      pscmp->UserDatap->pcode, pscmp->UserDatap->plen );
    if ( pscmp->ErrdPDUp )
	(void) printf ( "ErrdPDUp: %d %d\n",
		      pscmp->ErrdPDUp->pcode, pscmp->ErrdPDUp->plen );
    return;
}

/*
S* log_Route ( routep, subrp )
 */

  static void				/* Next bugid 0x55000 */
log_Route( routep, subrp )		/* from st2_CMPOutput */
  struct Route		*routep;
  char			*subrp;
{
    struct aNxtHop	*nxthopp = routep->nxthopp;
    unsigned short	*sh;


    PTRCHECK (routep, struct Route *, return;);

    (void) printf ( "%s (0x%x): route", subrp, routep );

    if ( nxthopp NE NO_NXTHOPP )
    {
	sh = Mkp (unsigned short *, &( nxthopp->lclsockaddr.sa_data[0] ), 0);
	(void) printf ( "sockaddr (%d) %4x %4x %4x %4x %4x %4x %4x\n",
		      nxthopp->lclsockaddr.sa_family,
		      sh[0], sh[1], sh[2], sh[3],
		      sh[4], sh[5], sh[6] );
    }
    else
	(void) printf ( "but no next agent pointer!\n" );

    return;
}

/*
S* log_SAP ( stsap, stringp )
 */

  static void				/* Next bugid 0x55100 */
log_SAP( stsap, stringp )		/* from st2_PCBUpdate*3 */
  struct STSAP		*stsap;
  char			*stringp;
{
    unsigned char	*sap = stsap->SAPp;
    int			 n = stsap->SAPBytes;


    (void) printf ( "%s protocol %d, sap %d bytes =",
		  stringp, stsap->NextPcol, stsap->SAPBytes );
    while ( --n >= 0 )
	(void) printf ( " %2x", *sap++ );

    (void) printf ( "\n" );

    return;
}


/*
S* log_Stream ( cp, subrp )
s*
s*	Routine to log the pcb and name of a stream.
s*
 */

  static void				/* Next bugid 0x55200 */
log_Stream( cp, subrp )			/* from st2_DoConnect st2_PCBDel */
  struct aST2pcb	*cp;
  char			*subrp;
{
    struct aName *sNamep = &( cp->Name );

    if ( cp NE NO_PCBP )
	(void) printf ( "%s (aST2pcb %x) named {%d:%x:%x}\n",
		      subrp, cp,
		      sNamep->UniqId,
		      sNamep->IPAdr,
		      sNamep->Timstmp );
    else
	(void) printf ( "%s (aST2pcb %x)\n", subrp, cp );

    return;
}

/*
S* log_Target ( targetp, subr )
s*
s*	Routine to log a target's state.
s*
 */

  static void				/* Next bugid 0x55300 */
log_Target( targetp, subr )		/* from CtrlMsgAddTarg CtrlMsgAddTrgSet DoConnect DoDisconnect DoRefuse*2 log_Forwarder st2_DoAccept*2 st2_PCBDel st2_PCBUpdate*8 TargAdd */
  struct Target		*targetp;
  char			*subr;
{
    PTRCHECK (targetp, struct Target *, return;);

    if ( targetp NE NO_TRGP )
    {
	(void) printf ( "%s (tp %x cp %x) %s %s %s $ IPAdr %x/%x/%x vlp %x err %d\n",
		subr, targetp, targetp->cp,
		(st2_ntarg_states > (int) targetp->targstate[(int)Overall])
		? st2_targ_states[(int)targetp->targstate[(int)Overall]].namep
		: "?",
		(st2_ntarg_states > (int) targetp->targstate[(int)Prevhop])
		? st2_targ_states[(int)targetp->targstate[(int)Prevhop]].namep
		: "?",
		(st2_ntarg_states > (int) targetp->targstate[(int)Nexthop])
		? st2_targ_states[(int)targetp->targstate[(int)Nexthop]].namep
		: "?",
		targetp->targetIPAdr, targetp->nxtSTIPAdr, targetp->routetoIPAdr,
		targetp->vlinkp, targetp->error );
    }
    else
	(void) printf ( "%s (targetp %x)\n", subr, targetp );

    return;
}

/*
S* log_VLink ( vlinkp, subr )
s*
s*	Routine to log information about the specified virtual link.
s*
 */

  static void				/* Next bugid 0x55400 */
log_VLink( vlinkp, subr )		/* from CtrlMsgSend */
  struct aVLink		*vlinkp;
  char			*subr;
{
    PTRCHECK (vlinkp, struct aVLink *, return;);

    if ( vlinkp NE NO_VLINKP )
	(void) printf ( "%s (vlinkp 0x%x) VLID %d %x [remote %d %x]\n",
		      subr, vlinkp,
		      vlinkp->local_vlid, vlinkp->lclIPAdr,
		      vlinkp->remote_vlid, vlinkp->rmtIPAdr );
    else
	(void) printf ( "%s (vlinkp 0x%x)\n", subr, vlinkp );

    return;
}

/*
S* MTUAdjust ( contblp, nxthopp )
s*
s*	Routine to update the minimum MTU encountered along the data path.
s*	??? to st2?routing c w/ifnet
s*
 */

  static void				/* Next bugid 0x55503 */
MTUAdjust( cp, nxthopp )		/* from st2_DoConnect st2_PCBUpdate */
  struct aST2pcb	*cp;
  struct aNxtHop	*nxthopp;
{
    struct aNetIf	*gifcp = nxthopp->gifcp;


    if ( gifcp NE NO_NETIFP )
    {
	if ( gifcp->ifc_mtu <= 0 )
	{
	    BUGSTOP (,gifcp->ifc_name,0x55501);
	}
	else if ( (cp->mtu EQ 0)  OR  (gifcp->ifc_mtu < cp->mtu) )
	    cp->mtu = gifcp->ifc_mtu;	/* ??? recompute on disconnect */
    }
    else
    {
	BUGSTOP (-,nxthopp,0x55502);
    }

    return;
}

/*
S* NbrCreate ( nxtstipadr, flags, pscmp, targetp )
s*
s*	Routine to add a new neighbor.  Depends on both ipadr and any
s*	IP Encapsulation or Source Routing.
s*
 */

  static struct aNeighbor *		/* Next bugid 0x55606 */
NbrCreate( nxtstipadr, flags, pscmp, targetp ) /* from HandleConnect HandleHello st2_PCBUpdate */
  INETADDR		 nxtstipadr;
  int			 flags;		/* FRF_USEIP, FRF_DIRECT, FRF_NOSKT */
					/* ARGSUSED */
					/* IP Source Route info for */
  struct aPrsSCM	*pscmp;		/* upstream nbr if not NULL */
  struct Target		*targetp;	/* downstream nbr if not NULL */
{
    enum ST2Errors	 error;
    int			 extra,
			 oldpri;
    struct aNeighbor	*np,
			*npe;
    struct aNeighborList *anlp,
			*unfilled_nlp,
			*lastnlp;
    struct aNxtHop	*anhp,
			 tmpanh;
    struct aPktDesc	*pp;
    struct aRouteChoiceList arcl;
    struct aVLink	*vlinkp;
    struct sockaddr_in	*sip;


    /* Build aNxtHop so we can compare encapsulation and/or source routing */
    anhp = &( tmpanh );
    Bzero (anhp,sizeof (tmpanh));
    anhp->refcount = 1;

    sip = (struct sockaddr_in *) &( anhp->lclsockaddr );
    sip->sin_family = ConfigParm (coip_family);

    if ( pscmp NE NO_PSCMP )		/* Base encap. on received message */
    {
	/* similar code in HandleConnect */
	anhp->gifcp = pscmp->gifcp;
	anhp->fwdtoIPAdr = pscmp->hoscmphdr.SndrIPAdr;
	sip->sin_addr.s_addr = pscmp->hoscmphdr.SndrIPAdr;
	anhp->lclhdrlen = pscmp->gifcp->ifc_lclhdrlen;

	/* Record any encapsulation info for link */
	if ( pscmp->encaplen > 0 )
	{
	    error = IPEncapRev( NEIGHBOR_VLINK, pscmp->encaphdrp, anhp );
	    if ( error NE NoError )
		return ( NO_NBRP );
	}

	Bzero (&( arcl ),sizeof (struct aRouteChoiceList));
	st2_FindRoutes( sip->sin_addr.s_addr, pscmp->hoscmphdr.SndrIPAdr,
		        &( arcl ), FRF_DIRECT | FRF_NOSKT |
		        (pscmp->encaplen > 0) ? FRF_USEIP : 0 );
	if ( (arcl.nxtfree EQ 0)  OR  (arcl.RouteChoices[0].error NE NoError) )
	{   /* ??? needs work when route table has mult. routes or srcruts */
	    BUGSTOP (-,arcl.RouteChoices[0].error,0x55601);
	    return ( NO_NBRP );
	}
	sip->sin_addr.s_addr = arcl.RouteChoices[0].fwdtoIPAdr;
	anhp->fwdtoIPAdr = arcl.RouteChoices[0].fwdtoIPAdr;
    }
    else if ( targetp NE NO_TRGP )	/* Base encap. on target parm */
    {
	/* Initialize aNxtHop; similar code in NextHopAdd */

	anhp->gifcp = targetp->curchoicep->gifcp;
	anhp->fwdtoIPAdr = targetp->curchoicep->fwdtoIPAdr;
	sip->sin_addr.s_addr = targetp->curchoicep->fwdtoIPAdr;
	anhp->lclhdrlen = targetp->curchoicep->gifcp->ifc_lclhdrlen;

	/* Build IP header if IP Encapsulation or IP Source Route */
	if (   (targetp->ipsrcrutp NE (struct aSrcRut *) NULL)
	    OR (targetp->curchoicep->IPencapsulate NE 0) )
	{
	    (void) IPEncapFwd( NEIGHBOR_VLINK, targetp->curchoicep->lclIPAdr,
			       targetp->curchoicep->nxtIPAdr,
			       targetp->ipsrcrutp, anhp );
	}
    }
    else				/* gripe! */
    {
	BUGSTOP (-,Inconsistency,0x55602);
	return ( NO_NBRP );
    }

    /* ==================================================================== */
    /* NOTE: there's similar code in NbrFind () */

    unfilled_nlp = NO_NBRLSTP;
    anlp = &( st2_neighbors );

    /* BEGIN HIGH PRIORITY */

    oldpri = splnet ();
    {
	/* for all neighbor list segments */
	for ( ; anlp NE NO_NBRLSTP ; anlp = anlp->nxtNeighborp )
	{
	    /* remember first neighborlist with a free entry */
	    if (    (anlp->nxtfree < anlp->allocated)
		AND (unfilled_nlp EQ NO_NBRLSTP) )
		unfilled_nlp = anlp;

	    /* for all neighbors in list, if any */
	    np = &( anlp->Neighbors[0] );
	    npe = np + anlp->nxtfree;
	    for ( ; np < npe ; ++np )
	    {
		if ( (np->flags & NbrFlgDelete) NE 0 )
		    continue;

		if ( np->rmtIPAdr NE nxtstipadr ) /* ??? check lclIPAdr, too */
		    continue;

		if ( np->nxthop.encaplen NE anhp->encaplen )
		    continue;

		if (    (anhp->encaplen > 0) /* 12 for IP src/dst/opts */
		    AND (Bcmp (&( ((struct sockaddr *) sip)
				    ->sa_data[anhp->     lclhdrlen + 12] ),
			       &( np->nxthop.lclsockaddr.
				      sa_data[np->nxthop.lclhdrlen + 12] ),
			       anhp->encaplen - 12) NE 0) )
		    continue;

		/* ??? make sure np->vlinkp not NO_VLINKP */
		(void) splx ( oldpri );

		/* -------------------------------------------------------- */
		/* Make sure remote VLId is up to date */

		if (    (pscmp NE NO_PSCMP)
		    AND (pscmp->hoscmphdr.OpCode EQ (int) OpHello)
		    AND (np->vlinkp NE NO_VLINKP)
		    AND (pscmp->hoscmphdr.SVLId NE np->vlinkp->remote_vlid) )
		{
		    BUGSTOP (,  (np->vlinkp->remote_vlid << 16)
			      | pscmp->hoscmphdr.SVLId,0x55605);
		    np->vlinkp->remote_vlid = pscmp->hoscmphdr.SVLId;
		}

		return ( np );		/* Already a neighbor */

	    } /* end of for all nbrs in segment loop */

	    lastnlp = anlp;
	}

	/* ---------------------------------------------------------------- */
	/* Neighbor isn't already known.  Add it. */

	anlp = unfilled_nlp;		/* may be NO_NBRLSTP */

	if ( anlp EQ NO_NBRLSTP )	/* all neighbor lists are full */
	{
	    /* new NeighborList */
	    pp = NO_PKTDSCP;
	    anlp = GetSpace (struct aNeighborList *, &( pp ),
			      PKT_NEIGHBOR, PKT_MD_NEW, PKT_XA_PS,
			      0,0, sizeof (struct aNeighborList), 0,0);

	    if ( anlp EQ NO_NBRLSTP )
	    {
		if ( pp NE NO_PKTDSCP )
		    FreePkts (pp);
		(void) splx ( oldpri );
#ifdef STIIDEBUG
		(void) printf ( "NbrCreate (GetSpace): can't add %4x\n",
			        nxtstipadr);
#endif /*STIIDEBUG*/
		BUGSTOP (-,nxtstipadr,0x55603);
		return ( NO_NBRP );
	    }

	    extra = pp->pkt_postpad / sizeof (struct aNeighbor);
	    if ( extra > 0 )
	    {
		AddPktSpace (pp,(extra * sizeof (struct aNeighbor)));
	    }

	    Bzero (anlp,  sizeof (struct aNeighborList)
			+ (extra * sizeof (anlp->Neighbors[0])));
	    anlp->allocated = DimensionOf (anlp->Neighbors) + extra;
	    anlp->selfp = pp;
	    anlp->objId = 0x1157;	/* (arbitrary number for now) */

	    lastnlp->nxtNeighborp = anlp;  /* lastly, link in this block */
	}

	np = &( anlp->Neighbors[ anlp->nxtfree++ ] );
	/* ??? may omit if zero on failure below */
	Bzero (np,sizeof (struct aNeighbor));
	np->flags |= NbrFlgInUse;
    }
    (void) splx ( oldpri );

    /* END HIGH PRIORITY */
    /* -------------------------------------------------------------------- */
    /* Fill in new neighbor info */

    uniqtime ( &( np->start ) );
    np->statechange = np->start;
    np->rmtIPAdr = nxtstipadr;
    np->avg_rtt_ms = (MSecs) ConfigParm (nbrRTT0); /* 4 sec. RTT to nbr */
    np->delay2nbr = (FracTime) 0x4200 * ConfigParm (nbrRTT0);
					/* 4200 >= 2**24/1000 */
    Count (nbrscreated,1);

    vlinkp = VLinkNew( NO_PCBP, DOWNSTREAM_VLINK, NEIGHBOR_VLINK );
    if ( vlinkp EQ NO_VLINKP )
    {
	Bzero (np,sizeof (struct aNeighbor));
        anlp->nxtfree--;		/* releases np entry */
	BUGSTOP (-,CantGetResrc,0x55604);
	return ( NO_NBRP );
    }

    np->vlinkp = vlinkp;
    vlinkp->np = np;

    np->nxthop = *anhp;			/* structure copy */
    anhp = &(np->nxthop);
    vlinkp->route.type = TO_NEIGHBOR;
    vlinkp->route.nxthopp = anhp;

    np->gifcp = anhp->gifcp;
    np->lclIPAdr = st2_LocalIPAdrOfGifc( np->gifcp );
    vlinkp->lclIPAdr = np->lclIPAdr;
    vlinkp->rmtIPAdr = np->rmtIPAdr;
    
    /* Make sure remote VLId is up to date */

    if (    (pscmp NE NO_PSCMP)
	AND (pscmp->hoscmphdr.OpCode EQ (int) OpHello)
	AND (pscmp->hoscmphdr.SVLId NE vlinkp->remote_vlid) )
    {
	BUGSTOP (,  (vlinkp->remote_vlid << 16)
		  | pscmp->hoscmphdr.SVLId,0x55605);
	np->vlinkp->remote_vlid = pscmp->hoscmphdr.SVLId;
    }

    /* np->nextp = Link to next neighbor (?on same network) */

    np->rtthello_ms = ConfigParm (hellobackoff); /* 20 sec.  adjusted later */
    np->rtt_var_ms = ConfigParm (hellobackoff) / 2; /* ??? until complete */
/*  np->rcvryto, hello_usec, and NbrFlgActive  set by NbrInsert */

    /* ??? send hello to get initial delay to neighbor estimate */

    return ( np );
}

/*
S* NbrDelete ( np )
s*
s*	Routine to delete a neighbor.	NOTYET
s*
 */
#ifdef NOTYET

  static void				/* Next bugid 0x55700 */
NbrDelete( np )				/* from */
   struct aNeighbor	*np;
{
    Count (nbrsdeleted,1);

    return;
}
#endif NOTYET

/*
S* NbrFailed ( np, reason )
s*
s*	Routine to notify streams using a failed neighbor.
s*
 */

  static void				/* Next bugid 0x55803 */
NbrFailed( np, reason )			/* from HandleHello st2_ClkSlow */
  struct aNeighbor	*np;
  enum ST2Errors	 reason;	/* RemoteRestart or NetworkFailure/
					   STAgentFailure/network */
{
    enum targ_state	 nxtstate;
    int			 upstream;
    struct aST2pcb	*cp;
    struct aVLink	*vlinkp,
			*nxtvlinkp;
    struct Target	*targetp;
    struct timeval	 died;
    unsigned char	*ucp = (unsigned char *) &( np->rmtIPAdr );


    Count (nbrsfailed,1);
    if ( reason EQ RemoteRestart )
	Count (nbrsfailedRemoteRestart,1);
    np->flags &= ~NbrFlgActive;
    Count (nbrsdeactivated,1);
    uniqtime ( &( died ) );
    np->active_msec += MsecBetween ( &( died ), &( np->statechange ) );
    np->statechange = died;
    np->hellorcvd.tv_sec = 0;		/* forget we ever heard from nbr */
    np->hellotimer_ms = 0;
    if ( np->vlinkp NE NO_VLINKP )
	np->vlinkp->remote_vlid = VLID_NONE;

    IfLog (DBG_NBR_ACT)
    {
	printf ( "## NbrFailed %u.%u.%u.%u ( %x, %d )\n",
		 ucp[0],ucp[1],ucp[2],ucp[3], np, reason);  /*DEBUG*/
    }
    BUGSTOP (,np->rmtIPAdr,0x55801);


    /* Use np->failurep to find all streams going to, or coming from, the
       neighbor and change the stream state.  Changing the state will cause
       other code to send the appropriate messages and take the appropriate
       actions */

    for ( vlinkp = np->failurep ; vlinkp NE NO_VLINKP ; vlinkp = nxtvlinkp )
    {
	nxtvlinkp = vlinkp->failurep;	/* next item, in case vlink flushed */

	if ( vlinkp->type EQ NEIGHBOR_VLINK ) /* ??? shouldn't be used now */
	{
	    NbrRemove( vlinkp );	/* now deactivated */
	    vlinkp->remote_vlid = VLID_NONE;
	    continue;
	}

	if ( upstream = (vlinkp->direction EQ UPSTREAM_VLINK) )
	{
	    Count (nbrsfailedstream,1);
	}

	cp = vlinkp->cp;		/* A stream via failed link */

	ForAllTargets (cp, targetp)
	{
	    if (   IsTargState (targetp, TS_DEAD)
		OR (	(targetp->vlinkp NE vlinkp)
		    AND (NOT upstream)) )
		continue;

	    IfLog (DBG_NBR_ACT)
	    {
		ucp = (unsigned char *) &( targetp->targetIPAdr );
		printf ( "##  target %-2u %u.%u.%u.%u\n",
			 targetp->target_index,
			 ucp[0],ucp[1],ucp[2],ucp[3] );  /*DEBUG*/
	    }
	    BUGSTOP (,targetp->targetIPAdr,0x55802);

	    if ( ConfigFlag (ST2FlgNoFail) )
		continue;

	    /* ------------------------------------------------------------ */

	    nxtstate = TS_FREE;

	    /* downstream vlink state machine */
#ifdef NOTYET
	    if ( IsTargStateIn (targetp, TS_NEW,  TS (TS_ROUTEOPT)
				TS (TS_ROUTING)  TS (TS_WAITNHOP)
				TS (TS_RSRCREPLY) TS (TS_WANTOPEN) ) )
	    {
		nxtstate = TS_SUSPENDED;
	    }
#else /*NOTYET*/
	    if ( IsTargStateIn (targetp, TS_NEW,
				TS (TS_ROUTEOPT)  TS (TS_ROUTING)) )
	    {
		nxtstate = TS_RELEASED;
	    }
	    else if ( IsTargStateIn (targetp, TS_WAITNHOP,
				     TS (TS_RSRCREPLY) TS (TS_WANTOPEN)) )
	    {
		nxtstate = TS_CLOSED;
	    }
#endif /*NOTYET*/
	    else if ( IsTargStateIn (targetp, TS_OPEN, TS (TS_OPENING)) )
	    {
#ifdef NOTYET
		nxtstate =  TS_WANTSUSPEND;
#else /*NOTYET*/
		nxtstate = TS_WANTCLOSED;
#endif /*NOTYET*/
	    }

	    if ( nxtstate NE TS_FREE )
	    {
		SetTargState (targetp,nxtstate);
		cp->needupdate ++ ;	/* work to do */
	    }

	    /* ------------------------------------------------------------ */

	    nxtstate = TS_FREE;

	    if ( upstream )		/* UPSTREAM_VLINK */
	    {
		/* upstream vlink state machine */
		if ( IsTargState (targetp, TS_CONNECTED) )
		{
#ifdef NOTYET
		    nxtstate = TS_WANTHOLD;
#else /*NOTYET*/
		    nxtstate = TS_FAILED;
#endif /*NOTYET*/
		}

		targetp->error = reason; /* What failed */
		cp->fwdrerror = reason;
	    }
	    else			/* DOWNSTREAM_VLINK */
	    {
		Count (nbrsfailedtarget,1);
		targetp->curchoicep->error = reason;
		targetp->curchoicep->flags |= RouteFailed | RouteDest;
	    }

	    if ( nxtstate NE TS_FREE )
	    {
		SetTargState (targetp,nxtstate);
		cp->needupdate ++ ;	/* work to do */
	    }

	} /* end of for all targets loop */

	if ( cp->needupdate > 0 )
	    st2_PCBUpdate (cp);
	    /* cp, vlinkp may be invalid */

    } /* end of for all streams with neighbor loop */

    return;
}

/*
S* NbrFind ( ipadr )	NOTYET
s*
s*	Return the struct aNeighbor * pointer for the specified ipadr, or
s*	NO_NBRP if no such neighbor entry.
 */
#ifdef NOTYET

  static struct aNeighbor *		/* Next bugid 0x55900 */
NbrFind( ipadr )			/* from */
  INETADDR		 ipadr;
{
    struct aNeighbor	*np, *npe;
    struct aNeighborList *anlp;


    /* NOTE: there's similar code in NbrCreate () */

    /* Linear search through the lists of neighbors */

    /* for all neighbor lists */
    for ( anlp = &( st2_neighbors ) ; anlp NE NO_NBRLSTP
	 ; anlp = anlp->nxtNeighborp )
    {
	/* for all neighbors in list, if any */
	np = &( anlp->Neighbors[0] );
	npe = np + anlp->nxtfree;
	for (  ;  np < npe;  ++np )
	{
	    if ( (np->flags & NbrFlgDelete) NE 0 )
		continue;

	    if ( np->rmtIPAdr EQ ipadr )
		return ( np );

	} /* end of for all neighbors in segment loop */
    } /* end of for all neighbor segments loop */

    return ( NO_NBRP );		/* neighbor isn't already known. */
}
#endif NOTYET

/*
S* NbrInsert ( vlinkp, RecoveryTimeout value )
s*
s*	Set vlinkp->rcvryto computed from the value passed, and place
s*	vlinkp on the failurep for neighbor vlinkp->remotehost, sorted
s*	by increasing values of rcvryto.  If rcvryto < previous
s*	minimum rcvryto, adjust np->interval and np->rcvryto.
*/

  static int				/* Next bugid 0x56006 */
NbrInsert( vlinkp, rcvryto )		/* from TargByApplIfc CtrlMsgRelay HandleAccept NbrSort st2_DoAccept */
  struct aVLink		*vlinkp;
  unsigned int		 rcvryto;	/* Host-order value desired */
{
    unsigned char	*ucp;
    long		 dt;
    unsigned long	 msto,
			 usecto;
    struct aNeighbor	*np;
    struct aVLink	*nvlp,
			*vlp;


    /* Require a valid (struct aNeighbor *) entry.  The pointer will be derived
       from vlinkp->np. */

    if ((np = vlinkp->np) EQ NO_NBRP )
    {
	BUGSTOP (-,vlinkp,0x56001);
	return ( 0 );
    }

    /* Make sure it is a vlink to a gateway. */

    if ( vlinkp->type EQ GATEWAY_VLINK )
    {
	if ( vlinkp->cp EQ NO_PCBP )
	{
	    BUGSTOP (-,vlinkp,0x56004);
	    return ( 0 );
	}
    }
#ifdef NOTWANTED
    else if ( vlinkp->type EQ NEIGHBOR_VLINK )
    {
	if (   (vlinkp->cp NE NO_PCBP)
	    OR (np->vlinkp NE vlinkp) )
	{
	    BUGSTOP (-,vlinkp,0x56004);
	    return ( 0 );
	}
    }
#endif NOTWANTED
    else				/* application vlink */
    {
	BUGSTOP (-,vlinkp,0x56005);
	return ( 0 );
    }


    /* See if this was already put on a failure list.*/
    if ( vlinkp->on_failure_np NE NO_NBRP )
    {
	if ( vlinkp->on_failure_np EQ np )
	    return ( 0 );		/* already on proper failure list */
	else
	{
	    BUGSTOP (-,vlinkp,0x56002);
	    return ( 0 );		/* on the wrong failure list */
	}
    } 

    /* -------------------------------------------------------------------- */
    /* Figure out effective recovery timeout */

    msto = rcvryto;			/* result in milliseconds (ms) */

    /* A RecoveryTimeout of 0 is interpreted as "use default" */
    if ( msto EQ 0 )
	msto = ConfigParm (rcvrytodef);
    /* msto is now >0, if default value is >0 */

    if ( msto < ConfigParm (rcvrytomin) )
	msto = ConfigParm (rcvrytomin);

    if ( msto < (500 /*msec clk_slow*/ * ConfigParm (hellolossfactor)) )
	msto = 500 * ConfigParm (hellolossfactor);

    vlinkp->rcvryto = msto;

    /* -------------------------------------------------------------------- */
    /* Insert vlink into failurep list, sorted by incr. recovery timeout */

    vlp = np->failurep;			/* head of linked list */
    if (   (vlp EQ NO_VLINKP)
	OR (msto < vlp->rcvryto) )
    {
	/* new head of chain, hence a new, smaller, recovery timeout */

	vlinkp->failurep = vlp;		/* insert vlinkp at head of chain */
	np->failurep = vlinkp;

	/* adjust HELLO interval */
	np->rcvryto = msto;		/* (ms) interpreted timeout */

	/* convert to usec and divide by Hello loss factor */
	usecto = 1000 * msto / ConfigParm (hellolossfactor);
	/*        old interval     -     new (>0) */
	dt = (long) np->hello_usec - (long) usecto;
	np->hello_usec = usecto;

	/* fix up np->sendhello */
	if (np->sendhello.tv_sec EQ 0)  /* sendhello needs initializing */
	{
	    uniqtime ( &( np->sendhello ) ); /* send first HELLO @ next tick */
	}
	else			/* adjust sendhello for new, smaller interval */
	{
	    /* subtract dt from sendhello */
	    np->sendhello.tv_sec -= dt / 1000000;
	    dt %= 100000;
	    dt = (long) np->sendhello.tv_usec - dt;
	    if (dt < 0)
	    {
		dt += 1000000;
		np->sendhello.tv_sec -= 1;
	    }
	    np->sendhello.tv_usec = dt;
	}

	IfLog (DBG_NBR_ACT)
	    printf ( "NbrInsert: (head) %x->%x, np %x, total %d\n",
		     vlinkp, vlinkp->failurep, np, np->link_count + 1 );

    }
    else
    {			/* vlinkp becomes item 2..N in chain */
	nvlp = vlp->failurep;		/* next item, if any */
	while (nvlp NE NO_VLINKP  AND  msto >= nvlp->rcvryto)
	{
	    vlp = nvlp;
	    nvlp = vlp->failurep;	/* next item, if any */
	}

	vlinkp->failurep = nvlp;
	vlp->failurep = vlinkp;

	/* no adjustment to np->hello_usec or np->rcvryto needed */

	IfLog (DBG_NBR_ACT)
	    printf ( "NbrInsert: (Nth) %x->%x->%x, np %x, total %d\n",
		     vlp, vlinkp, vlinkp->failurep, np, np->link_count + 1);
    }	

    vlinkp->on_failure_np = np;

    np->link_count++;
    Count (nbrsinserted,1);

    if ( (np->flags & NbrFlgActive) EQ 0 )
    {
	IfLog (DBG_NBR_ACT)
	{
	    ucp = (unsigned char *) &( np->rmtIPAdr );
	    printf ( "## Activating nbr %u.%u.%u.%u @ %u msec (%x)\n",
		    ucp[0],ucp[1],ucp[2],ucp[3], msto, np);
	}
	BUGSTOP (,np->rmtIPAdr,0x56003);

	np->hellotimer_ms = 0;		/* Not yet heard a HELLO */
	uniqtime ( &( np->hellorcvd ) ); /* start timeout interval */
	np->statechange = np->hellorcvd;
	np->flags |= NbrFlgActive;
#ifdef NOTWANTED
	NbrInsert( np->vlinkp, ConfigParm (rcvrytonbr) );
#endif NOTWANTED
	Count (nbrsactivated,1);
    }

    if ( (msto NE rcvryto)  AND  (rcvryto NE 0) )
	return ( msto );		/* not accepting requested */

    return ( 0 );			/* requested value accepted */
}

/*
S* NbrRemove ( vlinkp )
s*
s*	Remove vlinkp from vlinkp->np->failurep.  If
s*	vlinkp->rcvryto controlled np->hello_usec and
s*	np->rcvryto, recompute the hello interval.
*/

  static void				/* Next bugid 0x56105 */
NbrRemove( vlinkp )			/* from NbrSort VLinkDel */
  struct aVLink		*vlinkp;
{
    struct aNeighbor	*np;
    struct aVLink	*vlp,
			*nvlp;
    struct timeval	 deactivated;


    if ( vlinkp EQ NO_VLINKP )
	return;

    if (    (vlinkp->type NE GATEWAY_VLINK)
	AND (vlinkp->type NE NEIGHBOR_VLINK) )
	return;

    np = vlinkp->np;
    if ( np EQ NO_NBRP )
    {
	BUGSTOP (-,vlinkp,0x56101);
	return;
    }

    vlp = np->failurep;			/* head of linked list */
    if ( vlp EQ NO_VLINKP )
	return;				/* Possible if never Accepted */

    if ( np->link_count <= 0 )
    {
	BUGSTOP (-,np,0x56102);
	return;
    }

    if ( vlp EQ vlinkp )		/* vlinkp is at head of list */
    {
	vlp = vlinkp->failurep;
	np->failurep = vlp;

	if ( vlp EQ NO_VLINKP )		/* removing the only entry in list */
	{
	    np->rcvryto = 0;
	    np->hello_usec = 0;
	    np->sendhello.tv_sec = 0;	/* stop sending to nbr */
	    np->hellorcvd.tv_sec = 0;	/* forget we ever heard from nbr */
	    np->flags &= ~NbrFlgActive;	 /* no HELLOs needed if no streams */
	    Count (nbrsdeactivated,1);
	    uniqtime ( &( deactivated ) );
	    np->active_msec += MsecBetween ( &( deactivated ),
					     &( np->statechange ) );
	    np->statechange = deactivated;

	    IfLog (DBG_NBR_ACT)
	    {   unsigned char	*ucp = (unsigned char *) &( np->rmtIPAdr );
		printf ( "## Deactivating nbr %u.%u.%u.%u (%x)\n",
			 ucp[0],ucp[1],ucp[2],ucp[3], np);
	    }
	    BUGSTOP (,np->rmtIPAdr,0x56103);
	}
	else				/* compute new HELLO interval */
	{
	    np->rcvryto = vlp->rcvryto;
	    /* convert to usec and divide by Hello loss factor */
	    np->hello_usec =
		1000 * np->rcvryto / ConfigParm (hellolossfactor);

	    /* don't need to adjust sendhello */
	}

	IfLog (DBG_NBR_ACT)
	    printf ( "NbrRemove: (head) %x->%x, np %x, total %d\n",
		     vlinkp, np->failurep, np, np->link_count-1);
    }
    else
    {					/* find vlinkp in list */
	for (;; vlp = nvlp )
	{
	    nvlp = vlp->failurep;	/* next item in list, if any */
	    if ( nvlp EQ vlinkp )
	    {
		vlp->failurep = vlinkp->failurep;
		break;			/* from forever */
	    }

	    if ( nvlp EQ NO_VLINKP )
	    {
		BUGSTOP (,vlinkp,0x56104);
					/* Error recovery ??? */
		return;			/* not in list */
	    }

	} /* end of scan list forever loop */

	IfLog (DBG_NBR_ACT)
	    printf ( "NbrRemove: (Nth) %x->%x->%x, np %x, total %d\n",
		     vlp, vlinkp, vlinkp->failurep, np, np->link_count-1);
    }	

    np->link_count-- ;
    Count (nbrsremoved,1);

    vlinkp->on_failure_np = NO_NBRP;	/* mark not in any failure list */
    vlinkp->failurep = NO_VLINKP;

    return;
}

/*
S* NbrRTTUpdate ( np, refnum, tvp )
s*
s*	Compute the round trip time to neighbor ( = now - rtthello_sent )
s*	and then update the RTT history table, the average RTT, and the
s*	one-way delay (assumed to be RTT/2) to the neighbor.  If RTT >
s*	rtthello_ms, increase rtthello_ms.
 */

  static void				/* Next bugid 0x56200 */
NbrRTTUpdate( np, refnum, tvp )		/* from HandleAck */
  struct aNeighbor	*np;
  Reference		 refnum;
  struct timeval	*tvp;
{
    long		 rtt_ms;
    MSecs		*rtthp,
			 avgrtt;


    if ( refnum EQ np->rtthello_refnum )
    {					/* it's the one we're waiting for */
	np->rtthello_refnum = 0;	/* to ignore later duplicates */

	rtt_ms = MsecBetween (tvp,&(np->rtthello_sent)); /* time delta */

	IfLog (DBG_NBR_RTT)
	    printf ( "NbrRTTUpdate: now = %u,%u - rtthello_sent = %u,%u => rtt_ms %u\n",
		     tvp->tv_sec,tvp->tv_usec,np->rtthello_sent.tv_sec,
		     np->rtthello_sent.tv_usec,rtt_ms);

	if ( rtt_ms < 1 ) rtt_ms = 1;	/* don't confuse following code */
	/* ???wbe perhaps ignore values "way off" of average? */

	/* add rtt_ms to rtt_hist_ms and compute average. */
	if ( np->total_rtt_ms EQ 0 )	/* first HELLO ever from nbr */
	{
	    /* Method 1: keep a counter so that the first NBR_RTT_HISTORY-1
	       HELLO's RTTs can be averaged exactly.  Method 2 (used here):
	       the first one gets disproportionate weight, but eliminates
	       the almost useless counter and allows division by a constant
	       (which the compiler might optimize into a shift). */

	    rtthp = &(np->rtt_hist_ms[ NBR_RTT_HISTORY - 1 ]);
	    do  *rtthp = rtt_ms;  while ( --rtthp >= &(np->rtt_hist_ms[0]) );
	    np->total_rtt_ms = NBR_RTT_HISTORY * rtt_ms;
	    /* np->rtthistx = 0; */
	}
	else				/* HELLOs 2..N */
	{
	    rtthp = &(np->rtt_hist_ms[ np->rtthistx ]);
	    np->total_rtt_ms -= *rtthp;	/* remove old value from total */
	    np->total_rtt_ms += rtt_ms;	/* add new value to total */
	    *rtthp = rtt_ms;		/* save new value */
	    if ( ++np->rtthistx >= NBR_RTT_HISTORY )  np->rtthistx = 0;
	}

	avgrtt = (np->total_rtt_ms + NBR_RTT_HISTORY - 1) / NBR_RTT_HISTORY;
	np->avg_rtt_ms = avgrtt;

	if ( avgrtt >= 256*2000 )	/* >= 256*2 seconds in MSecs */
	    np->delay2nbr = ~0;		/* the largest unsigned number */
	else
	{
	    /* one way delay = (avgrtt/2) * 2**24(FracTime) / 1000(ms)
			     = (avgrtt << 13) * (1 + 3/125)  */
	    rtt_ms = avgrtt << 13;
	    np->delay2nbr = rtt_ms + (rtt_ms / 125) * 3;
	}

	/* keep rtthist_interval reasonably bigger than the average RTT
	   and not too small.  There are lots of ways to do this.  The
	   essential features of "+ConfigParm (hellobackoff)" are (1)
	   a minimum interval of 20 seconds and (2) x% of avgrtt, for
	   reasonable x and moderately large avgrtt, is ~20 seconds.
	   A fancier formula might use RTT variance. */

	rtt_ms = avgrtt + ConfigParm (hellobackoff); /* average RTT + 20 sec */
	if ( rtt_ms < np->rtthello_ms )
	    np->rtthello_ms = rtt_ms;	 /* reduce interval */
    }

    else				/* ACK for an old, forgotten HELLO */
    {
	/* RTT probably > rtthello_ms  and we've sent >= 1 RTT HELLOs
	   while waiting for this ACK.  If this is the case, increase
	   rtthello_ms (or implement overlapping RTT queries and
	   check history, and if history isn't deep enough, increase
	   rtthello_ms). */

	/* For each "ancient" ACK, increase rtthello_ms by one
	   basis amount.  This "automatically" puts off sending the next
	   RTT HELLO by the same amount.  Once rtthello_ms > RTT,
	   the code above will reduce rtthello_ms if appropriate. */

	np->rtthello_ms += ConfigParm (hellobackoff); /* 20 seconds */
    }

    return;
}

/*
S* NbrSort (vlinkp, new rcvryto value)	NOTYET
s*
s*	Equivalent to NbrRemove () followed by NbrInsert ().
s*	Sets vlinkp->rcvryto from the value passed, and adjusts
s*	np->interval and np->rcvryto if needed.
*/
#ifdef NOTYET

   static void				/* Next bugid 0x56300 */
NbrSort( vlinkp, recovery_timeout )	/* from */
  struct aVLink		*vlinkp;
  unsigned int		 recovery_timeout;
{
    /* somewhat inefficient, but not too much so */
    NbrRemove( vlinkp );
    (void) NbrInsert( vlinkp, recovery_timeout );

    return;
}
#endif NOTYET

/*
S* NextHopAdd ( ipsrcrutp, arcp, arganhp, anhlpp, pktdescpp )
s*
s*	Routine to add a new next-hop structure to a stream.
s*	Called ( p,    p, NULL, pp,   pp,   pp )  for streams,
s*	or     ( NULL, p, pp,   NULL, NULL, NULL) for neighbors.
s*
 */

  static struct NextHop *		/* Next bugid 0x56404 */
NextHopAdd( ipsrcrutp, arcp, arganhp, anhlpp, pktdescpp ) /* from st2_PCBUpdate */
  struct aSrcRut	 *ipsrcrutp;	/* Encoded IP Source Route, if any */
  struct aRouteChoice	 *arcp;		/* Routing choice for next hop */
  struct aNxtHop	 *arganhp;	/* aNxtHop to use, maybe NULL */
  struct aNxtHopList	**anhlpp;	/* Maybe NULL */
  struct aPktDesc	**pktdescpp;	/* Chain of resources, maybe NULL */
{
    struct NextHop	*nexthopp;
    struct aNxtHopList	*anhlp;
    struct aNxtHop	*anhp;
    int			 oldpri;


    /* Get a free next hop structure */

    oldpri = splnet ();
    {
	nexthopp = st2_nexthops.freep;
	if ( nexthopp NE NO_NEXTHOPP )
	    st2_nexthops.freep = nexthopp->nextp;
    }
    (void) splx ( oldpri );

    if ( nexthopp EQ NO_NEXTHOPP )
	goto abort0;

    Bzero (nexthopp,sizeof (struct NextHop));
    nexthopp->next_cnt = 1;
    SetState (NHP,nexthopp,nhop_state,NEW_HOP);

    IfLog (DBG_DTA_STR)
	printf ( "NextHopAdd at 0x%x for ST Agent 0x%x\n",
		 nexthopp, arcp->nxtSTIPAdr );

    /* ============================================================== */
    /* ??? this becomes conditional if multicast is to be/is being used */

    if ( (anhp = arganhp) EQ NO_NXTHOPP )
    {
	if (   (anhlpp EQ (struct aNxtHopList **) NULL)
	    OR (pktdescpp EQ (struct aPktDesc **) NULL) )
	    goto abort0;

	oldpri = splnet ();
	{
	    for (  anhlp = *anhlpp ; anhlp NE NO_NHLISTP
		 ; anhlpp = &( anhlp->nxtNxtHopp ), anhlp = *anhlpp )
	    {
		if ( anhlp->nxtfree >= anhlp->allocated )
		{
		    if ( anhlp->nxtNxtHopp NE NO_NHLISTP )
			continue;
		    break;		/* try to extend */
		}

		anhp = &( anhlp->NxtHops[ anhlp->nxtfree ] );
		anhlp->nxtfree++;	/* Already zeroed */
		break;			/* out of all array segments loop */

	    } /* end of for each aNxtHopList loop */

	    /* Did we did not find a free slot, extend list */
	    if ( anhp EQ NO_NXTHOPP )
	    {
#ifdef NOTYET
		/* ??? make sure not list head; how to find selfp */
		/* Try to extend the array in the last block */
		if ( *anhlpp EQ NO_NHLISTP )
		{
		    anhp = AddSpace (struct aNxtHop *, anhlpp, PKT_MD_PKT,
				     sizeof (struct aNxtHop));

		    if ( anhp NE NO_NXTHOPP )
		    {
			(*anhlpp)->allocated++;
			(*anhlpp)->nxtfree++;
			Bzero (anhp,sizeof (struct aNxtHop));
		    }
		} /* end of tried to extend */
#endif NOTYET

		/* Add another block to end of list, BUT NOT BEFORE ZEROED */
		if ( anhp EQ NO_NXTHOPP )
		{
		    struct aPktDesc	*pp = NO_PKTDSCP;

					/* new_nxthoplist */
		    anhlp = GetSpace (struct aNxtHopList *, &( pp ),
				       PKT_NXTHOP, PKT_MD_NEW, PKT_XA_PS,
				       0,0, sizeof (struct aNxtHopList), 0,0);

		    if ( anhlp EQ NO_NHLISTP )
		    {
			if ( pp NE NO_PKTDSCP )
			    FreePkts (pp);
#ifdef  STIIDEBUG
			(void) printf ( "aNxtHop dropped: out of space\n" );
#endif  STIIDEBUG
			(void) splx ( oldpri );
			BUGGOTO (-,1,0x56401,abort1);
		    }
		    /* In particular, clears NxtHopActive flags */	
		    Bzero (anhlp,sizeof (struct aNxtHopList));

		    /* Link into chain of aNxtHopLists, don't use selfp */
		    pp->nextp = *pktdescpp;
		    *pktdescpp = pp;

		    /* aNxtHopList struct has room for a # of nxthop slots. */

		    anhlp->allocated = DimensionOf (anhlp->NxtHops);
		    anhlp->objId = 0xf7; /*??? just picked something NE 0 */

		    anhp = &( anhlp->NxtHops[0] );    
		    anhlp->nxtfree = 1;

		    if ( *anhlpp NE NO_NHLISTP )
			BUGSTOP (-,Inconsistency,0x56402);

		    /* FOLLOWING LETS DRIVER-LEVEL ACCESS NEW BLOCK */
		    anhlp->nxtNxtHopp = *anhlpp;
		    *anhlpp = anhlp;

		} /* end of allocated a new aNxtHopList */
	    } /* end of wasn't a free NxtHop */
	} /* end of critical block */
	(void) splx ( oldpri );

	IfLog (DBG_MEM_ALO)
	    (void) printf ( "  nxthopsp 0x%x, used %d of %d\n",
			   anhlp, anhlp->nxtfree, anhlp->allocated );
    } /* end of get free aNxtHop entry */

    /* ============================================================== */
    /* Initialize aNxtHop ; similar code in NbrCreate */

    anhp->refcount = 1;
    anhp->gifcp = arcp->gifcp;
    anhp->fwdtoIPAdr = arcp->fwdtoIPAdr;
    {   struct sockaddr_in *sinp = (struct sockaddr_in * )
					&( anhp->lclsockaddr );
	sinp->sin_family = ConfigParm (coip_family);
	sinp->sin_addr.s_addr = arcp->fwdtoIPAdr;
    }
    anhp->lclhdrlen = arcp->gifcp->ifc_lclhdrlen;

    nexthopp->route.type = arcp->type;
    nexthopp->route.nxthopp = anhp;

    if ( arcp->type NE TO_APPL )
    {
	/* Build IP header if IP Encapsulation or IP Source Route */
	if (   (ipsrcrutp NE NO_SRP)
	    OR (arcp->IPencapsulate NE 0) )
	{
	    (void) IPEncapFwd( GATEWAY_VLINK, arcp->lclIPAdr, arcp->nxtIPAdr,
			       ipsrcrutp, anhp );
	}
    }

    /* ??? Could poke the ip layer with a dummy packet to trigger arp */

    if ( NOT RouteCheck( &( nexthopp->route ) ) )
	BUGGOTO (-,InvalidPointer,0x56403,abort1);

    return ( nexthopp );


 abort1:
    (void) NextHopDel( nexthopp );
    /* ??? error Inconsistency/Routing failure */

 abort0:
    return ( NO_NEXTHOPP );
}

/*
S* NextHopDel ( nexthopp )
s*
s*	Routine to delete a next-hop structure.  Caller is responsible for
s*	any associated state transisitions.
s*
 */

  static enum ST2Errors			/* Next bugid 0x56503 */
NextHopDel( nexthopp )			/* from NextHopAdd st2_PCBDel st2_PCBUpdate VLinkDel */
  struct NextHop	*nexthopp;
{
    int			 i = sizeof (struct aThruput) / sizeof (unsigned long),
			 oldpri;
    struct aNetIf	*ifst2p;
    struct aNxtHop	*anhp;
    struct aST2pcb	*cp;
    struct NextHop	*nh_entry;
    unsigned long	*in,
			*out;


    PTRCHECK (nexthopp, struct NextHop *,
	      BUGRETURN (-,InvalidPointer,0x56501,enum ST2Errors));

    cp = nexthopp->cp;
    kassert (cp NE NO_PCBP);

    SetState (NHP,nexthopp,nhop_state,DELETING_HOP);

    if ( (anhp = nexthopp->route.nxthopp) NE NO_NXTHOPP )
    {
	deactivate_aNxtHop(anhp);	/* Stops data forwarding */

	/* ---------------------------------------------------------------- */

	if ( nexthopp->route.type EQ TO_APPL )
	{
	    oldpri = splnet ();
	    {
		ifst2p = nexthopp->route.nxthopp->gifcp;

		/* ifst2p may be NULL if target has been closed/deleted */
		if ( ifst2p NE NO_NETIFP )
		{
		    if ( ifst2p->ifc_gwcp EQ cp ) 
			ifst2p->ifc_gwcp = NO_PCBP;
		    else
			BUGSTOP (-,Inconsistency,0x56502);

		    cp->refcnt--;
		    cp->refwho &= ~ RefGifcgwcpt;
		    IfLog (DBG_CNT_PCB)
			printf ( "refcnt: cp = %8x  refcnt = %2d  NextHopDel       RefGifcgwcpt\n",
				cp, cp->refcnt );
		}
	    }
	    (void) splx ( oldpri );
	}

	/* ??? currently, no resources reserved if TO_APPL */
	else
	{
	    (void) st2_RsrcRelease( nexthopp  /* multicast address list;
				    incremental */);
	}

	/* ---------------------------------------------------------------- */
	/* Accounting */

	(void) st2_RsrcAccount( anhp->gifcp, (unsigned long) 0 /*msec*/,
			        &( anhp->thruput ), &( cp->cost ) );

	out = (unsigned long *) &( cp->thruput );
	in  = (unsigned long *) &( anhp->thruput );
	i   = sizeof (struct aThruput) / sizeof (unsigned long);

	for ( ; i > 0 ; i-- ) { *out++ += *in; *in++ = 0; }

	nexthopp->route.nxthopp->gifcp = NO_NETIFP;
    }

					/* ??? what about a pending callback */
    nh_entry = cp->next_hopsp;
    if ( nh_entry EQ nexthopp )
    {
	cp->next_hopsp = nexthopp->nextp;
    }
    else for ( ; nh_entry NE NO_NEXTHOPP
	      ; nh_entry = nh_entry->nextp )
    {
	if ( nh_entry->nextp EQ nexthopp )
	{
	    nh_entry->nextp = nexthopp->nextp;
	    break;
	}
    }


    nexthopp->nextp = NO_NEXTHOPP;

    SetState (NHP,nexthopp,nhop_state,FREE_HOP);
    Bzero (nexthopp,sizeof (*nexthopp));
    nexthopp->nhop_state = FREE_HOP;

    oldpri = splnet ();
    {
	nexthopp->nextp = st2_nexthops.freep;
	st2_nexthops.freep = nexthopp;
    }
    (void) splx ( oldpri );

    IfLog (DBG_DTA_STR)
	printf ( "NextHopDel at 0x%x  cp 0x%x\n", nexthopp, cp );

    return ( NoError );
}

/*
S* PCBByName ( Namep )
s*
s*	Find another meander of the stream with the specified name.
s*
 */

  static struct aST2pcb *		/* Next bugid 0x56600 */
PCBByName( Namep )			/* from HandleStatus st2_DoConnect*2 */
  struct aName		*Namep;
{
    struct aST2pcb	*wcp;


    ForAllpcbs (wcp)
    {
	if ( Bcmp (&( wcp->Name ),Namep,sizeof (struct aName)) EQ 0 )
	    return ( wcp );

    } /* end of for all pcbs loop */

    return ( NO_PCBP );
}

/*
S* ReplyMsgSend ( request, reply )
s*
s*	Routine to send a response to a message received that does not
s*	have a valid virtual link.
s*
 */

  static enum ST2Errors			/* Next bugid 0x56703 */
ReplyMsgSend( request, reply )		/* from DoAcknowledge DoError */
					/* no valid vlinkp to use */
  struct aPrsSCM	*request,	/* caller keeps */
			*reply;		/* we keep */
{
    enum ST2Errors	 reason;
    octet4		 mandatory[ 1 ];
    struct sockaddr_in	*sip;

    /*??? We should have a "multicast" contbl entry for each
     *??? attached network.  Target address should be provided
     *??? individually with each message.
     */


    if ( request->gifcp EQ NO_NETIFP )
    {
	CtrlMsgDel( reply );		/* We dispose */
	BUGRETURN (-,InvalidPointer,0x56701,enum ST2Errors);
    }

    reply->hoscmphdr.RVLId = request->hoscmphdr.SVLId;
    reply->hoscmphdr.SVLId = request->hoscmphdr.RVLId;

    /* ??? this really the correct ipadr */
    sip = (struct sockaddr_in *) &( request->rcvingsa );
    reply->hoscmphdr.SndrIPAdr = sip->sin_addr.s_addr;
    reply->Namep = request->Namep;

    reply->pktp = NO_PKTDSCP;
    reply->pdulen = 0;

    mandatory[0] = st2_parmManByOp[ (int) reply->hoscmphdr.OpCode ];
    if (    (IsSet32 (mandatory, (int) STDetectorIPAdr ))
	AND (reply->detectoripadrp EQ (INETADDR) 0) )
    {
	/* ??? wrong address */
	reply->DetectorIPAdr = htonIP (reply->hoscmphdr.SndrIPAdr);
	reply->detectoripadrp = &( reply->DetectorIPAdr );
    }

    /* ??? encapsulation */
    reason = SCMPEncode( request->gifcp->ifc_mtu,
			  request->gifcp->ifc_lclhdrlen,
			  reply, &( reply->pktp ),
			  &( reply->pdulen ) );

    if ( reason NE NoError )
    {
#ifdef  STIIDEBUG
	(void) printf ( "Encode failed: reason code %d\n", reason );
#endif  STIIDEBUG
	CtrlMsgDel( reply );
	return ( reason );
    }

    /* ??? why isn't this cached per neighbor */
    sip = (struct sockaddr_in *) &( reply->rcvingsa );
    sip->sin_family = ConfigParm (coip_family);
    sip->sin_addr.s_addr = request->hoscmphdr.SndrIPAdr;

    if (   (request->gifcp->ifc_output EQ (int (*)()) NULL)
	OR (request->gifcp->ifc_output( request->gifcp->osifcp, reply->pktp,
			      &( reply->rcvingsa ) ) NE 0) )
    {
	reason = IntfcFailure;		/* ??? count xmit failures */
	BUGSTOP (,reason,0x56702);
    }
    reply->pktp = NO_PKTDSCP;

    CtrlMsgDel( reply );

    return ( reason );
}

/*
S* RouteCheck ( routep )
 */

  static boolean			/* Next bugid 0x56805 */
RouteCheck( routep )			/* from NextHopAdd st2_CMPOutput */
					/* INLINE/MACRO */
  struct Route	*routep;
{
    PTRCHECK (routep, struct Route *, return ( FALSE ));
    if ( routep->nxthopp EQ NO_NXTHOPP )
    {
	BUGSTOP (-,1,0x56801);
	return ( FALSE );
    }
    if ( routep->nxthopp->gifcp EQ NO_NETIFP )
    {
	BUGSTOP (,routep->nxthopp,0x56802);
	return ( FALSE );
    }
    if ( routep->nxthopp->gifcp->ifc_mtu <= 0 )
    {
	BUGSTOP (,routep->nxthopp->gifcp,0x56803);
	BUGSTOP (,routep->nxthopp->gifcp->ifc_name,0x56804);
	return ( FALSE );
    }

    return ( TRUE );
}

/*
S* SCMPDecode ( scmhdrp, scmlen, pscmpp, pscmlenp )
s 
s*	Routine to decode an SCMP PDU.
 *
** enum ST2Errors
** SCMPDecode ( scmhdrp, scmlen, pscmpp, pscmlenp )
**
** Input:
** struct aSCMPHeader	 *scmhdrp	Pointer to SCMPDU to be parsed.
** int			  scmlen	Valid length (from layer below).
** Out:
** struct aPktDesc	**pscmpp	Either address of pointer to area of
**					size *pscmlenp for resuting struct
**					aPrsSCM, if ** non-NULL, or location
**					to receive a pointer to aPktDesc
**					containing such a structure.
** int			 *pscmlenp	Size of area to be used when supplied
**					by caller; returns size of resulting
**					aPrsSCM structure.
** Result:
** InvalidArgument			Nothing was done due to invalid call
**					(Output parameters remain unchanged).
** <misc parsing errors>		Output parameters are valid, but parse
**					may not have been completed.
** NoError				Parse detected no errors.
**
 */
/* ??? cp for errors */

  static enum ST2Errors			/* Next bugid 0x56908 */
SCMPDecode( scmhdrp, scmlen, pscmpp, pscmlenp ) /* from st2_CMPInput */
  struct aSCMPHeader	 *scmhdrp;
  int			  scmlen;
  struct aPktDesc	**pscmpp;
  int			 *pscmlenp;
{
    boolean		 abortparse;
    int			 errorsdetected = 0;
    struct aPktDesc	*parsebufp;
    int			 targetsleft;
    octet4		*wp;
    struct aPrsSCM	*pscmp;
    octet4		 mandatory[ 1 ];
    octet4		 optional[ 1 ];


    if (   (scmhdrp EQ (struct aSCMPHeader *) NULL)
	OR (scmlen < sizeof (struct aSCMPHeader))
	OR (pscmpp EQ (struct aPktDesc **) NULL)
	OR (pscmlenp EQ (int *) NULL)
	OR (    (*pscmpp NE NO_PKTDSCP)
	    AND (*pscmlenp < sizeof (struct aPrsSCM))) )
    {
        if ( scmhdrp EQ (struct aSCMPHeader *) NULL )
	    BUGSTOP (,1,0x56901);
	if ( scmlen < sizeof (struct aSCMPHeader) )
	    BUGSTOP (,scmlen,0x56902);
	if ( pscmpp EQ (struct aPktDesc **) NULL )
	    BUGSTOP (,1,0x56903);
	if ( pscmlenp EQ (int *) NULL )
	    BUGSTOP (,1,0x56904);
	if (    (*pscmpp NE NO_PKTDSCP)
	    AND (*pscmlenp < sizeof (struct aPrsSCM)) )
	    BUGSTOP (,*pscmlenp,0x56905);

	/* ??? diagnostics */
	BUGRETURN (-,InvalidArgument,0x56906,enum ST2Errors);
    }


    if ( *pscmpp EQ NO_PKTDSCP )
    {
	parsebufp = NO_PKTDSCP;
	pscmp = GetSpace (struct aPrsSCM *, &( parsebufp ), PKT_PARSEDSCM,
			  PKT_MD_NEW, PKT_XA_PS,
			  0,0, sizeof (struct aPrsSCM), 0,0);
	if ( pscmp EQ NO_PSCMP )
	{
	    BUGRETURN (-,DropNoResrc,0x56907,enum ST2Errors);
	}
	Bzero (pscmp,sizeof (struct aPrsSCM));

	targetsleft = parsebufp->pkt_postpad / sizeof (struct aTargetDesc);
	*pscmlenp = parsebufp->pkt_size;

	*pscmpp = (struct aPktDesc *) parsebufp;
        pscmp->msgcnt = 1;
        pscmp->selfp = (struct aPktDesc *) parsebufp;
    }
    else
    {					/* Caller provided space */
	parsebufp = NO_PKTDSCP;
	pscmp = Mkp (struct aPrsSCM *, *pscmpp, 0);
	targetsleft =   (*pscmlenp - sizeof (struct aPrsSCM))
		      / sizeof (struct aTargetDesc);
	Bzero (pscmp,sizeof (struct aPrsSCM));
    }


#ifdef NetOrder
    Bcopy (scmhdrp,&( pscmp->hoscmphdr ),sizeof (struct aSCMPHeader));
#else !NetOrder
    {   struct aSCMPHeader	*hoscmhdrp = &( pscmp->hoscmphdr );
	hoscmhdrp->OpCode     = ntoh1 (scmhdrp->OpCode);
	hoscmhdrp->Options    = ntoh1 (scmhdrp->Options);
	hoscmhdrp->TotalBytes = ntoh2 (scmhdrp->TotalBytes);
	hoscmhdrp->RVLId      =        scmhdrp->RVLId;
	hoscmhdrp->SVLId      =        scmhdrp->SVLId;
	hoscmhdrp->Ref        =        scmhdrp->Ref;
	hoscmhdrp->LnkRef     =        scmhdrp->LnkRef;
	hoscmhdrp->SndrIPAdr  = ntohIP (scmhdrp->SndrIPAdr);
	hoscmhdrp->Checksum   = 0 /*   scmhdrp->Checksum */;
	hoscmhdrp->peropcode2 = ntoh2 (scmhdrp->peropcode2);

	mandatory[0] = mandatory[ (int) hoscmhdrp->OpCode ]; 
	if ( IsSet32 (mandatory, (int) STDetectorIPAdr) )
	    hoscmhdrp->peropcode4 = ntohIP (scmhdrp->peropcode4);
	else if ( IsSet32 (mandatory, (int) eHelloTimer) )
	    hoscmhdrp->peropcode4 = ntoh4 (scmhdrp->peropcode4);
	else
	    hoscmhdrp->peropcode4 = scmhdrp->peropcode4; /* SB: 0 */
    }
#endif NetOrder
    pscmp->scmhdrp = scmhdrp;
    pscmp->scmlen = scmlen;


    wp = Mkp (octet4 *,scmhdrp,sizeofaSCMPHeader);

    abortparse = FALSE;

    while ( wp < Mkp (octet4 *,scmhdrp,scmlen) )
    {
	register struct aParameter	*parmp = (struct aParameter *) wp;
	register struct aTarget		*aTargetp;
	register struct aTargetDesc	*tdp;
	int				 targetlistbytes,
					 targetlistcount,
					 targetbytes;
	enum ST2Errors			 errcod = NoError;


	if (    IsSet32 (pscmp->parmspresent, parmp->pcode)
	    AND (parmp->pcode NE (octet1) STpTargetList)
	    AND (parmp->pcode NE (octet1) STpHID) )
	{
	    errcod = DupParm;
	    goto tonxt;
	    /* ??? diag, error */
	}

	BitSet32 (pscmp->parmspresent, parmp->pcode);

	/* ??? verify min <= plen <= max */

	switch ( parmp->pcode )
	{
	case STpErrdPDU:			/* 1 */
	    pscmp->ErrdPDUp = (struct aErrdPDU *) wp;
	    break;


	case STpFlowSpec:			/* 2 */
	    pscmp->FlowSpecp = (struct aFlowSpec *) wp;
	    if ( ! IsSet32 (((octet4 *) &( ConfigParm (flowspec_versions) )),
			    pscmp->FlowSpecp->FlowVer) )
		errcod = FlowVerBad;
	    break;


	case STpFreeHIDs:			/* 3 */
	    pscmp->FreeHIDsp = (struct aFreeHIDs *) wp;
	    break;


	case STpGroup:			/* 4 */
	    pscmp->Groupp = (struct aGroup *) wp;
	    break;


	case STpHID:			/* 5 */
	    if ( pscmp->nhids >= MAX_HIDS_PER_STREAM )
		errcod = TooManyHIDs;
	    else if ( pscmp->nhids EQ 0 )
		pscmp->HIDp = (struct aHID *) wp;
	    else
		pscmp->altHIDp[ pscmp->nhids++ - 1 ] = (struct aHID *) wp;
	    break;


	case STpMcastAdr:			/* 6 */
	    pscmp->McastAdrp = (struct aMcastAdr *) wp;
	    break;


	case STpName:			/* 7 */
	    pscmp->Namep = (struct aName *) wp;
	    if ( pscmp->Namep->plen NE sizeofaName )
		errcod = NameUnknown;
		/* ??? wrong error */
	    break;


	case STpNxtHopIPAdr:		/* 8 */
	    pscmp->NxtHopIPAdrp = (struct aNxtHopIPAdr *) wp;
	    break;


	case STpOrigin:			/* 9 */
	    pscmp->Originp = (struct aOrigin *) wp;
	    break;


	case STpOrgTsmp:			/* 10 */
	    pscmp->OrgTsmpp = (struct aOrgTsmp *) wp;
	    break;


#ifdef NOTYET
|		case STpRecordRoute:		/* 11 */
|			/* ??? fill it in */
#endif NOTYET


	case STpRFlowSpec:		/* 12 */
	    pscmp->RFlowSpecp = (struct aFlowSpec *) wp;
	    if ( ! IsSet32 (((octet4 *) &( ConfigParm (flowspec_versions) )),
			    pscmp->RFlowSpecp->FlowVer) )
		errcod = FlowVerBad;
	    break;


	case STpRGroup:			/* 13 */
	    pscmp->RGroupp = (struct aGroup *) wp;
	    break;


	case STpRHID:			/* 14 */
	    pscmp->RHIDp = (struct aHID *) wp;
	    break;


	case STpRName:			/* 15 */
	    pscmp->RNamep = (struct aName *) wp;
	    if ( pscmp->RNamep->plen NE sizeofaName )
		errcod = NameUnknown;
		/* ??? wrong error */
	    break;


	case STpUserData:		/* 21 */
	    pscmp->UserDatap = (struct aUserData *) wp;
	    break;


	case STpTargetList:		/* 20 */

					/* ??? 4 */
	    targetlistbytes = ntoh1 (parmp->plen) - 4;
	    targetlistcount = ntoh2 (parmp->pdata2);
	    aTargetp = (struct aTarget *) &( parmp->pdata4 );

	    while (    (targetsleft > 0) /* ??? exparray */
		   AND (targetlistcount > 0)
		   AND (   targetlistbytes
			>= (targetbytes = ntoh1 (aTargetp->TargetBytes))) )
	    {
		if ( (targetbytes & 0x3) NE 0 )
		{
		    Count (lennot4x_target,1); /* count bad parameter lengths */
		    errcod = BadParmLen;
		    abortparse = TRUE;
		    break;		/* out of while loop */
		}

		tdp = &( pscmp->targlst.TargetDescs[ pscmp->targlst.nxtfree ] );

		tdp->targetp = NO_TRGP;
		tdp->Targetp = aTargetp;
#ifdef NOTYET
|			/* validate SAPs, source routes */
|			switch ( x )
|			{
|			/* ??? fill them in */
|			case STpIPLSrcRut:	/* 16 */
|			case STpIPSSrcRut:	/* 17 */
|			case STpSTLSrcRut:	/* 18 */
|			case STpSTSSrcRut:	/* 19 */
|			default:
|			}
#endif NOTYET

		++pscmp->targlst.nxtfree;
		--targetsleft;
  		--targetlistcount;
		targetlistbytes -= targetbytes;
		aTargetp = Mkp (struct aTarget *,aTargetp,targetbytes);

	    } /* end of while targetsin targetlist */

	    /* ??? check for leftover targetlistbytes */
	    break;


	default:
	    pscmp->unknownparms++;
	    BitSet32 (pscmp->errors, (int) PCodeUnknown);
	    /* ??? warning message */
	    /* ??? more extensible */

	} /* end of switch pcode */

tonxt:

	if ( errcod NE NoError )
	{
	    BitSet32 (pscmp->errors, (int) errcod);
	    errorsdetected++;
	}

	/* ??? move to top of loop */
	if ( (parmp->plen & 3) NE 0 )
	{
	    Count (lennot4x_parm,1);	/* count bad parameter lengths */
	    BitSet32 (pscmp->errors, (int) BadParmLen);
	    errorsdetected++;
	    break;
	}

	if ( abortparse )
	    break;

	wp = Mkp (octet4 *,wp,parmp->plen);

    } /* end of while parameters loop */


    /* Return actual length */

    *pscmlenp =
	(char *) &( pscmp->targlst.TargetDescs[ pscmp->targlst.nxtfree + 1 ] )
	- (char *) pscmp;


    /* More validity checking ... */

    mandatory[0] = st2_parmManByOp[ (int) pscmp->hoscmphdr.OpCode ];

/* The hid should be decoded if the element is mandatory *or* if it
 * is optional and the HBit is set.  Note that for requests where the
 * hid element is mandatory the setting HBit is not relevant.  And,
 * for all requests where the hid element is optional, the HBit indicates
 * whether it has been specified.  So sayeth the spec.
 */
    optional[0] = st2_parmOptByOp[ (int) pscmp->hoscmphdr.OpCode ];

    if (    IsSet32 (mandatory, (int) eHID)
	OR (    IsSet32 (optional, (int) eHID)
	    AND OptSet (scmhdrp, HBit) ) )
    {
	pscmp->hdrhidp = &( scmhdrp->peropcode2 );
	BitSet32 (pscmp->parmspresent, (int)eHID);
    }
    else if ( IsSet32 (mandatory, (int) STReasonCode) )
    {
	pscmp->reasoncodep = &( scmhdrp->peropcode2 );
	BitSet32 (pscmp->parmspresent, (int) STReasonCode);
    }

    if ( IsSet32 (mandatory, (int) eHelloTimer) )
    {
	pscmp->hellotimerp = &( pscmp->hoscmphdr.peropcode4 );
	BitSet32 (pscmp->parmspresent, (int) eHelloTimer);
    }
    else if ( IsSet32 (mandatory, (int) STDetectorIPAdr) )
    {
	pscmp->detectoripadrp = &( pscmp->hoscmphdr.peropcode4 );
	BitSet32 (pscmp->parmspresent, (int) STDetectorIPAdr);
    }

    if ( (pscmp->parmspresent[0] & mandatory[0]) NE mandatory[0] )
    {
	Count (parm_missing,1);		/* count missing parameters */
	BitSet32 (pscmp->errors, (int) ParmMissing);
	errorsdetected++;
    }


    return ( (errorsdetected EQ 0) ? NoError : ProtocolError );

} /* end of SCMPDecode */


/*
S* SCMPEncode ( mtu, xtrahdrlen, pscmp, pdupp, pdulenp )
s*
s*	Routine to encode an SCMP PDU.
s*
 */
/*ARGSUSED*/
#define MAX_TARGET_LISTS 10

  static enum ST2Errors			/* Next bugid 0x57022 */
SCMPEncode( mtu, xtrahdrlen, pscmp, pdupp, pdulenp ) /* from CtrlMsgRelay CtrlMsgSend ReplyMsgSend */
  int			  mtu,		/* MTU for packet */
			  xtrahdrlen;	/* Reserve for local net headers */
  struct aPrsSCM	 *pscmp;	/* What to encode */
  struct aPktDesc	**pdupp;	/* Set to ptr to aPktDesc of pdu */
  int			 *pdulenp;	/* Set to space used ??? OBS */
{
    enum OpCodes	  opcode;
    enum ST2Errors	  retcod = NoError;
    int			  datalen,
			  errorsdetected = 0,
			  i,
			  maxfraglen,
			  parmidx,	/* Count (FirstParm==1) for error msg */
			  pdunum = 0,	/* Nth PDU generated */
			  plen,
			  scmlen,	/* Length of SCM PDU */
			  targidx,	/* Iterate through targets */
			  targstogo;	/* # targets not yet encoded */
    octet1		 *basewp,	/* Ptr to ST2Hdr */
			 *datap,
			 *wp;		/* Ptr to next free octet */
    octet4		  mandatory[ 1 ], /* Bitmask of required parameters */
			  parmpresent[ 1 ]; /* Bitmask of actual parameters */
    struct aParameter	 *parmp,	/* Ptr to parameter being processed */
			**parmpp;	/* Work through FirstParm -> LastParm */
    struct aPktDesc	 *firstpktp,	/* Ptr to first PDU for aPrsSCM */
			 *lastpktp,	/* Ptr to last  PDU for aPrsSCM */
			 *pktp;		/* Ptr to PDU being built */
    struct aSCMPHeader	 *scmhdrp;	/* Ptr to SCMP Header */
    struct aST2Header	 *st2hdrp;	/* Ptr to ST2 Header */
    struct aTargetDesc	 *tdp;		/* Ptr to current target info */



    if (   (pscmp EQ NO_PSCMP)
	OR (pdupp EQ (struct aPktDesc **) NULL)
#ifndef STILL
	OR (*pdupp NE NO_PKTDSCP)	/* ??? How to handle fragments */
#endif STILL
	OR (pdulenp EQ (int *) NULL) )
	BUGRETURN (-,InvalidArgument,0x57001,enum ST2Errors);


    parmpresent[0] = 0;
    for ( i = DimensionOf (pscmp->errors) ; i > 0 ; )	pscmp->errors[--i] = 0;

    /* ------------------------------------------------------------ */
    /* Validate opcode */

    opcode = (enum OpCodes) pscmp->hoscmphdr.OpCode;
    if (   ((int) opcode <= (int) OpBadOpCode)
	OR ((int) ConfigParm (num_st_opcodes) <= (int) opcode) )
    {
	Count (encodewrongparms,1);	/* count encode errors */
	retcod = OpCodeUnknown;
	BitSet32 (pscmp->errors, (int) retcod);
	errorsdetected++;

	BUGSTOP (-,opcode,0x57002);
    }

    pscmp->hoscmphdr.OpCode = (octet1) opcode;

    /* ------------------------------------------------------------ */
    /* Check required parameters */

    mandatory[0] = st2_parmManByOp[ (int) opcode ];

    if ( pscmp->hdrhidp NE (octet2 *) NULL )
    {
	/* HBit should be set only if an optional hid is being provided.
	 * If hid is required, the bit is never set. */
        if ( !IsSet32 (mandatory, (int) eHID) )
	    pscmp->hoscmphdr.Options |= HBit;
	pscmp->hoscmphdr.peropcode2 = *(pscmp->hdrhidp);
	BitSet32 (parmpresent, (int) eHID);

	if ( pscmp->reasoncodep NE (octet2 *) NULL )
	{
	    Count (encodewrongparms,1);	/* count encode errors */
	    retcod = HIDandReasonCode;
	    BitSet32 (pscmp->errors, (int) retcod);
	    errorsdetected++;
	    BUGSTOP (-,retcod,0x57003);
	}
    }
    else if ( pscmp->reasoncodep NE (octet2 *) NULL )
    {
	pscmp->hoscmphdr.peropcode2 = (octet2)
		FormalReasonFor ((enum ST2Errors) *(pscmp->reasoncodep));
	BitSet32 (parmpresent, (int) STReasonCode);
    }

    if ( pscmp->detectoripadrp NE (octet4 *) NULL )
    {
	pscmp->hoscmphdr.peropcode4 = *(pscmp->detectoripadrp);
	BitSet32 (parmpresent, (int) STDetectorIPAdr);

	if ( pscmp->hellotimerp NE (octet4 *) NULL )
	{
	    Count (encodewrongparms,1);	/* count encode errors */
	    retcod = HelloTimerandDetectorIPAdr;
	    BitSet32 (pscmp->errors, (int) retcod);
	    errorsdetected++;
	    BUGSTOP (-,retcod,0x57004);
	}
    }
    else if ( pscmp->hellotimerp NE (octet4 *) NULL )
    {
	pscmp->hoscmphdr.peropcode4 = *(pscmp->hellotimerp);
	BitSet32 (parmpresent, (int) eHelloTimer);
    }

    /* ------------------------------------------------------------ */
    /* Compute length of & basic validity check regular parameters */

    scmlen = sizeofaSCMPHeader;		/* Fixed portion */

    parmpp = (struct aParameter **) &( pscmp->FirstParm );
    parmidx = 0;
    for (  ; parmpp <= (struct aParameter **) &( pscmp->LastParm ) ; ++parmpp )
    {
	parmidx++;			/* For error messages */

	if ( (parmp = *parmpp) EQ (struct aParameter *) NULL ) 
	    continue;

	if (   (parmp->pcode EQ (octet1) 0)
	    OR ((int) parmp->pcode >= (int) ConfigParm (num_st_parms))
	    OR ((parmp->plen & 3) NE 0)
	    OR (parmp->plen < parmlens[ parmp->pcode ].min)
	    OR (parmlens[ parmp->pcode ].max < parmp->plen) )
	{
	    Count (encodebadparm,1);	/* count encode errors */
	    retcod = BadParmLen;
	    BitSet32 (pscmp->errors, (int) retcod);
	    errorsdetected++;

	    BUGSTOP (-,parmidx,0x57005);
	    if ( parmp->pcode EQ (octet1) 0 )
		BUGSTOP (-,1,0x57006);
	    if ( (int) parmp->pcode >= (int) ConfigParm (num_st_parms) )
		BUGSTOP (-,parmp->pcode,0x57007);
	    if ( (parmp->plen & 3) NE 0 )
		BUGSTOP (-,parmp->plen,0x57008);
	    if ( parmp->plen < parmlens[ parmp->pcode ].min )
		BUGSTOP (-,parmp->plen,0x57009);
	    if ( parmlens[ parmp->pcode ].max < parmp->plen )
		BUGSTOP (-,parmp->plen,0x57010);
	}
	else
	{
	    BitSet32 (parmpresent, parmp->pcode);
	    scmlen += parmp->plen;
	}

    } /* end of for all regular parameters loop */

    /* ------------------------------------------------------------ */
    /* Check if required parameters present */

    if ( pscmp->targlst.nxtfree > 0 )
	BitSet32 (parmpresent, STpTargetList);

    if ( (parmpresent[0] & mandatory[ 0 ]) NE mandatory[ 0 ])
    {
	Count (encodewrongparms,1);	/* count encode errors */
	retcod = ParmMissing;
	BitSet32 (pscmp->errors, (int) retcod);
	errorsdetected++;

	BUGSTOP (,opcode,0x57011);
	BUGSTOP (,st2_parmManByOp[ (int) opcode ],0x57012);
	BUGSTOP (,parmpresent[0],0x57013);
    }

    if ( errorsdetected > 0 )
    {
	BUGRETURN (-,retcod,0x57014,enum ST2Errors);
    }

    /* ------------------------------------------------------------ */
    /* Find space in a PDU available for target (s) */

    maxfraglen = mtu;

    /* If the mtu has not yet been discovered, use a guess. */
    if ( maxfraglen EQ 0 )
    {
	maxfraglen = 512+24;		/* Internet default. */
	BUGSTOP (-,MTUtosmall,0x57015);
    }

    if ( maxfraglen > ConfigParm (api_MTU) )
	maxfraglen = ConfigParm (api_MTU); /* ??? MCLBYTES */

    if ( pscmp->nxthopp NE NO_NXTHOPP )
    {
	xtrahdrlen += pscmp->nxthopp->lclhdrlen + pscmp->nxthopp->encaplen;
	maxfraglen -= pscmp->nxthopp->encaplen;
    }
/*  else ... ???
 */

    maxfraglen -= 16 /*align*/ + xtrahdrlen + sizeofaST2Header + scmlen;

    if ( maxfraglen < minsizeofaTarget )
    {
	Count (misconfiguration,1);	/* count configuration errors */
	retcod = MTUtosmall;
	BitSet32 (pscmp->errors, (int) retcod);
	errorsdetected++;

	BUGSTOP (,(scmlen + minsizeofaTarget),0x57016);
	BUGSTOP (,mtu,0x57017);
	BUGSTOP (,maxfraglen,0x57018);
	BUGRETURN (-,retcod,0x57019,enum ST2Errors);
    }

    /* ------------------------------------------------------------ */
    /* Append new PDU (s) to end of possibly preexisting list */

    firstpktp = Mkp (struct aPktDesc *, *pdupp, 0);
    lastpktp = firstpktp;
    if ( lastpktp NE NO_PKTDSCP )
    {
	while ( lastpktp->pkt_nxtmsgp NE NO_PKTDSCP )
	    lastpktp = lastpktp->pkt_nxtmsgp;
    }

    pdunum = 0;				/* Nth PDU generated */
    targstogo = pscmp->targlst.nxtfree;	/* # targets to be encoded */

    do
    {
	pdunum++;			/* Next PDU to be generated */

	/* Must get memory for packet */
	pktp = NO_PKTDSCP;
	basewp = GetSpace (octet1 *, &( pktp ), PKT_DATA, PKT_MD_NEW,
			    PKT_XA_PS, 0, xtrahdrlen + 8,
			    sizeofaST2Header + scmlen,
			    maxfraglen,0);
	if ( basewp EQ (octet1 *) NULL )
	{
	    retcod = DropNoResrc;
	    BUGSTOP (,retcod,0x57020);
	    break;			/* out of do all PDUs loop */
	}

	Bzero (basewp,pktp->pkt_length + pktp->pkt_postpad);

	/* Either first packet, or link this to last */
	if ( lastpktp EQ NO_PKTDSCP )
	    firstpktp = pktp;
	else
	    lastpktp->pkt_nxtmsgp = pktp;
	lastpktp = pktp;

	wp = basewp;			/* Start of ST Header */

	/* ------------------------------------------------------------ */
	/* Build ST-II and SCMP headers */

	st2hdrp = Mkp (struct aST2Header *, wp, 0); /* ST-II header */

	st2hdrp->IPSTVer  = hton1 (STV2);
	/* Ctl have priority ??? */
	st2hdrp->STPriOpt = hton1 (pscmp->st2options);
	st2hdrp->HID      = HID_CONTROL;
	st2hdrp->STChecksum = 0;
	
	wp += sizeofaST2Header;
	scmhdrp = Mkp (struct aSCMPHeader *, wp, 0); /* SCMP header */
	pscmp->scmhdrp = scmhdrp;	/* ??? for printout, multiple */

	scmhdrp->OpCode     = hton1 (pscmp->hoscmphdr.OpCode);
	scmhdrp->Options    = hton1 (pscmp->hoscmphdr.Options);
	scmhdrp->RVLId      =        pscmp->hoscmphdr.RVLId;
	scmhdrp->SVLId      =        pscmp->hoscmphdr.SVLId;
	scmhdrp->Ref        =        pscmp->hoscmphdr.Ref;
	scmhdrp->LnkRef     =        pscmp->hoscmphdr.LnkRef;
	scmhdrp->SndrIPAdr  = htonIP (pscmp->hoscmphdr.SndrIPAdr);
	scmhdrp->Checksum   = 0;
	scmhdrp->peropcode2 = hton2 (pscmp->hoscmphdr.peropcode2);
	scmhdrp->peropcode4 = hton4 (pscmp->hoscmphdr.peropcode4);

	wp += sizeofaSCMPHeader;

	/* ------------------------------------------------------------ */
	/* Insert all parameters except target list, in "any order" */

	for ( parmpp = (struct aParameter **) &( pscmp->FirstParm )
	     ; parmpp <= (struct aParameter **) &( pscmp->LastParm )
	     ; ++parmpp )
	{
	    if ( (parmp = *parmpp) EQ (struct aParameter *) NULL ) 
		continue;

	    plen = parmp->plen;
	    Bcopy (parmp,wp,plen);
	    wp += plen;

	} /* end of for all regular parameters loop */

	/* ------------------------------------------------------------ */
	/* Insert a TargetList (if needed), only one per SCM PDU for now */

	if ( pscmp->targlst.nxtfree > 0 )
	{
	    struct aTargetList	*tlp = Mkp (struct aTargetList *, wp, 0);


	    datalen = maxfraglen;	/* Limit based on MTU */
	    if ( datalen > pktp->pkt_postpad )
		datalen = pktp->pkt_postpad; /* Limit to what will fit */

	    tlp->pcode = (octet1) STpTargetList;
	    tlp->TargetCount = 0;
	    wp += sizeofaTargetListHeader;
	    datalen -= sizeofaTargetListHeader;
	    AddPktSpace (pktp,sizeofaTargetListHeader);

	    tdp = &( pscmp->targlst.TargetDescs[ 0 ] );
	    targidx = 0;

	    for ( ; targidx < pscmp->targlst.nxtfree ; targidx++, tdp++ )
	    {
		if ( tdp->pktnum != 0 )
		    continue;		/* Already been processed */

		plen = ntoh1 (tdp->Targetp->TargetBytes);
		if ( plen > datalen )
		    continue;		/* Won't fit */

		datalen -= plen;
		targstogo--;		/* Either encode or error */

		Bcopy (tdp->Targetp,wp,plen);

		tdp->pktnum = pdunum;
		++tlp->TargetCount;
		wp += plen;
		datalen -= plen;
		AddPktSpace (pktp,plen);

	    } /* end of all targets loop */

	    tlp->plen = wp - (octet1 *) tlp;

	    if ( tlp->TargetCount == 0 )
	    {
		wp = (octet1 *) tlp;
		/* unlink & free lastpktp; break; ??? flush packet */
	    }
	    else
		tlp->TargetCount = hton2 (tlp->TargetCount);

	} /* end of have targets */

	/* ------------------------------------------------------------ */
	/* Finish headers: lengths and checksums */

	datap = (octet1 *) scmhdrp;
	datalen = wp - datap;
	scmhdrp->TotalBytes = hton2 (datalen);
	st2hdrp->TotalBytes = hton2 (sizeofaST2Header + datalen);

	{   unsigned long	 cksum = 0;
	    octet2		*cksump = (octet2 *) scmhdrp;
	    int			 nc;


	    for ( nc = datalen / 2 ; nc > 0 ; --nc )
		cksum += *cksump++;

	    if ( (datalen & 1) NE 0 )	/* should be a multiple of 4 !! */
		cksum += *cksump++ & hton2 (0xFF00);

	    cksum = (cksum >> 16) + (cksum & (octet2) 0xFFFF);
	    cksum = (cksum >> 16) + (cksum & (octet2) 0xFFFF);
	    scmhdrp->Checksum = ~ cksum;

	    /* IP/ST-II Versions and total bytes */
	    cksum = * Mkp (unsigned long *, st2hdrp, 0);
	    cksum += st2hdrp->HID;

	    cksum = (cksum >> 16) + (cksum & (octet2) 0xFFFF);
	    cksum = (cksum >> 16) + (cksum & (octet2) 0xFFFF);
	    st2hdrp->STChecksum = ~ cksum;
	}

	/* ------------------------------------------------------------ */
	/* Insert IP Encapsulation, if required */

	if (    (pscmp->nxthopp NE NO_NXTHOPP)
	    AND ((i = pscmp->nxthopp->encaplen) > 0) )
	{
	    struct aNxtHop	*anhp = pscmp->nxthopp;
	    struct ip		*ipp;
	    unsigned long	 cksum;


	    ipp = AddSpace (struct ip *, &( pktp ), PKT_MD_HDR, i);
	    if ( ipp EQ (struct ip *) NULL )
	    {
		BUGRETURN (-,Inconsistency,0x57021,enum ST2Errors);
	    }
	    
	    Bcopy (&(anhp->lclsockaddr.sa_data[ anhp->lclhdrlen ]),ipp,i);

	    /* if IP encapsulation, must add ST length to IP total length
	       and update IP checksum (computed with (IP total length = 0))
	       NB: nxthopp->encaplen EQ IP total length before adding ST */

	    ipp->ip_len = hton2 (ipp->ip_len + ntoh2 (st2hdrp->TotalBytes));

	    cksum = ipp->ip_sum + ONES_NEGATE (ipp->ip_len);
	    ipp->ip_sum = (cksum >> 16) + (cksum & (octet2) 0xFFFF);
	}

	*pdulenp += pktp->pkt_length;

    } while ( targstogo > 0 );		/* end of do loop */

    /* ------------------------------------------------------------ */
    /* Return info to caller */

    *pdupp = (struct aPktDesc *) firstpktp;

    return ( retcod );

} /* end of SCMPEncode */


/*
S* st2_ApplEntityMatch ( entp, originp )
s*
s*	Routine to match Application Entity Addresses - NextPcol,
s*	SAP length & SAP data.
s*
 *	Points for exactness are as follows:
 *		exact match	wild card
 *	NextPcol    +1		   N/A
 *	IPAdr	    +4		   +1
 *	SAP	    +8		   +1
 */


  int					/* Next bugid 0x57100 */
st2_ApplEntityMatch( entp, originp )	/* from api_control st2_DoConnect */
  struct aApplEntity	*entp;
  struct aOrigin	*originp;	/* really an soname.applent */
{
    int			 exactness = 0;


    /* Protocols must match. */
    if ( entp->NextPcol NE originp->NextPcol )
	return ( 0 );
    exactness++;

    if ( (entp->SAPBytes EQ 0) OR (originp->SAPBytes EQ 0) )
	exactness++;
    else if (     (entp->SAPBytes EQ originp->SAPBytes)
	     AND (Bcmp (&( entp->SAP[0] ),&( originp->SAP[0] ),
			entp->SAPBytes) EQ 0) )
	exactness += 8;
    else
	return ( 0 );

    if ( entp->IPAdr EQ originp->IPAdr )
	exactness += 4;
    else if ( (entp->IPAdr EQ INADDR_ANY) OR (originp->IPAdr EQ INADDR_ANY) )
	exactness++;
    else
	return ( 0 );

    return ( exactness );
}

/*
 * st2_ClkFast ()
 */
#ifdef NOTYET
|	
|	  int
|	st2_ClkFast()				/* from <nowhere/vector> */
|	{
|		Count (fasttics,1);
|						/* ??? what priority */
|		return ( 0 );
|	
|	} /* end of st2_ClkFast */
#endif NOTYET

/*
S* st2_ClkSlow ()
 */

static int		 sweepcount;
static struct aST2pcb	*sweepptr;


  int					/* Next bugid 0x57200 */
st2_ClkSlow()				/* from */
{
    int			 i,
			 oldpri,
			 updatetimer;
    long		 ms_last_heard;
    struct aNeighbor	*np;
    struct aNeighborList*nlp;
    struct aPrsSCM	*scan,
			*php;
    struct timeval	 tv;


    Count (slowtics,1);
    uniqtime ( &tv );			/* What time it is */

    /* ??? where to put this; st2_init is too early */
    if ( local_inetaddr EQ INADDR_ANY )
    {	boolean			 go = TRUE;
	extern struct ifnet	*ifnet;	/* BSD's global list of interfaces */
	struct ifnet		*wifnetp;
	struct ifaddr		*wifaddrp;
	struct sockaddr_in	*sinp;


	for ( wifnetp = ifnet ; go AND (wifnetp NE (struct ifnet *) NULL)
	     ; wifnetp = wifnetp->if_next )
	{
	    wifaddrp = wifnetp->if_addrlist;
	    for ( ; go AND (wifaddrp NE (struct ifaddr *) NULL)
		 ; wifaddrp = wifaddrp->ifa_next )
	    {
		if ( wifaddrp->ifa_addr.sa_family EQ AF_INET )
		{
		    sinp = (struct sockaddr_in *) &( wifaddrp->ifa_addr );
		    if ( sinp->sin_addr.s_addr EQ 0x7F000001 )
			continue;
		    local_inetaddr = sinp->sin_addr.s_addr;
		    go = FALSE;	/* Found an address */
		}
	    }
	}

	st2_stats.agent = local_inetaddr;
    }

    if (    ((st2_stats.butSeconds & 0xFFFF0000) EQ 0)
	AND ((tv.tv_sec & 0xFFFF0000) NE 0) )
    {
	st2_stats.butSeconds = tv.tv_sec;
	st2_stats.butFract   = tv.tv_usec;
					/* Start HelloTimer */
	st2_stats.hellomsec  = 0;
	st2_stats.hellosec   = tv.tv_sec;
	st2_stats.hellousec  = tv.tv_usec;

	/* ??? randomize st2_stats.sap_next = 0xC000 + 1024; */
	/* ??? randomize VLID */
    }

    st2_stats.rptSeconds = tv.tv_sec;
    st2_stats.rptFract   = tv.tv_usec;

    /* ============================================================== */
    /* Send HELLOs & detect none received */

    if ( NOT ConfigFlag (ST2FlgNoHello) )
    {
	updatetimer = 1;
	nlp = &( st2_neighbors );
	for ( ; nlp NE NO_NBRLSTP ; nlp = nlp->nxtNeighborp )
	{
	    np = &( nlp->Neighbors[0] );
	    i = nlp->nxtfree;
	    for ( ; i > 0 ; i--, np++ )
	    {
		if (   ((np->flags & NbrFlgDelete) NE 0)
		    OR ((np->flags & NbrFlgActive) EQ 0) )
		    continue;

		/* Check if time to send neighbor a HELLO */
		if (    (np->sendhello.tv_sec NE 0) /* Enabled */
		    AND (        (tv.tv_sec  > np->sendhello.tv_sec)
			 OR (    (tv.tv_sec EQ np->sendhello.tv_sec)
			     AND (tv.tv_usec > np->sendhello.tv_usec))) )
		{
		    /* Update our msec up aka HelloTimer */
		    if ( updatetimer )
		    {
			updatetimer = 0; /* doing it */
			st2_stats.hellomsec = MsecBetween (&(tv),
				(struct timeval *) &(st2_stats.hellosec));
		    }

		    /* Send HELLO to neighbor */
		    DoHello( np, &( tv ) );

		    np->sendhello.tv_usec += np->hello_usec;
		    if ( np->sendhello.tv_usec > 1000000 )
		    {
			np->sendhello.tv_sec += np->sendhello.tv_usec / 1000000;
			np->sendhello.tv_usec %= 1000000;
		    }
		} /* end of send hello */

		/* Check if time to declare neighbor failure */
		ms_last_heard = MsecBetween (&(tv),&(np->hellorcvd));

		IfLog (DBG_NBR_RTT)
		    printf ( "hello interval = %u, np->rcvryto = %u\n",
			     ms_last_heard, np->rcvryto);

		if (    (ms_last_heard > np->rcvryto)
		    AND (np->totheard NE 0)) /* ??? backward compatibility */
		{			/* (ignore wrap to zero) */
		    np->totfailed++;
		    NbrFailed( np, NetworkFailure /*STAgentFailure*/ );
		}
	    } /* end of for all neighbors in list loop */
	} /* end of for all neighbor lists loop */
    } /* HELLOs not disabled */

    /* ============================================================== */
    /* st2_process_retries */
					/* ??? what priority */

    php = NO_PSCMP;
    scan = st2_retry_q.nextp;

    oldpri = splnet ();
    {
	for ( ; scan NE EmptyRetryQ ; scan = scan->nextp )
	{
	    if ( tv.tv_sec < scan->retry_tod )
		break;
	    php = st2_retry_q.nextp;

	} /* end of for loop */

	if ( php NE NO_PSCMP )
	{
	    php->prevp->nextp = scan;	/* i.e., st2_retry_q.nextp = scan; */
	    scan->prevp->nextp = NO_PSCMP;
	    scan->prevp = php->prevp;
	}
    }
    (void) splx ( oldpri );

    /* send them on at lower priority */

    while ( php NE NO_PSCMP )
    {
	scan = php->nextp;
	php->nextp = NO_PSCMP;
	php->prevp = NO_PSCMP;

	IfLog (DBG_CTL_MSG)
	    (void) printf ( "Retry #%d of opn %d scm message (%x)\n",
			    php->tries, php->hoscmphdr.OpCode, php );

	(void) st2_CMPOutput( php, NO_VLINKP );
	php = scan;

    } /* end of while loop */

    /* ============================================================== */


    if ( sweepcount-- <= 0 )
    {
	sweepcount = ConfigParm (sweep_interval);
	if ( sweepcount <= 0 )
	    sweepcount = 10;
	if ( sweepptr EQ NO_PCBP )
	    sweepptr = &( st2_ConTbl[0] );
	if ( ++sweepptr >= &( st2_ConTbl[ ConfigParm (max_streams) ] ) )
	    sweepptr = &( st2_ConTbl[1] );

	if ( IsState (PCB,sweepptr,pcb_state,CINS_free) )
	{
	    sweepptr->needupdate++;
	    softcall ( (int (*) ()) st2_PCBUpdate, (int) sweepptr );
	}
    }

#ifdef NOTYET
|	    struct aST2pcb	*toacctp;
|	    oldpri = splnet ();
|	    {
|		toacctp = st2_acctp;
|		st2_acctp = NO_PCBP;
|	    }
|	    (void) splx ( oldpri );
|	
|	    if ( toacctp NE NO_PCBP )
|		toacctp = st2_accounting( toacctp );
|	
|	    if ( toacctp NE NO_PCBP )
|	    {
|		oldpri = splnet ();
|		{
|		    if ( st2_ConTblTailp EQ NO_PCBP )
|			st2_ConTblFreep = toacctp;
|		    else
|			st2_ConTblTailp->nextp = toacctp;
|
|		    while ( toacctp->nextp NE NO_PCBP )
|			toacctp = toacctp->nextp;
|
|		    st2_ConTblTailp = toacctp;
|		}
|		(void) splx ( oldpri );
|	    }
#endif NOTYET

    return ( 0 );

} /* end of st2_ClkSlow */

/*
S* st2_CMPInput ( gifcp, cp, msgp, st2p, datap, encaplen )
s*
s*	Routine to do initial processing of received SCMP messages.
s*	gifcp is a pointer to the generic interface from which the PDU
s*		was received
s*	cp is a pointer to the ST2pcb corresponding to the HID - 0 if here
s*	msgp is a ptr to the first mbuf in the chain for the received PDU,
s*		it may or may not contain the ST2 Header ...
s*	st2p is a ptr to the mbuf containing the ST2 Header; it may be
s*		identical to msgp; if here,
s*		1) the mbuf contains at least the whole ST2 header plus
s*		   timestamp, but there are no timestamps with scmp/HID 0
s*		2) any IP encapsulation header immediately preceeds the
s*		   ST2 Header, its length is ...
s*	encaplen is the number of bytes in IP encapsulation header
s*	datap is a ptr after the ST2 Header (+ Timestamp)
s*
 *	should it take via AF_INET vs AF_COIP ???
 */

  enum ST2Errors			/* Next bugid 0x57320 0x57311 */
st2_CMPInput( gifcp, cp, argmsgp, argst2p, argdatap, encaplen ) /* from st2_intr */
					/* dsp dspin via st2_init */
  struct aNetIf		*gifcp;		/* Generic receiving interface ptr */
  struct aST2pcb	*cp;		/* (ST2pcb from HID (0) hash) */
  struct aPktDesc	*argmsgp,	/* we dispose, first mbuf in chain */
			*argst2p;	/* aPktDesc with <encaplen of ip>
					   & ST2 header */
  octet1		*argdatap;	/* Ptr after ST2 Header */
  int			 encaplen;	/* IP bytes PRECEEDING ST2 Header */
{
    enum ST2Errors	 retcod;
    int			 actlen,
			 hash_index,
			 scmlen,	/* Bytes in SCM PDU */
			 st2len;	/* Bytes in ST2 + SCMP PDU */
    octet2		 ref;
    struct aPktDesc	*parsebufp;	/* Ptr to buffer holding pscmp */
    struct aPrsSCM	*pscmp;		/* Ptr to parsed SCM PDU struct */
    struct aSCMPHeader	*scmhdrp;	/* Ptr to SCMP Header */
    struct aST2Header	*st2hdrp;	/* Ptr to ST2 Header */
    struct aVLink	*vlinkp;
    struct Target	*targetp;
    struct ifaddr	*ifaddrp;
    struct timeval	 timestamp;

    struct aPktDesc	*msgp = argmsgp,
			*st2p = argst2p;
    octet1		*datap = argdatap;



    uniqtime ( &(timestamp) );		/* When received */

    /* -------------------------------------------------------------------- */
    /* Find ST-II Header */

    st2hdrp = pkt2data (st2p, struct aST2Header *);
    st2len = ntoh2 (st2hdrp->TotalBytes);

    /* -------------------------------------------------------------------- */
    /* Make sure SCM PDU is in contiguous memory */
    /* ??? make sure properly aligned */

    if ( st2len > st2p->pkt_length )	/* Are all bytes in this buffer? */
    {
#if 1					/* Sometimes get fragmented pkt */
	char		*dp;
	int		 cl,
			 dl,
			 rl = st2len;
	struct aPktDesc	*tmp,
			*pp = NO_PKTDSCP;

	dp = GetSpace (char *, &( pp ),PKT_DATA,PKT_MD_NEW,PKT_XA_PS,
		       16,ROUND (encaplen,octet4),st2len,0,0);
	if ( dp EQ (char *) NULL )
	{
	    cp->prvhop.thruput.ctrl_in_drop_bytes += st2len;
	    cp->prvhop.thruput.ctrl_in_drop_pkts++;
	    FreePkts (pp);
	    FreePkts (msgp);
	    BUGRETURN (-,DropNoResrc,0x57301,enum ST2Errors);
	}

	if ( encaplen > 0 )
	    Bcopy ((char *) st2hdrp - encaplen,dp - encaplen,encaplen);

	dl = pp->pkt_length;

	for ( tmp = st2p ; tmp NE NO_PKTDSCP ; tmp = tmp->pkt_nxtbufp )
	{
	    cl = tmp->pkt_length;
	    if ( cl > rl )
		cl = rl;
	    if ( cl > dl )
		/* ??? should never be */
		cl = dl;
	    Bcopy (pkt2data (tmp, char *),dp,cl);
	    dp += cl;
	    rl -= cl;
	    dl -= cl;
	}

	if ( rl > 0 )			/* really inconsistency */
	{
	    cp->prvhop.thruput.ctrl_in_drop_bytes += st2len;
	    cp->prvhop.thruput.ctrl_in_drop_pkts++;
	    FreePkts (pp);
	    FreePkts (msgp);
	    BUGRETURN (-,DropNoResrc,0x57302,enum ST2Errors);
	}

	pp->pkt_length = st2len;

	if ( msgp EQ st2p )		/* Replace whole chain */
	{
	    FreePkts (msgp);
	    msgp = pp;
	}
	else				/* Replace tail of chain */
	{
	    tmp = msgp;
	    for ( ; tmp->pkt_nxtbufp NE st2p ; tmp = tmp->pkt_nxtbufp )
		continue;		/* Locate tail */
	    tmp->pkt_nxtbufp = pp;
	    FreePkts (st2p);
	}
	st2p = pp;

	rl = datap - (octet1 *) st2hdrp;
	st2hdrp = pkt2data (st2p, struct aST2Header *);
	datap = Mkp (octet1 *, st2hdrp, rl);
#else
	Count (scmp_fragmented,1);	/* count ctrl msg spanning buffers */
	cp->prvhop.thruput.ctrl_in_drop_bytes += st2len;
	cp->prvhop.thruput.ctrl_in_drop_pkts++;
	FreePkts (msgp);		/* ??? message to fix device driver! */
	BUGRETURN (-,TruncatedCtl,0x57303,enum ST2Errors);
#endif 1
    }

    /* -------------------------------------------------------------------- */
    /* Find SCMP Header */

    scmlen = st2len - (datap - (octet1 *) st2hdrp);
    scmhdrp = Mkp (struct aSCMPHeader *, datap, 0);

    IfLog (DBG_CTL_MSG)
	(void) printf ( "RcvSCMP:%2x %2x %4x RV:%4x %4x Ref:%4x %4x IP:%8x (%4x %8x)\n",
		        ntoh1 (scmhdrp->OpCode), ntoh1 (scmhdrp->Options),
		        ntoh2 (scmhdrp->TotalBytes), ntoh2 (scmhdrp->RVLId),
		        ntoh2 (scmhdrp->SVLId) , ntoh2 (scmhdrp->Ref),
		        ntoh2 (scmhdrp->LnkRef), ntohIP (scmhdrp->SndrIPAdr),
		        ntoh2 (scmhdrp->peropcode2), scmhdrp->peropcode4 );

    /* -------------------------------------------------------------------- */
    /* Check for broken packets */

    if ( scmlen < sizeofaSCMPHeader )
    {
	Count (scmp_tooshort,1);	/* count scm too short */
	cp->prvhop.thruput.ctrl_in_drop_bytes += st2len;
	cp->prvhop.thruput.ctrl_in_drop_pkts++;
	FreePkts (msgp);
	BUGRETURN (-,TruncatedCtl,0x57304,enum ST2Errors);
    }

    if ( ntoh2 (scmhdrp->TotalBytes) < scmlen )
    {
	Count (scmp_leninconsist,1);	/* count bad length */
	cp->prvhop.thruput.ctrl_in_drop_bytes += st2len;
	cp->prvhop.thruput.ctrl_in_drop_pkts++;
	FreePkts (msgp);
	BUGRETURN (-,TruncatedCtl,0x57305,enum ST2Errors);
    }

    scmlen = ntoh2 (scmhdrp->TotalBytes);

    /* -------------------------------------------------------------------- */
    /* Validate checksum */

    {	octet2		*cksump = Mkp (octet2 *, datap, 0);
        unsigned long	 cksum = 0;
	int		 i = scmlen / 2;


	for ( ; i > 0 ; --i )	cksum += *cksump++;

	/* should be a multiple of 4 !! */
	if ( (scmlen & 1) NE 0 )
	    cksum += *cksump++ & hton2 (0xFF00);

	cksum = (cksum >> 16) + ( cksum & 0xFFFF );
	cksum = (cksum >> 16) + ( cksum & 0xFFFF );
	if (    (cksum NE 0xFFFF) AND (cksum NE 0)
#ifdef STIIDEBUG
	    AND (scmhdrp->Checksum NE 0) AND (NOT ConfigFlag (ST2FlgSCMPCksm))
#endif STIIDEBUG
	    )
	{
	    Count (scmp_cksum,1);	/* count bad scmp checksum */
	    cp->prvhop.thruput.ctrl_in_drop_bytes += st2len;
	    cp->prvhop.thruput.ctrl_in_drop_pkts++;
	    FreePkts (msgp);
	    BUGSTOP (,scmlen,0x57306);
	    BUGSTOP (,cksum,0x57307);
	    BUGRETURN (-,CksumBadCtl,0x57308,enum ST2Errors);
	}
    }

    /* -------------------------------------------------------------------- */
    /* Have SCM PDU Parsed */

    parsebufp = NO_PKTDSCP; /* SCMPDecode gets memory */
    retcod = SCMPDecode( scmhdrp, scmlen, &parsebufp, &actlen );
    if ( retcod NE NoError )
    {
	cp->prvhop.thruput.ctrl_in_drop_bytes += st2len;
	cp->prvhop.thruput.ctrl_in_drop_pkts++;
	/* ??? record last error for cp */
	/* ??? count parse errors */
	if ( parsebufp NE NO_PKTDSCP )
	    FreePkts (parsebufp);
	FreePkts (msgp);
	return ( retcod );	
    }

    parsebufp->pkt_nxtbufp = msgp;	/* Parsed msg owns input PDU */

    cp->prvhop.thruput.ctrl_in_bytes += st2len;
    cp->prvhop.thruput.ctrl_in_pkts++;

    /* Locate parsed PDU structure */
    pscmp        = pkt2data (parsebufp,struct aPrsSCM *);
    pscmp->selfp = (struct aPktDesc*) parsebufp; /* Parsed msg owns itself */
    pscmp->gifcp = gifcp;		/* Save receoving interface */
    pscmp->timestamp = timestamp;	/* When received */

    /* -------------------------------------------------------------------- */
    /* Insert our receiving address */
    /* ??? do this here, only used in HandleConnect */

    for ( ifaddrp = gifcp->osifcp->if_addrlist
	 ; ifaddrp NE (struct ifaddr *) NULL
	 ; ifaddrp = ifaddrp->ifa_next )
    {
	if ( ifaddrp->ifa_addr.sa_family EQ AF_INET )
	    break;
	/* ??? more than one */
    }

    if ( ifaddrp EQ (struct ifaddr *) NULL )
    {
	Count (other,1);		/* Not have an IP address?? */
	BUGSTOP (-,InvalidPointer,0x57309);
	/* ??? rcvingsa is trash here */
    }
    else
    {
	Bcopy (&( ifaddrp->ifa_addr ),&( pscmp->rcvingsa ),
					sizeof (pscmp->rcvingsa));
    }

    /* ??? if have this is rcvingsa required too/on-demand lookup */

    if ( (pscmp->encaplen = encaplen) > 0 )
	pscmp->encaphdrp = Mkp (struct ip *, st2hdrp, (- encaplen));

    /* ==================================================================== */
    /* Process Parsed SCMP Message */

    /* Find VirtualLink */

    retcod = NoError;
    vlinkp = NO_VLINKP;

    /* First get the virtual link using any VLId available. */

    if ( pscmp->hoscmphdr.OpCode EQ (int) OpHello )
	;				/* skip this */
    else if (pscmp->hoscmphdr.RVLId NE VLID_NONE)
    {
	vlinkp = VLinkByVLId(pscmp->hoscmphdr.RVLId);
	/* ??? Validate any SVLId/sender; what if differ */
    }
    else if ( pscmp->hoscmphdr.SVLId NE VLID_NONE )
    {
	/* Remote lookups scan along a chain of values with same
	 * hash index.
	 */
	hash_index = VLinkHash (pscmp->hoscmphdr.SndrIPAdr,
			        pscmp->hoscmphdr.SVLId);
	vlinkp = st2_vlinks.remote_hash[ hash_index ];
	for ( ; vlinkp NE NO_VLINKP ; vlinkp = vlinkp->nextp )
	{
	    /* We're looking for entry with matching vlid and host */
	    if (    (vlinkp->type NE FREE_VLINK) /* ??? more states */
		AND (pscmp->hoscmphdr.SVLId EQ vlinkp->remote_vlid)
		AND (pscmp->hoscmphdr.SndrIPAdr EQ vlinkp->rmtIPAdr) )
		/* ??? check name */
	    {
		break;			/* out of for loop */
	    }
	} /* end of for all vlinks in hash bucket loop */
    }

    /* -------------------------------------------------------------------- */
    /* This switch statement is used to verify:
     * 1. Whether a VLID is required at all.
     * 2. How it must be provided.  E.g., some op'ns require RVLId,
     *    some either SVLId or RVLId, some don't require a VLID.
     * 3. Whether a VLID must reference an active link.
     *    This switch statement could be table driven.
     */

    switch ( pscmp->hoscmphdr.OpCode )
    {
    /* These operations flow upstream and require an active virtual
     * link identified by the receiver vlid field.
     */
    case OpAccept:
    case OpChangeRequest:
    case OpHidApprove:
    case OpHidChangeRequest:
    case OpHidReject:
    case OpRefuse:
					/* ??? also SVLId */
	if (pscmp->hoscmphdr.RVLId EQ VLID_NONE)
	{
	    retcod = ProtocolError;
	    break;
	}
	/* Falls through */

    /* These requests flow downstream after the "connect" request
     * so they must identify an active virtual link with either
     * the SVLId or RVLId field.
     */
    case OpChange:
    case OpHidChange:
    case OpDisconnect:

	if (vlinkp EQ NO_VLINKP)
	{
	    if (pscmp->hoscmphdr.RVLId NE VLID_NONE)
		retcod = InvalidRVLId;
	    else if (pscmp->hoscmphdr.SVLId NE VLID_NONE)
		retcod = InvalidSVLId;
	    else
		retcod = ProtocolError;
	}
	break;

    /* Connect requires an SVLId, but generally not an active link.
     * However, if an RVLId is included, then it's an add and
     * the virtual link must exist.
     */
    case OpConnect:

	if (pscmp->hoscmphdr.SVLId EQ VLID_NONE)
	{
	    retcod = ProtocolError;
	}
	else if (    (pscmp->hoscmphdr.RVLId NE VLID_NONE)
		 AND (vlinkp EQ NO_VLINKP) )
	{				/* ??? compare name */
	    retcod = InvalidRVLId;
	}
	break;

    /* These may be acknowledging disconnect or refuse so the
     * virtual link may have been dismantled even if a VLId
     * field is provided.
     */
    case OpAck:
    case OpErrorInRequest:
    case OpErrorInResponse:
    case OpNotify:			/* ??? */
	break;

    /* ??? Some may act like connect and should establish a vlinkp when
     * VLIDs provided.
     */
    case OpHello:
    case OpStatus:
    case OpStatusResponse:
	break;

    default:
	retcod = OpCodeUnknown;
	break;
    } /* end of opcode switch */

    /* -------------------------------------------------------------------- */
    /* If VLId matches an active link, make sure that it isn't a
     * stale reference by checking the other VLId fields and
     * stream Name (if provided in the message).
     * Note: Hellos have NO_VLINKP here so this block is skipped.
     */
    if ( (retcod EQ NoError) AND (vlinkp NE NO_VLINKP) )
    {
	/* If we don't know the vlid for the other side of the
	 * link yet, grab it now.
	 */
	if ( vlinkp->remote_vlid EQ VLID_NONE )
	{
	    if (    (pscmp->hoscmphdr.SVLId NE VLID_NONE)
		AND (vlinkp->type NE NEIGHBOR_VLINK) )
	    {
		(void) VLinkSetRemote( vlinkp, pscmp->hoscmphdr.SVLId,
				       pscmp->hoscmphdr.SndrIPAdr);
		/* ??? boolean error */
	    }
	}
	/* Otherwise make sure the vlid info matched. */
	else if (   (vlinkp->remote_vlid NE pscmp->hoscmphdr.SVLId)
		 OR (vlinkp->rmtIPAdr NE pscmp->hoscmphdr.SndrIPAdr) )
	{
	    if ( vlinkp->remote_vlid NE pscmp->hoscmphdr.SVLId )
	    {
		BUGSTOP (,((  vlinkp->remote_vlid << 16)
			    | pscmp->hoscmphdr.SVLId),0x57310);
	    }
	    if ( vlinkp->rmtIPAdr NE pscmp->hoscmphdr.SndrIPAdr )
	    {
		BUGSTOP (,vlinkp->rmtIPAdr,0x57312);
		BUGSTOP (,pscmp->hoscmphdr.SndrIPAdr,0x57313);
	    }
	    retcod = InconsistVLID;
	    BUGSTOP (,vlinkp->local_vlid,0x57314);
	    vlinkp = NO_VLINKP;
	}

	/* If ctrl msg contains a stream Name, make sure it matches too. */
	if (    (retcod EQ NoError)
	    AND (vlinkp->cp NE NO_PCBP) )
	{
	    if (    (pscmp->Namep NE (struct aName *) NULL)
		AND (Bcmp (pscmp->Namep,&( vlinkp->cp->Name ),
			   sizeof (struct aName)) NE 0) )
	    {
		retcod = InconsistVLID;	/* ???better - wrong stream name */
		BUGSTOP (,pscmp->hoscmphdr.OpCode,0x57315);
		vlinkp = NO_VLINKP;
	    }
	    else
	    {				/* ??? is this binding necessary */
		vlinkp->cp->refcnt++;
		IfLog (DBG_CNT_PCB)
		    printf ( "refcnt: cp = %8x  refcnt = %2d  st2_CMPInput     pscmp %x on vlink %x\n",
			     vlinkp->cp, vlinkp->cp->refcnt, pscmp, vlinkp );
		pscmp->cp = vlinkp->cp;
	    }
	}
    } /* no error & have vlinkp */


    /* ==================================================================== */
    /* Protocol Processing for the SCM PDU */

    if ( retcod EQ NoError )
    {
	/* ---------------------------------------------------------------- */
	/* Common pre-dispatch operations */

	/* Call completion routines of acked messages & release them */

	switch ( pscmp->hoscmphdr.OpCode )	/* ??? control bit */
	{				/* stop retransmitting */
	case OpAccept:
	case OpRefuse:
	    ref = pscmp->hoscmphdr.LnkRef;
	    goto acked;

	case OpAck:
	case OpErrorInRequest:
	case OpErrorInResponse:
	case OpHidApprove:
	case OpHidReject:
	    ref = pscmp->hoscmphdr.Ref;
  acked:
	    if ( ref EQ (octet2) 0 )
		break;

	    /* BEWARE: this may trigger deletion of a target, its vlinkp, and
	     * the aST2pcb
	     */
	    (void) CtrlMsgDropAcked( pscmp, ref );

	    if (   (vlinkp EQ NO_VLINKP)
		OR (vlinkp->type EQ FREE_VLINK) )
		vlinkp = NO_VLINKP;	/* bullet proof handlers */

	    break;			/* We keep pscmp */

  	default:
	    break;
	} /* end of opcode switch */

	/* ---------------------------------------------------------------- */
	/* Dispatch to SCM PDU handler */

	if ( (int) pscmp->hoscmphdr.OpCode >= (int) ConfigParm (num_st_opcodes) )
	    retcod = OpCodeUnknown;
	else
	{
	    /* HandleStar( pscmp, vlinkp ) */ /* from st2_CMPInput */

	    retcod = ops[ (int) pscmp->hoscmphdr.OpCode ]( pscmp, vlinkp );

	    /* ------------------------------------------------------------ */
	    /* Common post-disptch functions */

	    switch ( (int) pscmp->hoscmphdr.OpCode )
	    {
	    case OpAck:
	    case OpErrorInRequest:
	    case OpErrorInResponse:
	    case OpHidApprove:		/* does own DoError */
	    case OpHidReject:		/* does own DoError */

		if ( vlinkp EQ NO_VLINKP )
		{
		    BUGSTOP (,(int) pscmp->hoscmphdr.OpCode,0x57316);
		    break;
		}

		if ( vlinkp->cp EQ NO_PCBP )
		{
		    if ( vlinkp->np EQ NO_NBRP )
			BUGSTOP (,vlinkp,0x57317);
		    break;
		}

		ForAllTargets (vlinkp->cp, targetp)
		{
		    if ( NOT IsTargState (targetp, TS_BUSY) )
			continue;

		    if (    targetp->pending_reference
			EQ (Reference) pscmp->hoscmphdr.Ref )
		    {
			/* ??? Phasing out, whats left */
			BUGSTOP (,targetp->pending_reference,0x57318);
					/* notify_pending_Target */
					/* cancel_pending_Target */
			targetp->pending_reference = (Reference) 0;
					/* ??? obsolete, but tested above */
		    }
#ifdef STIIDEBUG
		    else if ( targetp->pending_reference NE (Reference) 0 )
		    {
			BUGSTOP (-,(  ((targetp->pending_reference) << 16)
				    | ((pscmp->hoscmphdr.OpCode & 0xF) << 12)
				    | (pscmp->hoscmphdr.Ref & 0xFFF)),0x57319);
		    }
#endif STIIDEBUG
		} /* end of for all targets loop */

		/* Run state machine if required */
		if ( vlinkp->cp->needupdate NE 0 )
		    st2_PCBUpdate( vlinkp->cp );
		/* cp may be invalid */

		retcod = NoError;	/* for DoError test below */
		break;

	    default:
	        break;
	    } /* end of opcode switch */
	}
    }

    /* ==================================================================== */
    /* Send an error message if needed, cleanup */

    if ( retcod NE NoError )
    {
	/* ??? record error */
	(void) DoError( pscmp, vlinkp, retcod );
    }

    CtrlMsgDel( pscmp );

    return ( retcod );

} /* end of st2_CMPInput */

/*
S* st2_DoAccept ( targetp, flowspecp, opts, userdatap )
s*
s*	Send an ACCEPT message for the specified target.
s*
 */

  enum ST2Errors			/* Next bugid 0x57402 */
st2_DoAccept( targetp, flowspecp, opts, userdatap ) /* from api_control */
  struct Target		*targetp;
  struct aFlowSpec	*flowspecp;	/* May be NULL (volitile) */
  unsigned int		 opts;
  struct aUserData	*userdatap;	/* May be NULL (volitile) */
					/* ??? error return */
{
    enum ST2Errors	 reason;
    int			 oldpri;
    struct aFlowSpec3	*fsp;
    struct aPrsSCM	*accept;


    if ( (accept = CtrlMsgCreate()) EQ NO_PSCMP )
    {    
#ifdef  STIIDEBUG
	(void) printf ( "st2_DoAccept: out of memory\n" );
#endif  STIIDEBUG
	IfLog (DBG_UPDTARG)
	    log_Target( targetp, "accept_Message - no buffers" );

	return ( CantGetResrc );
    }

    IfLog (DBG_CTL_MSG)
	(void) printf ( "accept_Message\n" );
    IfLog (DBG_UPDTARG)
	log_Target( targetp, "accept_Message" );

    kassert (targetp->error EQ NoError); /* ??? */

    accept->comp = CompAccept;	/* Completion routine */

    accept->DetectorIPAdr =
			(octet4) targetp->cp->upstream_vlinkp->lclIPAdr;
    accept->detectoripadrp = &( accept->DetectorIPAdr );
    accept->hoscmphdr.OpCode = (octet1) OpAccept;
    accept->hoscmphdr.Options = opts & (int) STTSR;
    accept->hoscmphdr.Ref = NxtRefNum ();
    targetp->pending_reference = accept->hoscmphdr.Ref;
					/* ??? obsolete */
    accept->hoscmphdr.LnkRef = targetp->linked_reference;

    if ( flowspecp EQ (struct aFlowSpec *) NULL )
	flowspecp = targetp->flow_specp;
    /* ??? else cache final */

    (void) FlowSpecEncode( accept, flowspecp );

    accept->UserDatap = userdatap;

    if ( CtrlMsgAddTarg( accept, targetp, UPSTREAM_VLINK ) )
    {
	/* Must hold ACCEPTs until sending HID is valid */
	if ( targetp->cp->hid_ready )
	{
	    reason = CtrlMsgSend( accept, targetp->cp->upstream_vlinkp );
					/* It CtrlMsgDels connmsgp */
	    if (    (reason EQ NoError)
		AND (targetp->cp->upstream_vlinkp->np NE NO_NBRP) )
	    {
		/* ??? use received so application cannot decrease it */
		/* ??? flowspecp is  struct aFlowSpec3 */
		fsp = (struct aFlowSpec3 *) flowspecp;
		(void) NbrInsert( targetp->cp->upstream_vlinkp,
				  ntoh2 (fsp->RecoveryTimeout) );
		/* sets targetp->cp->upstream_vlinkp->rcvryto */
	    }
	}
	else
	{
	    oldpri = splnet ();
	    {
		accept->nextp = targetp->cp->accepts;
		targetp->cp->accepts = accept;
	    }
	    (void) splx ( oldpri );
	    reason = NoError;
	}

	/* ??? make state change for successfully accepted targets */
	if ( reason EQ NoError )
	{
	    targetp->lnkrefflg = 0;

	    if ( IsTargState (targetp,TS_WANTOPEN) )
	    {
		SetTargState (targetp, TS_OPEN);
	    }				/* No reply from internal gateway */
	}
	/* else returning error will notify application */
    }
    else
    {
#ifdef  STIIDEBUG
	(void) printf ( "Failed to add target to accept\n" );
#endif  STIIDEBUG
	BUGSTOP (-,targetp,0x57401);
	reason = ErrorUnknown;
    }

    return ( reason );
}

/*
S* st2_DoConnect ( cp, sockadrp, maxsockbytes, flowspecp, userdatap )
s*
s*	Routine to perform application connect request.
s*
 */

  int	/* enum ST2Errors */		/* Next bugid 0x57518 */
st2_DoConnect( cp, sockadrp, maxsockbytes, flowspecp, userdatap ) /* from api_control st2_usrreq */
  struct aST2pcb	*cp;
  struct sockaddr_st2	*sockadrp;
  int			 maxsockbytes;
  struct aFlowSpec	*flowspecp;	/* ARGSUSED ??? more work */
  struct aUserData	*userdatap;
{
    int			 match,
			 plen,
			 retcod,
			 unique_id;
    struct aApplEntity	*lclentp;
    struct aName	 streamName;
    struct aOrigin	*op,
			*Originp;
    struct aParameter	*nextparmp,
			*parmp,
			*parmlimitp;
    struct aST2pcb	*wcp;
    struct aVLink	*vlinkp;
    struct timeval	 tv;


    /* ??? via usrreq/connect @ splnet */

    /* ??? too soon, after routing for good local address to use */
    /* Use local origin binding if the application specified one. */
    lclentp = (struct aApplEntity *) NULL;

    if ( cp->soname.applent.pcode NE (octet1) 0 )
    {
	lclentp = (struct aApplEntity *) &( cp->soname.applent );

	/* Make sure entity is not already bound to another stream. */
	/* (May need to adjust "st2_ApplEntityMatch ()") */

	/* ???  Should the origin parameters be bound to the stream or
		the aST2pcb?  The intermediate gateway agents do not
		care about origin, only the host agents.  Yet,stream
		needs to have control of the origin parameter so that
		it can update the IPAdr field to match address of the
		primary interface.  For either approach, we need to
		check for a conflict with the binding of any other
		aST2pcb because target aST2pcb are not bound to
		streams and so will not be recognized by this call.
	 */

	retcod = 0;

	/* ??? maybe a function of 1socket option */
	ForAllpcbs (wcp)
	{
	    if ( wcp EQ cp )
		continue;		/* only check other pcbs */

	    if ( (wcp->flags & LocalConnection) EQ 0 )
		continue;		/* skip intermediates */

	    if (    (( cp->sop->so_options & SO_REUSEADDR) NE 0)
		AND ((wcp->sop->so_options & SO_REUSEADDR) NE 0) )
		continue;

	    if ( 0 < st2_ApplEntityMatch( lclentp,
			      (struct aOrigin *) &( wcp->soname.applent ) ) )
	    {
		retcod = EADDRINUSE;
		break;
	    }
	} /* end of for loop */

	if ( retcod NE 0 )
	    return ( retcod );
    }

    /* ------------------------------------------------------------ */
    /* Create the Name for the new stream.  We control the Names. */
    if ( cp->Name.pcode EQ 0 )
    {
	streamName.pcode = hton1 ((octet1) STpName);
	streamName.plen  = hton1 (sizeof (struct aName));
	streamName.IPAdr = htonIP (local_inetaddr);
	do
	{
	    uniqtime ( &tv );
	    unique_id = UNIQ_MASK & Count (unique_id,1);
	    if ( unique_id EQ 0 )
	    {
		unique_id = UNIQ_MASK & Count (unique_id,1);
		BUGSTOP (,0,0x57501);	/* Amazing! */
	    }

	    streamName.UniqId = hton2 (unique_id);
	    streamName.Timstmp = hton4 (tv.tv_sec);

	} while ( PCBByName( &streamName ) NE NO_PCBP );


	/* ---------------------------------------------------------------- */
	/* Find other meanders of the specifed stream and link them all
	 * together.
	 */


	/* lock aST2pcbs */
	{
	    wcp = PCBByName( &( streamName ) );

	    /* ??? find any stream with given Name */
	    /* ??? won't work for meanders */

	    if ( wcp NE NO_PCBP )
	    {
		if ( wcp->nxtNamep EQ NO_PCBP )
		{
		    cp->nxtNamep = wcp;	/* Form ring */
		    wcp->nxtNamep = cp;
		}
		else
		{
		    cp->nxtNamep = wcp->nxtNamep; /* Add to ring */
		    wcp->nxtNamep = cp;
		}
	    }
	    else
	    {
		kassert (wcp EQ NO_PCBP);
	    }

	    cp->Name = streamName;	/* structure copy */

	} /* unlock aST2pcbs */


	op = &( cp->Origin );	
	Originp = (struct aOrigin *) lclentp;

	/* ??? else ever used */
	if ( Originp NE (struct aOrigin *) NULL )
	{
	    Bcopy (Originp,op,sizeof (*op));

	    if ( op->pcode NE hton1 ((octet1) STpOrigin) )
		op->pcode = hton1 ((octet1) STpOrigin);
	}
	else
	{
	    op->pcode = hton1 ((octet1) STpOrigin);
	    op->plen = hton1 (sizeof (*op));
	    op->NextPcol = hton1 (0);
	    op->SAPBytes = hton1 (2);
	    op->IPAdr = htonIP (INADDR_ANY);

	    match = 1;
	    while ( match EQ 1 )
	    {
		if ( st2_stats.sap_next EQ (octet2) 0 )
		    st2_stats.sap_next = 0xC000 + 1024;
		op->SAP[0] = 0xFF & (st2_stats.sap_next >> 8);
		op->SAP[1] = 0xFF &  st2_stats.sap_next++;

		match = 0;

		ForAllpcbs (wcp)
		{
		    /* ??? should only check local origins (?+ targets),
		       not all sockets */

		    if ( wcp EQ cp )
			continue;

		    if (    (wcp->Origin.SAPBytes > 0)
			AND (   (op->SAP[0] EQ wcp->Origin.SAP[0])
			     OR (    (wcp->Origin.SAPBytes > 1)
				 AND (op->SAP[1] EQ wcp->Origin.SAP[1]))) )
		    {
			match = 1;
			break;		/* match, try st2_stats.sap_next */
		    }
		} /* end of scanning all pcbs loop */
	    } /* end of while looking for free SAP loop */
	}

	IfLog (DBG_INI_DS_MSG)
	    log_Stream( cp, "StreamCreate" );

	/* ---------------------------------------------------------------- */
	/* ??? merge */
	cp->optapi = sockadrp->st_options & (STOptPBit | STOptSBit | STTSR);
	/* Assume request will succeed */
	cp->optlvl = cp->optapi & (unsigned char) STTSR;
	/* ??? ought to be set after all ACCEPTs */

	switch ( cp->optlvl )
	{
	default:
	case STTSRNotImp:
	case STTSRNo:
	    cp->optmask &= ~ (unsigned char) STDataTBit; /* Force it off */
	    break;
	case STTSRYes:
	    cp->optneg  |=   (unsigned char) STDataTBit; /* Force it on */
	    /* Fall through */
	case STTSROpt:
	    cp->optmask |=   (unsigned char) STDataTBit; /* Permit it */
	    break;
	} /* end of desired timestamp mode switch */
    }

    /* ------------------------------------------------------------ */
    /* Need to have cp->hid_ready to propagate accepts */

    if ( cp->hop_id[0] EQ (HID) 0 )
    {
	cp->hop_id[0] = HID_RESV3;
	cp->hid_ready = TRUE;

	if ( cp->accepts NE NO_PSCMP )
	    relay_accepts(cp);		/* SB:rror */
    }

    /* ------------------------------------------------------------ */
    /*
     * Create upstream control link entry, attached to forwarder, and
     * add remote markings to link entry index.
     */

    if ( cp->upstream_vlinkp EQ NO_VLINKP )
    {
	vlinkp = VLinkNew( cp, UPSTREAM_VLINK, APPL_VLINK );
	if ( vlinkp EQ NO_VLINKP )
	{
	    BUGSTOP (-,1,0x57502);
	    /* ??? shouldn't trash connection/forwarder on error */
	    st2_PCBDel( cp );		/* WARNING: cp is no longer valid */
	    return ( ENOBUFS );
	}

	/* Set previous hop to be the application. */
	cp->upstream_vlinkp = vlinkp;
	vlinkp->route.type = TO_APPL;
	vlinkp->route.nxthopp = &( cp->prvhop );

	cp->prvhop.gifcp = cp->gifcp;
	vlinkp->lclIPAdr = local_inetaddr;
	vlinkp->rmtIPAdr = local_inetaddr;
	cp->prvhop.fwdtoIPAdr = local_inetaddr;
	cp->prvhop.HID = VLID_RESV3;

	/* Connect our application interface to ourself */
	if ( st2_ApplIfcBind( cp->gifcp, cp ) NE NoError )
	{
	    BUGSTOP (-,1,0x57503);
	    /* ??? shouldn't trash connection/forwarder on error */
	    st2_PCBDel( cp );		/* WARNING: cp is no longer valid */
	    return ( EINVAL );
	}

	MTUAdjust( cp, &( cp->prvhop ) );

	/* ??? origin per departing interface */
	/* ??? should not be required for socket */
    }

    /* ------------------------------------------------------------ */
    /* ??? should be able to get here via setsockopt () */

    nextparmp = &( sockadrp->st_parms[0] );
    parmlimitp = Mkp (struct aParameter *, sockadrp, maxsockbytes);
    retcod = 0;
    while ( (retcod EQ 0) AND (nextparmp < parmlimitp) )
    {
	parmp = nextparmp;
	plen = ntoh1 (parmp->plen);

	if ( (plen <= 0) OR ((plen & 3) NE 0) )
	{
	    BUGSTOP (-,1,0x57504);
	    retcod = EINVAL;
	    break;
	}

	nextparmp = Mkp (struct aParameter *,parmp,plen);

	if ( parmlimitp < nextparmp )
	{
	    BUGSTOP (-,1,0x57505);
	    retcod = EINVAL;
	    break;
	}

	switch ( ntoh1 (parmp->pcode) )
	{
	case STpFlowSpec:		/* 2 */
					/* ??? What if already have one? */
	    retcod = st2_FlowSpecFromAppl( (struct aFlowSpec *) parmp,
					        &( cp->flow_spec ) );
	    if ( retcod NE 0 )
		BUGSTOP (-,1,0x57506);
	    break;

	/* ??? should dispatch by flowspec version */
	case STpRFlowSpec:		/* 12 */
#ifdef NOTYET
	    /* ??? separate pcb for reverse stream?? */
	    retcod = st2_FlowSpecFromAppl( (struct aFlowSpec *) parmp,
					        &( cp->rflow_spec ) );
	    if ( retcod NE 0 )
		BUGSTOP (-,1,0x57507);
#else
	    /* ??? cleanup */
	    /* ??? not yet support FullDuplex */
	    BUGSTOP (,EOPNOTSUPP,0x57508);
	    retcod = EOPNOTSUPP;
#endif NOTYET
	    break;

	case STpTargetList:		/* 20 */
	{
	    struct aTargetList	*tlp = (struct aTargetList *) parmp;
	    int			 count = ntoh2 (tlp->TargetCount),
				 tlen;
	    enum ST2Errors	 targerror;
	    struct aTarget	*aTargetp = &( tlp->targets[0] ),
				*nxtaTargetp,
				*targlimitp;


	    targlimitp = Mkp (struct aTarget *, tlp, ntoh1 (tlp->plen));

	    while ( --count >= 0 )	/* ??? vs paramerter length */
	    {
		/* Make sure this parameter is within the limits. */
		if ( aTargetp >= targlimitp )
		{
		    BUGSTOP (-,1,0x57509);
		    retcod - EINVAL;
		    break;		/* out of while */
		}

		tlen = ntoh1 (aTargetp->TargetBytes);

		if ( (tlen EQ 0) OR ((tlen & 3) NE 0) )
		{
		    BUGSTOP (-,1,0x57510);
		    retcod = EINVAL;
		    break;		/* out of while */
		}

		nxtaTargetp = Mkp (struct aTarget *,aTargetp,tlen);

		/* Make sure all of this parameter is within the limits. */
		if ( nxtaTargetp > targlimitp )
		{
		    BUGSTOP (-,1,0x57511);
		    retcod = EINVAL;
		    break;		/* out of while */
		}

		targerror = TargAdd( cp, aTargetp, (Reference) 0 );
		if ( targerror NE NoError )
		    retcod = EINVAL;

		aTargetp = nxtaTargetp;

	    } /* end of while more targets */
					/* ??? what if more data */
	} /* end of target list block */

	    if ( retcod NE 0 )
		BUGSTOP (-,1,0x57512);
	    break;


	case STpUserData:		/* 21 */
					/* ??? detect multiple */
	    userdatap = (struct aUserData *) parmp;
	    break;


	default:
	    BUGSTOP (-,maxsockbytes,0x57513);
	    BUGSTOP (-,parmp->pcode,0x57514);
	    retcod = EINVAL;
	    break;
	} /* end of switch */

    } /* end of while more parameters */

    /* ------------------------------------------------------------ */
    /* Cache User data away */

    if (    (retcod EQ 0)
	AND (userdatap NE (struct aUserData *) NULL) )
    {
	plen = ntoh1 (userdatap->plen);

	if ( cp->userdatap NE (struct aUserData *) NULL )
	{
	    BUGSTOP (-,DupParm,0x57515); /* ??? be more forgiving */
	}
	else
	{
	    char *udp = GetSpace (char *, &( cp->fwdrfreespace ), PKT_DATA,
				  PKT_MD_NEW, PKT_XA_PS, 0,0, plen, 0,0);

	    if ( udp NE (char *) NULL )
	    {
		Bcopy (userdatap,udp,plen);
		cp->userdatap = Mkp (struct aUserData *, udp, 0);
	    }
	    else
	    {
		BUGSTOP (,UserDataSize,0x57516);
		retcod = EMSGSIZE;
	    }
	}
    }

    /* ------------------------------------------------------------ */

    if ( retcod NE 0 )
    {
	BUGSTOP (-,retcod,0x57517);
	st2_PCBDel( cp );		/* WARNING: cp is no longer valid */
	return ( retcod );
    }

    /* ------------------------------------------------------------ */

    if ( IsState (PCB,cp,pcb_state,CINS_PREOPEN) )
    {
	soisconnecting ( cp->sop );
	SetState (PCB,cp,pcb_state,CINS_OPENING);

	/* For now, let it be open immediately. */
	SetState (PCB,cp,pcb_state,CINS_OPEN);
    }

    st2_PCBUpdate( cp );
    /* cp may be invalid */

    return ( 0 );
}

/*
S* st2_DoDisconnect ( cp, reason, targetp, userdatap )
s*
s*	Routine to process application disconnect request.
s*	??? need G-Bit and target list versions.
s*
 */

  int	/* enum ST2Errors */		/* Next bugid 0x57600 */
st2_DoDisconnect( cp, reason, targetp, userdatap ) /* from api_control st2_usrreq*2 */
  struct aST2pcb	*cp;
  enum ST2Errors	 reason;
  struct Target		*targetp;	/* May be NULL */
  struct aUserData	*userdatap;	/* May be NULL */
{
    /*??? The LocalOrigin flag is defined if the socket/aST2pcb is bound to
     * the source end of the stream.  We need to generate a refuse if the
     * socket/aST2pcb is bound to the target end of a stream (cp->gifcp).
     * To do so, we map each of the aST2pcb's gifcps back to a target and
     * call for the target to be disconnected.  Also make sure that the
     * aNxtHops are deactivated, and the ifnet structures properly released
     * as part of that process. */

    if ( IsState (PCB,cp,pcb_state,CINS_LISTEN) )
    {
	IfLog (DBG_CTL_MSG)
	    (void) printf ( "st2_DoDisconnect (%x/%x/%u) disconnecting listener\n",
			    cp, targetp, reason );

	SetState (FWD,cp,fwdr_state,FW_DISCONNECT);
    }
    else
    {
	IfLog (DBG_CTL_MSG)
	    (void) printf ( "st2_DoDisconnect (%x/%x/%u) disconnecting forwarder\n",
			    cp, targetp, reason );

	/* ??? drop_pending_CtrlMsg ( cp ); */
	/* ??? propagate G-Bit */

	if ( targetp EQ NO_TRGP )
	{
	    disconnect_Forwarder(cp, reason, userdatap);
	}
	else if ( NOT IsTargState (targetp,TS_DEAD) )
	{
	    if ( NOT IsTargStateIn (targetp,TS_RELEASED,TS (TS_CLOSED)
				    TS (TS_CLOSING) TS (TS_WANTCLOSED)) )
	    {
		SetTargState (targetp,TS_WANTCLOSED);
	    }
	    SetTargState (targetp,TS_DISCONNECTED);

	    targetp->discon_userdatap = userdatap;
	    targetp->error = reason;
	}
	else
	    return ( ENOTCONN );
    }

    cp->needupdate++;			/* In case no targets left */
    st2_PCBUpdate( cp );
    /* cp may be invalid */

    return ( 0 );
}

/*
S* st2_CMPOutput ( pscmp, vlinkp )
s*
s*	Routine to send an SCMP message via a virtual link.
s*
 */

  enum ST2Errors			/* Next bugid 0x57709 */
st2_CMPOutput( pscmp, vlinkp )		/* from CtrlMsgRelay CtrlMsgSend st2_ClkSlow */
  struct aPrsSCM	*pscmp;		/* we CtrlMsgDel it */
  struct aVLink		*vlinkp;	/* only if set_hopinfo_message */
{
    enum ST2Errors	 retcod;
    int			 len = pscmp->pdulen,
			 result;
    struct aPktDesc	*pktp;


    if ( vlinkp NE NO_VLINKP )
    {
	if ( NOT RouteCheck( &( vlinkp->route ) ) )
	{
#if 0					/* route.nxthopp may be invalid */
	    vlinkp->route.nxthopp->thruput.ctrl_out_drop_pkts++;
	    vlinkp->route.nxthopp->thruput.ctrl_out_drop_bytes += len;
#endif 0
	    CtrlMsgDel( pscmp );
	    /* ??? error type */
	    BUGRETURN (-,InvalidPointer,0x57701,enum ST2Errors);
	}

	len += vlinkp->route.nxthopp->encaplen;

	if ( pscmp->pktp EQ NO_PKTDSCP )
	{
	    vlinkp->route.nxthopp->thruput.ctrl_out_drop_pkts++;
	    vlinkp->route.nxthopp->thruput.ctrl_out_drop_bytes += len;;
	    CtrlMsgDel( pscmp );
	    /* ??? error type */
	    BUGRETURN (-,InvalidPointer,0x57702,enum ST2Errors);
	}

	IfLog (DBG_CTL_MSG)
	    log_Route( &( vlinkp->route ), "CtrlMsgSend route" );

	if ( vlinkp->cp NE NO_PCBP )
	{
	    if ( pscmp->cp NE NO_PCBP )	/* Only case should be CtrlMsgRelay */
	    {
		if ( pscmp->cp NE vlinkp->cp ) /* shouldn't be here */
		{
		    BUGSTOP (-,pscmp->cp,0x57703);
		    pscmp->cp->refcnt--;
		    IfLog (DBG_CNT_PCB)
			printf ( "refcnt: cp = %8x  refcnt = %2d  st2_CMPOutputAd pscmp %x on vlink %x\n",
				 pscmp->cp, pscmp->cp->refcnt, pscmp, vlinkp );
		    if ( pscmp->cp->refcnt <= 0 )
			st2_PCBDel( pscmp->cp );
			/* WARNING: pscmp->cp is no longer valid */

		    vlinkp->cp->refcnt++;
		    IfLog (DBG_CNT_PCB)
			printf ( "refcnt: cp = %8x  refcnt = %2d  st2_CMPOutputAd pscmp %x on vlink %x\n",
				 vlinkp->cp, vlinkp->cp->refcnt, pscmp,
				 vlinkp );
		}
		/* else same cp so no adjustment needed */
	    }
	    else			/* Bind pscm to cp */
	    {
		vlinkp->cp->refcnt++;
		IfLog (DBG_CNT_PCB)
		    printf ( "refcnt: cp = %8x  refcnt = %2d  st2_CMPOutput    pscmp %x on vlink %x\n",
			     vlinkp->cp, vlinkp->cp->refcnt, pscmp, vlinkp );
		pscmp->cp = vlinkp->cp;
	    }
	}
	else if ( pscmp->cp NE NO_PCBP )
	{
	    BUGSTOP (-,pscmp->cp,0x57704); /* should not get here */
	}


	pscmp->nxthopp = vlinkp->route.nxthopp;

	 switch ( pscmp->hoscmphdr.OpCode )
	{
	case OpAccept:
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_Accept) );
	    break;
	case OpAck:
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_Ack) );
	    break;
	case OpChange:
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_Change) );
	    break;
	case OpChangeRequest:
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_ChangeRequest) );
	    break;
	case OpConnect:
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_Connect) );
	    break;
	case OpDisconnect:
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_Disconnect) );
	    break;
	case OpErrorInRequest:
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_ErrorInRequest) );
	    break;
	case OpErrorInResponse:
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_ErrorInResponse) );
	    break;
	case OpHello:			/* Retransmit would fuzz timestamp */
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_Hello) );
	    break;
	case OpHidApprove:
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_HIDApprove) );
	    break;
	case OpHidChange:
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_HIDChange) );
	    break;
	case OpHidChangeRequest:
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_HIDChangeRequest) );
	    break;
	case OpHidReject:
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_HIDReject) );
	    break;
	case OpNotify:
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_Notify) );
	    break;
	case OpRefuse:
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_Refuse) );
	    break;
	case OpStatus:
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_Status) );
	    break;
	case OpStatusResponse:
	    pscmp->retryparms = * (struct stRetryParms *)
				  &( ConfigParm (rx_StatusResponse) );
	    break;
	default:
	    pscmp->retryparms.retry_limit = 0;
	    pscmp->retryparms.msec_interval = 0;
	    break;
	} /* end of switch */

	if (   (vlinkp->type EQ NEIGHBOR_VLINK)
	    OR (vlinkp->route.type EQ TO_APPL) )
	{
	    pscmp->retryparms.retry_limit = 0;
	    pscmp->retryparms.msec_interval = 0;
	}

	IfLog (DBG_CTL_MSG)
	{
	    (void) printf ( "pscmp %x op %d with %d retries, interval %d\n",
			    pscmp, pscmp->hoscmphdr.OpCode,
			    pscmp->retryparms.retry_limit,
			    pscmp->retryparms.msec_interval );
	}

	pscmp->retry_tod = 0;
    }

    /* ------------------------------------------------------------ */

    kassert (pscmp->nextp EQ NO_PSCMP); /* ??? what if it isn't */

    /*
     * We only send messages that are one-try-only or which haven't
     * reached their retry limit.
     */

    uniqtime ( &(pscmp->timestamp) );

    if ( pscmp->tries <= pscmp->retryparms.retry_limit )
    {
	if  ( pscmp->pktp EQ NO_PKTDSCP )
	{
	    retcod = Inconsistency;
	    BUGSTOP (-,retcod,0x57705);
	    CtrlMsgDel( pscmp );	/* decrement pcb's use count */
	    return ( retcod );
	}

	/* ??? only if cannot share buffers serially */
	/* ??? multiple use */

	if ( pscmp->tries < pscmp->retryparms.retry_limit )
	{   /* For messages that will be retried, copy the PDU. */
	    pktp = (struct aPktDesc *)	/* ??? mbufs */
			m_copy ( (struct mbuf *) pscmp->pktp, 0, PKT_COPYALL );
	    /* ??? what if copy failed to get space, bump retries? */
	}
	else
	{   /* On last try, use the existing PDU. */
	    pktp = pscmp->pktp;
	    pscmp->pktp = NO_PKTDSCP;
	}

	pscmp->tries++;

	/* ------------------------------------------------------------ */

	/* Queue before sending so ack will can be matched to it. */
	if ( pscmp->retryparms.msec_interval > 0 )
	{
	    int			 oldpri;
	    long		 new_tod;
	    struct aPrsSCM	*scan;


	    /* ??? only good for a second */
	    new_tod =   pscmp->retryparms.msec_interval
		      * ConfigParm (timeoutfactor);
	    new_tod /= 1000;
	    new_tod += pscmp->timestamp.tv_sec + (new_tod > 0) ? 0 : 1;
	    pscmp->retry_tod = new_tod;

	    /* Insert in time order.
	     *   - If queue is empty, scan will match EmptyRetryQ (i.e. the
	     *	queue header address), so new entry will be added as
	     *	the only one of the queue
	     *   - If queue isn't empty we scan backward from the end of the
	     *	queue until we find an entry with an earlier time of day.
	     *	We add the new entry after the one with an earlier time
	     *	of day.  If we reach the queue head first, we add the new
	     *	entry just after the queue head.
	     */

	    oldpri = splnet ();
	    {
		for ( scan = st2_retry_q.prevp;
		      scan NE EmptyRetryQ AND new_tod < scan->retry_tod; )
		{
		    scan = scan->prevp;
		}

		pscmp->nextp = scan;
		pscmp->prevp = scan->prevp;
		scan->prevp->nextp = pscmp;
		scan->prevp = pscmp;
	    }
	    (void) splx ( oldpri );
	}

	/* ------------------------------------------------------------ */

	if ( pktp EQ NO_PKTDSCP )
	{
	    BUGSTOP (-,CantGetResrc,0x57706);
	    retcod = CantGetResrc;
	}
	else if ( pscmp->nxthopp->gifcp EQ NO_NETIFP )
	{
	    BUGSTOP (-,InvalidPointer,0x57707);
	    FreePkts (pktp);		/* We must free packet */
	    retcod = IntfcFailure;
	}
	else
	{
	    IfLog (DBG_CTL_MSG)
	    {
		struct aSCMPHeader *scmhdrp = pscmp->scmhdrp;

		(void) printf ( "SndSCMP:%2x %2x %4x RV:%4x %4x Ref:%4x %4x IP:%8x (%4x %8x)\n",
		    scmhdrp->OpCode, scmhdrp->Options, scmhdrp->TotalBytes,
		    scmhdrp->RVLId, scmhdrp->SVLId, scmhdrp->Ref,
		    scmhdrp->LnkRef, scmhdrp->SndrIPAdr,
		    scmhdrp->peropcode2, scmhdrp->peropcode4 );
	    }

	    if ( pscmp->nxthopp->gifcp->ifc_output EQ st2_ToAppl )
	    {
		result = st2_ToAppl( pscmp->nxthopp->gifcp, pktp,
		    (struct sockaddr_st2 *) &( pscmp->nxthopp->lclsockaddr ) );
		retcod = (result EQ 0) ? NoError : IntfcFailure;
	    }
	    else if ( pscmp->nxthopp->gifcp->ifc_output NE (int (*)()) NULL )
	    {
		result = pscmp->nxthopp->gifcp->ifc_output (
				pscmp->nxthopp->gifcp->osifcp, pktp,
				&( pscmp->nxthopp->lclsockaddr ) );
		retcod = (result EQ 0) ? NoError : IntfcFailure;
	    }
	    else
	    {
		FreePkts (pktp); /* We must free packet */
		retcod = Misconfiguration; /* no ifc_output routine */
	    }
	}

	/* Was it sent */
	if ( retcod EQ NoError )
	{
	    /* pkt was sent/held by arp */
	    pscmp->nxthopp->thruput.ctrl_out_pkts++;
	    pscmp->nxthopp->thruput.ctrl_out_bytes += len;
	}
	else
	{
	    pscmp->nxthopp->thruput.ctrl_out_drop_pkts++; /* pkt was dropped */
	    pscmp->nxthopp->thruput.ctrl_out_drop_bytes += len;;
	    BUGSTOP (,retcod,0x57708);
	    /* ??? diagnose error / ? initiate recovery */
	}

	/* ------------------------------------------------------------ */

	if ( pscmp->retryparms.retry_limit EQ 0 )
	{
	    CtrlMsgDel( pscmp );
	}
	else
	    retcod = NoError;		/* to try again */
    }

    /* All retries exhausted, call handler and say no more retries. */
    else
    {
	if ( pscmp->comp NE (void (*)()) NULL )
	{
	    IfLog (DBG_MSG_CMP)
		printf ( "Calling message %8x (op %u) completion routine %8x after reply timeout\n",
			pscmp, pscmp->hoscmphdr.OpCode, pscmp->comp );

		/* CompAccept() CompConnect() CompDisconnect() CompRefuse()  */
	    pscmp->comp ( pscmp, NO_PSCMP /*,RetransTimeout*/ );
	}				/* It CtrlMsgDels pscmp */
	else
	    CtrlMsgDel( pscmp );	/* decrement pcb's use count */

    }

    return ( retcod );
}

/*
S* st2_ctrlinput ( m, pcb, ptype, ifp, lcid )
s*
s*	NOTYET
s*
 */
#ifdef NOTYET
|
|  int
|	_st2_CtrlInput ( m, pcb, ptype, ifp, lcid ) /* from <vector> */
|	  struct mbuf		**m;		/* ARGSUSED */
|	  struct aST2pcb	 *pcb;
|	  int			  ptype;
|	  struct ifnet		 *ifp;
|	  unsigned short	  lcid;
|	{
|	} /* end of st2_CtrlInput */
|	
#endif NOTYET

/*
S* st2_Ioctl ( m, pcb, ptype, ifp, lcid )
s*
s*	NOTYET
s*
 */
#ifdef NOTYET
|	
|	  int
|	_st2_Ioctl ( sop, mp, namp, rightsp )w /* from <vector> */
|	  struct socket		*sop;		/*ARGSUSED*/
|	  struct mbuf		*mp,
|				*namp,
|				*rightsp;
|	{
|	#ifdef  STIIDEBUG
|	    (void) printf ( "st2_ioctl\n" );
|	#endif  STIIDEBUG
|	}
#endif NOTYET

/*
S* st2_PCBDel ( cp )
s*
s*	Called after all protocol activity has ended to reclaims resources
s*	associated with a aST2pcb.
s*	Everything should already have been released by previous protocol
s*	processing, this makes sure in a non-graceful way.
s*	WARNING: cp is no longer valid when this routine returns.
s*
 */

  void					/* Next bugid 0x57814 */
st2_PCBDel( cp )			/* from st2_usrreq CtrlMsgDel HandleConnect*3 st2_DoConnect*3 st2_CMPOutput st2_PCBUpdate*2 */
  struct aST2pcb	*cp;
{
    int			 conhshx,
			 hididx,
			 i,
			 oldpri;
    struct aConHshEntry	*achep = &( st2_ConHsh[0] );
    struct aVLink	*vlinkp;
    struct Target	*targetp;
    unsigned long	*in,
			*out;


    IfLog (DBG_DTA_STR)
	log_Forwarder( cp, "st2_PCBDel" );

    if ( cp EQ NO_PCBP )
    {
	BUGSTOP (-,InvalidPointer,0x57801);
	return;
    }


    if ( cp->stop.tv_sec EQ 0 )
	uniqtime ( &( cp->stop ) );

    cp->Name.pcode = (octet1) 0;	/* stream Name no longer valid */


    /*
     * Next clear the hid table entries.  That will shutdown
     * data forwarding activities.
     */

    cp->hid_ready = FALSE;

    for ( hididx = 0 ; hididx < MAX_HIDS_PER_STREAM ; hididx++ )
    {
	if ( cp->hop_id[ hididx ] NE (HID) 0 )
	{
	    HIDClear( cp->hop_id[ hididx ] );
	    cp->hop_id[ hididx ] = (HID) 0;
	}
    } /* end of for all HIDs loop */



    for ( conhshx = ConfigParm (hid_mask) ; conhshx >= 0 ; conhshx-- )
    {
	if ( achep->cp NE cp )
	    continue;

	BUGSTOP (-,achep->hid,0x57802);

	HIDClear( achep->hid );
    }


    if ( cp->ppcb NE (caddr_t) NULL )
	BUGSTOP (-,Inconsistency,0x57803);


    (void) drop_pending_CtrlMsg( cp );	/* Flush retry_q */


    /* Flush link to socket layer (should already be gone) */
    if ( cp->sop NE (struct socket *) NULL )
    {
	/* ??? ought to be in api c */
	oldpri = splnet ();

	if ( cp->sop NE (struct socket *) NULL )
	{
	    cp->sop->so_error = ECONNABORTED; /* ??? or ECONNRESET by peer */
	    soisdisconnected ( cp->sop ); /* wakeup upper layer */

	    cp->sop->so_pcb = (caddr_t) NULL;
	    cp->sop = (struct socket *) NULL;
	    cp->refcnt--;
	    IfLog (DBG_CNT_PCB)
		printf ( "refcnt: cp = %8x  refcnt = %2d  st2_PCBDel       RefSocket\n",
			 cp, cp->refcnt );
	    cp->refwho &= ~ RefSocket;

	    cp->flags &= ~ LocalConnection;
	}

	(void) splx ( oldpri );
    }


    /* Three cases depending on whether an origin (if_gwcp EQ if_cp) */
    /* or target (if_gwcp NE if_cp) or a listener/gateway (if_gwcp EQ NULL) */

    if (  cp->gifcp NE NO_NETIFP )
    {
	struct aST2pcb *wcp;
	struct aNetIf *ifst2p;

	oldpri = splnet ();

	ifst2p = cp->gifcp;
	/* May be more than one if 1sock mode */
	while ( ifst2p NE NO_NETIFP )
	{
	    wcp = ifst2p->ifc_gwcp;
	    if ( wcp NE NO_PCBP )
	    {
		if ( wcp EQ cp )	/* Origin */
		{
		    if ( cp->prvhop.gifcp EQ (struct aNetIf *) ifst2p )
			cp->prvhop.gifcp = NO_NETIFP;
		    else
			BUGSTOP (-,Inconsistency,0x57804);

		    cp->refcnt--;
		    IfLog (DBG_CNT_PCB)
			printf ( "refcnt: cp = %8x  refcnt = %2d  st2_PCBDel       RefGifcgwcpo\n",
				cp, cp->refcnt );
		    cp->refwho &= ~ RefGifcgwcpo;
		}
		else			/* Target for wcp */
		{
		    /* Update target and aNxtHop state in "gateway" pcb */
		    struct Target *wtp = wcp->targetsp;
		    struct aNxtHop *anhp;


		    for ( ; wtp NE NO_TRGP ; wtp = wtp->nextp )
		    {
			if (   (wtp->vlinkp EQ NO_VLINKP)
			    OR (wtp->vlinkp->next_hopp EQ NO_NEXTHOPP)
			    OR (   (anhp = wtp->vlinkp->next_hopp->route.nxthopp)
				EQ NO_NXTHOPP)
			    OR (anhp->gifcp NE (struct aNetIf *) ifst2p) )
			{
			    continue;
			}

			deactivate_aNxtHop (anhp);
			ifst2p->osifcp->if_flags &= ~ IFF_UP;

			ifst2p->ifc_gwcp->refcnt--;
			IfLog (DBG_CNT_PCB)
			    printf ( "refcnt: cp = %8x  refcnt = %2d  st2_PCBDel       RefGifcgwcpt\n",
				     ifst2p->ifc_gwcp,
				     ifst2p->ifc_gwcp->refcnt );
			ifst2p->ifc_gwcp->refwho &= ~ RefGifcgwcpt;
			ifst2p->ifc_gwcp = NO_PCBP;
			anhp->gifcp = NO_NETIFP;

			if ( NOT IsTargState (wtp,TS_BUSY) )
			    break;

			if ( NOT IsTargStateIn (wtp,TS_RELEASED,
						TS (TS_CLOSED)) )
			{
			    SetTargState (wtp,TS_CLOSED);
			    wtp->cp->needupdate++; /* Something to do */
			}

			if ( NOT IsTargStateIn (wtp,TS_DISCONNECTED,
					TS (TS_REFUSING) TS (TS_FAILED)) )
			{
			    wtp->error = cp->fwdrerror;
			    wtp->pending_reference = (Reference) 0;
					/* ??? obsolete */
			    SetTargState (wtp,TS_FAILED);
			    wtp->cp->needupdate++; /* Something to do */
			}

			if ( wtp->cp->needupdate > 0 )
			    st2_PCBUpdate( wtp->cp );

			break;		/* out of for loop */

		    } /* end of searching for target loop */

		    if ( wtp EQ NO_TRGP ) /* Not found */
			BUGSTOP (-,Inconsistency,0x57805);
		} /* end of else target */
	    }
	    /* else listener or pure gateway */

	    ifst2p->ifc_cp = NO_PCBP;
	    cp->refcnt--;
	    IfLog (DBG_CNT_PCB)
		printf ( "refcnt: cp = %8x  refcnt = %2d  st2_PCBDel       RefGifccp\n",
			cp, cp->refcnt );
	    cp->refwho &= ~ RefGifccp;

	    cp->gifcp = ifst2p->ifc_nextp;
	    /* shouldn't be linked until 1sock support */
	    FreePkts (ifst2p->ifc_selfp);
	    ifst2p = cp->gifcp;

	} /* end of while st2 ifs loop */

	(void) splx ( oldpri );
    }


    RemRing (nxtNamep,
	    IfLog (DBG_INI_DS_MSG)
		log_Stream( cp, "st2_PCBDel" );
	    cp->Name.pcode = (octet1) 0;
	    /* notify other meanders ??? */
    );


    RemRing (fdxp,
	    /* notify other direction of fwdr error ??? */
    );


    RemRing (nxtGroupp,			/* ??? multiple groups */
	    /* notify other groups ??? */
    );


    /* flow_spec, rflow_spec not currently dynamic */


    if ( cp->accepts NE NO_PSCMP )
    {
	struct aPrsSCM *pscmp, *listp;

	oldpri = splnet ();

	listp = cp->accepts;
	cp->accepts = NO_PSCMP;

	(void) splx ( oldpri );

	while ( (pscmp = listp) NE NO_PSCMP )
	{
	    listp = pscmp->nextp;
	    pscmp->nextp = NO_PSCMP;

	    CtrlMsgDel( pscmp );

	} /* end of while */
    }


    /* ??? Origin has a count here for the ApplIfc */
    oldpri = splnet ();
    {
	vlinkp = cp->upstream_vlinkp;
	cp->upstream_vlinkp = NO_VLINKP;
    }
    (void) splx ( oldpri );

    if ( vlinkp NE NO_VLINKP )
    {
	/* Accounting */

	(void) st2_RsrcAccount( cp->prvhop.gifcp, (unsigned long) 0 /*msec*/,
			        &( cp->prvhop.thruput ), &( cp->cost ) );

	out = (unsigned long *) &( cp->thruput );
	in  = (unsigned long *) &( cp->prvhop.thruput );
	i   = sizeof (struct aThruput) / sizeof (unsigned long);

	for ( ; i > 0 ; i-- ) { *out++ += *in; *in++ = 0; }

	VLinkDel( vlinkp );
    }

    /* ??? all targets should be TS_DEAD by now */

    /* Delete target entries (and their data can control links). */
    ForAllTargets (cp, targetp)
    {
	if ( IsTargState (targetp,TS_DEAD) )
	    continue;

	BUGSTOP (-,PackTState (targetp,TS_DEAD),0x57806);

	IfLog (DBG_INI_DS_MSG)
	    log_Target( targetp, "st2_PCBDel: target not TS_DEAD" );

	/* These are not required ... */
	if ( NOT IsTargState (targetp,TS_RELEASED) )
	{
	    SetTargState (targetp,TS_RELEASED);
	}

	if ( NOT IsTargState (targetp,TS_DISCONNECTED) )
	{
	    SetTargState (targetp,TS_DISCONNECTED);
	}

	targetp->stop = cp->stop;

	(void) TargDel( targetp );
	/* which does a VLinkDel */
	/* which does a NextHopDel */
	/* and a st2_VLinkReset */

    } /* end of for all targets loop */

    /* fwdrfreespace holsd target memory */
    if ( cp->fwdrfreespace NE NO_PKTDSCP )
    {
	cp->targetsp = NO_TRGP;
	cp->targettailp = NO_TRGP;

	FreePkts (cp->fwdrfreespace);
	cp->fwdrfreespace = NO_PKTDSCP;
    }


    /* Data links should have all been deleted as side effect. */
    while ( cp->next_hopsp NE NO_NEXTHOPP )
    {
	BUGSTOP (-,cp->next_hopsp,0x57807);

	(void) NextHopDel( cp->next_hopsp );
    }

    if ( cp->nxthopsp NE NO_NHLISTP )
	cp->nxthopsp = NO_NHLISTP;

    if ( cp->nxthopbufp NE NO_PKTDSCP )
    {
	FreePkts (cp->nxthopbufp);
	cp->nxthopbufp = NO_PKTDSCP;
    }

    if ( cp->trashp NE NO_PKTDSCP )
    {
	FreePkts (cp->trashp);
	cp->trashp = NO_PKTDSCP;
    }


    if ( cp->refcnt > 0 )
    {
	if ( (cp->refwho & RefSelf) EQ 0 )
	{
	    BUGSTOP (-,cp->refcnt,0x57808);
	    IfLog (DBG_CNT_PCB)
		printf ( "refcnt: cp = %8x  refcnt = %2d  st2_PCBDel       [who=%8x] ought to have one for self\n",
			 cp, cp->refcnt, cp->refwho );
	}

	cp->refcnt--;
	IfLog (DBG_CNT_PCB)
	    printf ( "refcnt: cp = %8x  refcnt = %2d  st2_PCBDel       RefSelf\n",
		     cp, cp->refcnt );
	cp->refwho &= ~ RefSelf;

	if ( cp->refcnt > 0 )
	{
	/*  BUGSTOP (,(cp->refwho | cp->refcnt),0x57809); /* ??? temp */
	    cp->refwho |= RefLastPrsSCM;
	    return;			/* wait for CtrlMsgDel */
	}
    }
    else
    {
	if ( (cp->refwho & RefLastPrsSCM) NE 0 )
	{
	/*  BUGSTOP (,(cp->refwho | cp->refcnt),0x57810); /* ??? temp */
	}
	else
	{
	    BUGSTOP (,cp->refcnt,0x57811); /* ??? temp */
	}
	cp->refwho &= ~ RefLastPrsSCM;
    }

    if ( (cp->refcnt > 0)  OR  (cp->refwho NE 0) )
    {
	BUGSTOP (-,cp->refcnt,0x57812); /* ??? Now is this possible */
    }

    if ( cp->refcnt <= 0 )
    {
	oldpri = splnet ();
	{
	    if (   IsState (PCB,cp,pcb_state,CINS_POSTACC)
		OR IsState (PCB,cp,pcb_state,CINS_free)
		OR (cp->prevp EQ NO_PCBP) )
	    {
		(void) splx ( oldpri );
		BUGSTOP (-,Inconsistency,0x57813); /* Already freed */
		return;
	    }

	    IfLog (DBG_CNT_PCB)
		printf ( "refcnt: cp = %8x  refcnt = %2d  st2_PCBDel       freeing it\n",
			cp, cp->refcnt );

	    cp->nextp->prevp = cp->prevp;
	    cp->prevp->nextp = cp->nextp;
	    cp->nextp = NO_PCBP;
	    cp->prevp = NO_PCBP;

	    SetState (FWD,cp,fwdr_state,FW_FREE);
	    cp->flags &= ~ ConnActive;
	    SetState (PCB,cp,pcb_state,CINS_POSTACC);

#ifdef NOTYET
|		    /* Queue it for post analysis/accounting */
|		    cp->nextp = st2_acctp;
|		    st2_acctp = cp;
#else
	    /* ------------------------------------------------------------ */
	    /* Update global stats */

#define aCostVar(name,format,desc) \
	    Count (Ident (cost_)name,cp->cost.name);
	    CostList
#undef  aCostVar

	    out = (unsigned long *) &( st2_stats.ctrl_in_bytes ),
	    in  = (unsigned long *) &( cp->thruput );
	    i = sizeof (struct aThruput) / sizeof (unsigned long);
	    for ( ; i > 0 ; i-- ) { *out++ += *in; *in++ = 0; }

	    /* ------------------------------------------------------------ */

	    SetState (PCB,cp,pcb_state,CINS_free);

	    if ( st2_ConTblTailp EQ NO_PCBP )
		st2_ConTblFreep = cp;
	    else
		st2_ConTblTailp->nextp = cp;

	    st2_ConTblTailp = cp;

	    Count (pcb_rel,1);
	    IfLog (DBG_DTA_STR)
		printf ( "st2_PCBDel aST2pcb at 0x%x deleted\n", cp );
#endif NOTYET
	}
	(void) splx ( oldpri );
    }

    return;
}

/*
S* st2_PCBGet ()
 */

  struct aST2pcb *			/* Next bugid 0x57900 */
st2_PCBGet()				/* from st2_usrreq HandleConnect */
{
    int			  oldpri;
    struct aST2pcb	 *cp = NO_PCBP;


    oldpri = splnet ();

    cp = st2_ConTblFreep;
    if ( cp NE NO_PCBP )
    {
        st2_ConTblFreep = cp->nextp;

#ifdef STIIDEBUG
	kassert ((&( st2_ConTbl[0] ) <= cp)
		 AND (cp < &( st2_ConTbl[ ConfigParm (max_streams) ] )));
	kassert (cp->refcnt EQ 0);
#endif STIIDEBUG

	Bzero (cp,sizeof (struct aST2pcb));

	cp->nextp = &( st2_ConTbl[0] );
	cp->prevp = st2_ConTbl[0].prevp;
	st2_ConTbl[0].prevp->nextp = cp;
	st2_ConTbl[0].prevp = cp;

	IfLog (DBG_DTA_STR)
	    printf ( "st2_PCBGet at 0x%x\n", cp );

	Count (pcb_get,1);
	uniqtime ( &( cp->start ) );
	cp->dspin = st2_Forward;
	cp->optmask = (unsigned char) STDataPriority;
	cp->refcnt = 1;
	cp->refwho |= RefSelf;
	IfLog (DBG_CNT_PCB)
	    printf ( "refcnt: cp = %8x  refcnt = %2d  st2_PCBGet       RefSelf\n",
		    cp, cp->refcnt );
	SetState (PCB,cp,pcb_state,CINS_PREOPEN);
	SetState (FWD,cp,fwdr_state,FW_BUSY);
    }
    else
    {
	Count (pcb_fail,1);
        Count (nofreecon,1);		/* count no free connection blocks */
	/* ??? bug return CantGetResrc */
    }

    (void) splx ( oldpri );

    return ( cp );
}

/*
S* st2_PCBUpdate ( cp )
s*
s*	Routing to advance the state of the connection or its targets,
s*	by moving each target to its next state.
s*	It needs a bit of work.  BEWARE: cp may be invalid on return.
**
 */

  void					/* Next bugid 0x58013 */
st2_PCBUpdate( cp )			/* from api_control CompAccept CompConnect CompDisconnect CompRefuse HandleConnect HandleDisconnect HandleRefuse st2_CMPInput st2_DoConnect st2_DoDisconnect st2_PCBDel */
  struct aST2pcb	*cp;		/* 	non-0		non-0 */
{
    boolean		 direct,
			 here,		/* First srcrut IPAdr is one of ours */
    			 loop;		/* Dropped first parm, back for next */
    unsigned char	*sap;
    enum PCodes		 srpc;
    enum ST2Errors	 error,
			 reason;
    INETADDR		 lclipadr,
			 tipadr;
    int			 i,
			 lenrt,
			 lentg,
			 mismatch,
			 n,
			 oldpri,
			 plen,
			 srplen;
    static int		 recurse = 0;
    unsigned int	 route_offset,	/* Bytes to active srcrut */
			 slen,		/* Length of a srcrut parm */
			 tlen;		/* Length of target */
    struct aApplEntity	*entp;
    struct aRouteChoice	*arcp,
			*warcp;
    struct aSrcRut	*asrp,		/* Ptr to active srcrut parm */
			*asrrtp,
			*asrtgp,
			*wsrp;
    struct aST2pcb	*wcp;
    struct aTarget	*atp,
			*tparminp;
    struct aVLink	*vlinkp;
    struct NextHop	*next_hopp,
			*tailp;
    struct STSAP	 entsap;
    struct Target	*targetp,
			*wtp;


    PTRCHECK (cp, struct aST2pcb *, return);

    if (   IsState (FWD,cp,fwdr_state,FW_FREE)
	OR (cp->needupdate EQ 0) )
    {
	return;
    }

    if ( IsState (PCB,cp,pcb_state,CINS_LISTEN) )
    {
	if ( cp->sop EQ (struct socket *) NULL )
	{
	    st2_PCBDel( cp );
	    /* cp may be invalid */
	}

	return;
    }

    while ( cp->needupdate NE 0 )
    {
	/* Advance state machine for each target */

	cp->needupdate = 0;
	ForAllTargets (cp, targetp)
	{
	    if ( NOT IsTargState (targetp, TS_BUSY) )
		continue;

	    IfLog (DBG_UPDTARG)
		log_Target( targetp, "update_Target-starting" );

	    /* ============================================================ */
	    /* Work the downstream state machine */

	    switch ( targetp->targstate[ (int) Nexthop ] )
	    {
	    case TS_NEW:		/* New targets need a route */
		/* We can get flowspec from target for type of service routing.
		 * But maybe there's a better way to handle the abstraction.
		 */
		/* ??? Should take flow spec as argument and share pointer only
		 * if they match.  Otherwise grab some memory and make a copy.
		 * Free on TargDel or keep as part of forwarder trashp list?
		 */
		targetp->flow_specp = (struct aFlowSpec *)
					&( targetp->cp->flow_spec );


		/* Setup for route lookup: process source route parameters */

		direct = FALSE;		/* Strict route to next entity */
		tparminp = targetp->tparminp;
		tipadr = ntohIP (tparminp->IPAdr);

		/* i.e., ntoh1 (targetp->tparminp->TargetBytes) */
		tlen = targetp->tparmin_bytes;

		targetp->targetIPAdr = tipadr; /* Headed toward target */
		targetp->nxtSTIPAdr = tipadr;
		targetp->routetoIPAdr = tipadr;

		targetp->sap.NextPcol = targetp->cp->Origin.NextPcol;
		targetp->sap.SAPBytes = ntoh1 (tparminp->SAPBytes);
		targetp->sap.SAPp     = &( tparminp->SAP[0] );

		/* The offset to the source route portion is 4 for the IP
		 * address, plus one for the TargetBytes plus one for SAPBytes
		 * plus SAPBytes then rounded to a 4 byte boundary.
		 */
		/* vs targetp->prefixlen */
		route_offset = ROUND (  (OffsetOf (SAP[0],struct aTarget)
				+ ntoh1 (tparminp->SAPBytes)), octet4);
		targetp->prefixlen = route_offset;
		targetp->dropfirst = 0;	/* Not first IPAdr in srcrut parm */

		IfLog (DBG_SKT_RUT)
		{
		    log_Target( targetp, "update_Target/FindRoutes" );
		    (void) printf ("Target on %x", targetp->targetIPAdr);
		    log_SAP( &( targetp->sap ), "");
		}

		tlen -= route_offset;	/* Remaining bytes in target param */
		loop = FALSE;

  nxtsrcrut:
		if ( tlen > 0 )
		{
		    IfLog (DBG_SRC_RUT)
			(void) printf ( "Target has source route (offset %d, tbytes %d)\n",
				        route_offset, tlen );

		    asrp = Mkp (struct aSrcRut *,tparminp,route_offset);
		    slen = ntoh1 (asrp->plen);
		    tipadr = ntohIP (asrp->IPAdr[0]);
		    here = st2_IPAdrFunc( IPAdrFuncValidate, &( tipadr ) );

		    switch ( ntoh1 (asrp->pcode) )
		    {
			/* ??? Should be error, but it can't happen since
			 * the parser ensures we only get nice, validly
			 * formatted parameters.  But there's applications ...
			 * ??? Source Route error or target format error
			 */
		    default:
			error = PCodeUnknown;
			continue;	/* on to next target */

		    case STpIPSSrcRut:	/* 17 */
			direct = TRUE;
		    case STpIPLSrcRut:	/* 16 */
			if ( here )
			{
			    /* perv-hop should have moved it to IP */
			    BUGSTOP (-,SourceRouteError,0x58001);
			    continue;	/* on to next target */
			}
			targetp->iphop = 1; /* Last IPAdr is next ST agent */
			targetp->nxtSTIPAdr = ntohIP (asrp->IPAdr[
					(slen - sizeof (struct aSrcRut)) / 4]);
			/* First IPAdr is IP entity */
			targetp->routetoIPAdr = tipadr;
			break;

		    case STpSTSSrcRut:	/* 19 */
			if (    NOT here AND NOT loop /* Not "strict" */
			    AND ((targetp->cp->flags & LocalOrigin) EQ 0) )
			{
			    BUGSTOP (-,SourceRouteError,0x58002);
			    continue;	/* on to next target */
			}

			direct = TRUE;
			/* Fall through */

		    case STpSTLSrcRut:	/* 18 */
			if ( here )
			{
			    if ( slen <= sizeof (struct aSrcRut) )
			    {		/* End of srcrut parameter */
				tlen -= slen; /* Drop completed srcrut parm */
				route_offset += slen;
				direct = FALSE;
				loop = TRUE;
				goto nxtsrcrut; /* Maybe another */
			    }

			    targetp->dropfirst = 1;
			}
			tipadr = ntohIP (asrp->IPAdr[targetp->dropfirst]);
			targetp->nxtSTIPAdr = tipadr;
			targetp->routetoIPAdr = tipadr;
			break;

		    } /* end of switch */

		    targetp->targsrp = asrp; /* NB: incl. leading IP SrcRut */

		} /* have source route */

		targetp->targsrlen = tlen;

		/* ======================================================== */

		arcp = &( targetp->routes.RouteChoices[0] );
		targetp->curchoicep = NO_RUTCHOICEP;

		/* Find (all) routes to target */
		st2_FindRoutes( targetp->routetoIPAdr, targetp->nxtSTIPAdr,
			        &( targetp->routes ),
			        ((targetp->iphop NE 0) ? FRF_USEIP : 0)
			        | ((direct EQ TRUE) ? FRF_DIRECT : 0) );

		if ( arcp->type EQ TO_APPL )
		{
		    mismatch = RouteNextPcol;
		    IfLog (DBG_BND_SKT)
			log_SAP( &( targetp->sap ), "Look for socket" );

		    ForAllpcbs (wcp)	/* ??? lock out API */
		    {
			/* ??? really if listening-bit set in state (1sock) */
			if ( NOT IsState (PCB,wcp,pcb_state,CINS_LISTEN) )
			    continue;

			entp = (struct aApplEntity *) &( wcp->soname.applent );

			IfLog (DBG_SKT_RUT)
			{
			    /* APIEQ */
			    if ( wcp->soname.applent.pcode NE (octet1) 0 )
			    {
				entsap.SAPp = (unsigned char *)
						&( entp->SAP[0] );
				entsap.SAPBytes = entp->SAPBytes;
				entsap.NextPcol = entp->NextPcol;
				log_SAP( &( entsap ), "... compare to " );
			    }
			    else
				(void) printf ( "Skipping unbound socket %x\n",
					        wcp );
			}

			/* Skip sockets that aren't bound yet. */
			if ( wcp->soname.applent.pcode EQ (octet1) 0 )
			    continue;

			/* Protocols must match. */
			if ( targetp->sap.NextPcol NE entp->NextPcol )
			    continue;

			mismatch = RouteSAP; /* have matching protocol */

			if (   (targetp->sap.SAPBytes EQ 0)
			    OR (entp->SAPBytes EQ 0) )
			    ;
			else if (    (targetp->sap.SAPBytes EQ entp->SAPBytes)
				 AND (Bcmp (targetp->sap.SAPp,&( entp->SAP[0] ),
					    targetp->sap.SAPBytes) EQ 0) )
			    ;
			else
			    continue;

			/* IPAdrs must not differ */
			if ( entp->IPAdr NE INADDR_ANY )
			{
			    if ( entp->IPAdr NE targetp->targetIPAdr )
				continue;
			}

			/* ................................................ */
			/* Logically breaking out of loop ... */

			IfLog (DBG_SKT_RUT)
			    (void) printf ( "...found listening %s socket for cp = %x\n",
					        (wcp->soname.st_options
					      & (int) STOpt1socket)
					    ? "shared" : "normal", wcp );

			wcp = st2_ApplIfcCreate( wcp, arcp ); /* New socket */

			break;		/* out of for all pcb loop */

		    } /* end of for all pcbs loop */


		    if ( arcp->gifcp EQ NO_NETIFP )
		    {
			if ( arcp->error EQ NoError )
			{
			    if ( mismatch EQ RouteNextPcol )
			    {
				arcp->flags |= RouteFailed | RouteNextPcol;
				arcp->error = PcolIdUnknown;
			    }
			    else
			    {
				arcp->flags |= RouteFailed | RouteSAP;
				arcp->error = SAPUnknown;
			    }
			}
			targetp->routes.nxtfree = 0; /* failed */
		    }

		    IfLog (DBG_SKT_RUT)
		    {
			sap = targetp->sap.SAPp;

			if ( arcp->gifcp EQ NO_NETIFP )
			    (void) printf ( "...failed to find listening socket\n" );

			(void) printf ( "Local route to %x, sap = (",
				        targetp->targetIPAdr );
			for ( n = 0 ; n < targetp->sap.SAPBytes ; n++ )
			    (void) printf ( "%s%x", n?" ":"", *sap++ );
			(void) printf ( ")\n" );
		    }
		} /* end of TO_APPL */

		targetp->cp->needupdate++;
		SetTargState (targetp, TS_ROUTEOPT);

		break;			/* to let other targets catch up */


	    case TS_ROUTEOPT:

		SetTargState (targetp,TS_ROUTING);

		if ( st2_RouteOptimizep NE (void (*)()) NULL )
		{
		    (*st2_RouteOptimizep) ( targetp->cp );

		    targetp->cp->needupdate++;
		    break;		/* to let other targets catch up */
		}

		IfLog (DBG_UPDTARG)
		    log_Target( targetp, "update_Target-starting route search" );
		/* Falls through */


	    case TS_ROUTING:
		/* If no routes were found, target fails at this hop */
		if ( targetp->routes.nxtfree EQ 0 )
		{
		    targetp->error = targetp->routes.RouteChoices[0].error;
		    /* (if targetp->src_route) w/NoRouteToAgent : w/NoRouteToDest */

		    SetTargState (targetp, TS_RELEASED); /* No resources */
		    if ( IsTargState (targetp, TS_CONNECTED) )
			SetTargState (targetp, TS_FAILED); /* Run other side */
		    /* not need, FAILED processed below */
		    /* targetp->cp->needupdate++; */

		    break;		/* out of routing case */
		}

  reroute:
		/* Find untried route with lowest "order" */
		arcp = NO_RUTCHOICEP;
		warcp = &( targetp->routes.RouteChoices[0] );
		for ( i = targetp->routes.nxtfree ; i > 0 ; i--, warcp++ )
		{
		    if ( (warcp->flags & RouteFailed) NE 0 )
			continue;	/* Skip previously tried routes */

		    if ( arcp EQ NO_RUTCHOICEP )
		    {
			arcp = warcp;
			continue;	/* First untried route */
		    }

		    if ( warcp->order < arcp->order )
			arcp = warcp;	/* Route with higher priority */

		} /* end of searching for next to try */

		if ( arcp EQ NO_RUTCHOICEP )
		{
		    /* tried one so no resources, unless result of failure */
		    targetp->error = error; /* ??? multiple route case */

		    SetTargState (targetp, TS_RELEASED);
		    if ( IsTargState (targetp, TS_CONNECTED) )
			SetTargState (targetp, TS_FAILED); /* Run other side */
		    targetp->cp->needupdate++;

		    break;		/* out of routing case */
		}

		targetp->curchoicep = arcp; /* Trying out this choice */

		/* -------------------------------------------------------- */
		/* Construct outgoing target parm for selected route choice */

		atp = targetp->tparmoutp; /* Pre-reserved space */
		i = targetp->prefixlen;	/* Less any source routing */
		Bcopy (targetp->tparminp,atp,i);
		atp->TargetBytes = hton1 (i);
		targetp->ipsrcrutp = NO_SRP;

		if (   (targetp->targsrlen > 0)
		    OR (arcp->rtblsrlen > 0) )
		{			/* Have Source Route */
		    asrrtp = arcp->rtblsrp;
		    lenrt = arcp->rtblsrlen;
		    asrtgp = targetp->targsrp;
		    lentg = targetp->targsrlen;
		    plen = i;
		    srplen = 0;
		    srpc = (enum PCodes) 0;

		    /* Have two levels of source routing to deal with.
		     * "tg" level was specified in received target
		     * parameter.  "rt" level was specified in local
		     * routing table.  Combining both levels results
		     * in "rt" level followed by "tg" level, and maybe
		     * an IP Source Route Option to reach the next ST
		     * Agent.
		     */

		    wsrp = Mkp (struct aSrcRut *,atp,targetp->prefixlen);

		    if ( lenrt > 0 )	/* NOTYET ??? */
		    {
			if (   (   (enum PCodes) asrrtp->pcode
				EQ hton1 (STpIPLSrcRut))
			    OR (   (enum PCodes) asrrtp->pcode
				EQ hton1 (STpIPSSrcRut)) )
			{
			    /* IP Routes handled in nxthop */
			    targetp->ipsrcrutp = asrrtp;
			    lenrt -= ntoh1 (asrrtp->plen);
			    asrrtp = Mkp (struct aSrcRut *,asrrtp,
					  ntoh1 (asrrtp->plen));

			    /* Route Tbl only has one IP srcrut */
			    if (    (lenrt <= 0)
				AND (lentg > 0)
				AND (   (   (enum PCodes) asrtgp->pcode
					 EQ hton1 (STpIPLSrcRut))
				     OR (   (enum PCodes) asrtgp->pcode
					 EQ hton1 (STpIPSSrcRut)) ) )
			    {
				/* ??? trouble, concatinated IP srcruts */
				BUGSTOP (-,targetp,0x58003);
			    }
			}

			if ( lenrt > 0 )
			{
			    Bcopy (asrrtp,wsrp,lenrt);
			    wsrp = Mkp (struct aSrcRut *,wsrp,lenrt);
			}
		    } /* end of route table has srcrut */

		    /* .................................................... */
		    /* Target has ST SrcRut & this ST Agent is first entry */

		    if ( targetp->dropfirst )
		    {
			IfLog (DBG_SRC_RUT)
			    (void) printf ( "Contracting source route parameter from %d bytes\n",
					    lentg );

			if ( ntoh1 (asrtgp->plen) <= sizeof (struct aSrcRut) )
			{		/* I.e., minimal size */
			    lentg -= ntoh1 (asrtgp->plen); /* Drop it */
			    asrtgp = Mkp (struct aSrcRut *,asrtgp,
					  ntoh1 (asrtgp->plen));
			}
			else		/* remove first entry */
			{
			    srpc = (enum PCodes) asrtgp->pcode;
			    srplen = ntoh1 (asrtgp->plen) - 4;
			    lentg -= 4;
			    asrtgp = Mkp (struct aSrcRut *,asrtgp,4);
			    /* NB: this doesn't really point to aSrcRut */
			}
		    }

		    /* .................................................... */
		    /* If target now has initial IP Source Route, stash it */

		    if (    (lentg > 0)
			AND (targetp->ipsrcrutp EQ NO_SRP)
			AND (   (   (enum PCodes) asrtgp->pcode
				 EQ ntoh1 (STpIPLSrcRut))
			     OR (   (enum PCodes) asrtgp->pcode
				 EQ ntoh1 (STpIPSSrcRut))) )
		    {
			/* IP Routes handled in nxthop */
			targetp->ipsrcrutp = asrtgp; /* IP SrcRut */
			lentg -= ntoh1 (asrtgp->plen);	/* Flush IP */
			asrtgp = Mkp (struct aSrcRut *,asrtgp,
				      ntoh1 (asrtgp->plen));
		    }

		    /* .................................................... */
		    /* Copy srcruts from target */

		    if ( lentg > 0 )
		    {
			Bcopy (asrtgp,wsrp,lentg);

			if ( srpc NE (enum PCodes) 0 )
			{		/* Dropped first address above */
			    wsrp->pcode = (octet1) srpc;
			    wsrp->plen = hton1 (srplen);
			    wsrp->fill2 = 0;
			}
			wsrp = Mkp (struct aSrcRut *,wsrp,lentg);
		    }

		    /* .................................................... */

		    if ( lenrt + lentg > 0 )
			IfLog (DBG_SRC_RUT)
			    (void) printf ( "Final source route parameters %d bytes\n",
					    lenrt + lentg );
		    plen += lenrt + lentg;
		    atp->TargetBytes = hton1 (plen);

		    if ( plen NE ((char *) wsrp - (char *) atp) )
			BUGSTOP (,(plen - ((char *) wsrp - (char *) atp)),0x58004);

		} /* end of processing source routing options */

		/* -------------------------------------------------------- */
		/* Now start processing next hop & its state machine */

		/* Must then wait for next hop's machine to finish */
		/* Up front to avoid race */
		SetTargState (targetp, TS_WAITNHOP);

		if ( arcp->type NE TO_APPL )
		{
		    /* Check to see if a virtual link for an existing target
		     * goes to the same next-hop ST Agent.  If it does, try to
		     * use it.  If no virtual link to that agent exists,
		     * create a new link.  For the new link check to see if
		     * there is a nexthop link that can be shared--if one
		     * exists, that means it is a multicast group.
		     */

		    /* Scan through target table to see if any of the control
		     * links connect to nxtSTIPAdr next-hop agent.
		     */

		    ForAllTargets (targetp->cp, wtp)
		    {
			if ( NOT IsTargState (wtp, TS_BUSY) )
			    continue;

			/* skip ourself & targets without a selected route */
			if (   (wtp EQ targetp)
			    OR (wtp->curchoicep EQ NO_RUTCHOICEP) )
			    continue;

			/* Matching routes must have same non-local next ST
			 * agent
			 */
			if (   (   wtp->curchoicep->nxtSTIPAdr
				NE arcp->nxtSTIPAdr)
			    OR (wtp->curchoicep->type EQ TO_APPL) )
			    continue;

			/* Must match IP Source Route, too */
			if ( targetp->ipsrcrutp EQ NO_SRP )
			{
			    if ( wtp->ipsrcrutp NE NO_SRP )
				continue;
			}
			else
			{
			    if (   (wtp->ipsrcrutp EQ NO_SRP)
				OR (   targetp->ipsrcrutp->plen
				    NE wtp->ipsrcrutp->plen) )
				continue;

			    if ( Bcmp (targetp->ipsrcrutp,wtp->ipsrcrutp,
				       ntoh1 (targetp->ipsrcrutp->plen))
				 NE 0 )
				continue;
			}

			if ( wtp->vlinkp NE NO_VLINKP )
			{
			    /* Increment the number of targets that this
			     * control link points to.
			     */

			    { /* lock virtual link */
				kassert (wtp->vlinkp->vlink_cnt > 0);
				wtp->vlinkp->vlink_cnt++;
			    } /* unlock virtual link */

			    targetp->vlinkp = wtp->vlinkp;

			    /* If other target is waiting we wait, too,
			     * otherwise continue to open
			     */
			    /* Could check nexthop state ... */

			    if ( IsTargState (wtp, TS_WAITNHOP) )
				;	/* already in TS_WAITNHOP */
			    else	/* ??? what if dead/dying */
			    {
#ifdef NOTYET				/* ??? want this, generic */
|				Bcopy (&( wtp->cp->flow_spec ),
|				       &( targetp->cp->flow_spec ),
|				       sizeof (struct aFlowSpec3));
#endif NOTYET
				SetTargState (targetp, TS_WANTOPEN);
				targetp->cp->needupdate++; /* more to do */
			    }

			    break;	/* out of for all targets loop */
			}

			/* Good match but no existing virtual link; will create
			 * it
			 */

		    } /* end of for all targets loop */
		} /* end of not TO_APPL */

		/* -------------------------------------------------------- */

		if ( targetp->vlinkp NE NO_VLINKP )
		{
		    IfLog (DBG_UPDTARG)
		    {
		    /*	log_Target ( targetp, "update_Target-share link" ); */
			log_Target( targetp, "update_Target-routing succeeded" );
		    }

		    break; /* out of routing case to let other targets get
			    * resources for batched CONNECT */
		}

		/* -------------------------------------------------------- */
		/* Need a new virtual link & nexthop */

		vlinkp = VLinkNew( targetp->cp, DOWNSTREAM_VLINK,
				   (arcp->type EQ TO_APPL)
				   ? APPL_VLINK : GATEWAY_VLINK );
		if ( vlinkp EQ NO_VLINKP )
		{
		    targetp->error = CantGetResrc;
		    BUGSTOP (-,CantGetResrc,0x58005);

		    if ( IsTargState (targetp, TS_CONNECTED) )
			SetTargState (targetp, TS_FAILED); /* Run other side */
		    SetTargState (targetp, TS_RELEASED); /* No resources */
		    targetp->cp->needupdate++;
		    break; /* out of routing case to let other targets get
			    * resources for batched CONNECT */
		}

		vlinkp->lclIPAdr = arcp->lclIPAdr;
		vlinkp->rmtIPAdr = arcp->nxtSTIPAdr;
		targetp->vlinkp = vlinkp;

#ifdef MULITCAST_IDEA
|		vlinkp->next_hopp = NO_NEXTHOPP; /* share_next_hop_Forwarder */
|	
|		/* scan through target table to see if any of the
|		 * data links connect us to the requested host. */
|	
|		/* check the state of the target */
|		ForAllTargets (targetp->cp, wtp)
|		{
|		    if ( NOT IsTargState (targetp, TS_BUSY) )
|			continue;
|	
|		    if (    (targetp NE wtp)
|			AND (! IsTargStateIn (wtp,TS_NEW,TS (TS_NEW) TS (TS_DEAD)))
|	
|			/* Matching route sets must have same non-local next agent */
|			AND (targetp->curchoicep->nxtIPAdr EQ wtp->curchoicep->nxtIPAdr)
|			AND (targetp->curchoicep->curtype NE TO_APPL)
|			AND (wtp->curchoicep->curtype NE TO_APPL) )
|		    {
|	
|			PTRCHECK (wtp->vlinkp->next_hopp, struct NextHop *, break;); /* ??? or continue */
|			/* lock next_hop */
|			{
|			    kassert (wtp->vlinkp->next_hopp->vlink_cnt >= 0);
|			    wtp->vlinkp->next_hopp->vlink_cnt++;
|			} /*  unlock */
|	
|			vlinkp->next_hopp = wtp->vlinkp->next_hopp;
|			break;
|		    }
|		} /* end of for loop */
|	
|		if ( vlinkp->next_hopp NE NO_NEXTHOPP )
|		{
|		    /*?? Multicast handling needs to be added. */
|		    /*?? Need to modify flow spec in case
|		      different targets have different specs? */
|		    /*?? If we allow modifies, need to allow multiple,
|			pending modifications? */
|		    /*?? For source routing, we should check here to
|		     *?? see if we have a data link to the next hop's
|		     *?? network and make plans to add the next hop
|		     *?? to multicast group */
|		    /* For now, getting a match is simply a bug since
|		       we do not support multicast we should not get
|		       a match on next hop without having already
|		       matched on the virtual link! */
|		    targetp->error = CantGetResrc;
|		    BUGSTOP (-,CantGetResrc,0x58006);
|		    /* ??? might want to collect stats before they are flushed */
|		    targetp->vlinkp = NO_VLINKP;
|		    VLinkDel ( vlinkp );
|		    break;
|		}
#endif MULITCAST_IDEA

		error = CantGetResrc;
		arcp = targetp->curchoicep;

		next_hopp = NextHopAdd( targetp->ipsrcrutp,
				        targetp->curchoicep, NO_NXTHOPP,
				        &( cp->nxthopsp ),
				        &( cp->nxthopbufp ) );
		if ( next_hopp NE NO_NEXTHOPP )
		{
		    next_hopp->cp = cp;

		    /* If it's a route to a local socket, bind to it. */
		    if (    (arcp->type EQ TO_APPL)
			AND (   (error = st2_ApplIfcBind( arcp->gifcp, cp ))
			     NE NoError) )
		    {
			(void) NextHopDel( next_hopp );
			next_hopp = NO_NEXTHOPP;

			/* ??? anything else that needs to be bound here */
			/* ??? how about fullduplex, where does merging of
			 * connections to single socket happen, etc. */
			/* Too soon for defaulted IPAdrs ...  */
		    }
		    else
		    {

			MTUAdjust( cp, next_hopp->route.nxthopp );

			/* The socket delivery routine identifies data
			 * messages by looking for a non-zero HID (and control
			 * messages by looking for a zero HID).  So we have to
			 * set the HID value in the aNxtHop entry, otherwise
			 * HID field will be zero for both data and control.
			 */

			if ( arcp->type EQ TO_APPL )
			{

			    activate_aNxtHop (next_hopp->route.nxthopp,
					      (HID) HID_RESV3);
			}
					/* ??? vs head/tail ptrs */
			if ( cp->next_hopsp EQ NO_NEXTHOPP )
			    cp->next_hopsp = next_hopp;
			else
			{
			    tailp = cp->next_hopsp;

			    while ( tailp->nextp NE NO_NEXTHOPP )
				tailp = tailp->nextp;

			    tailp->nextp = next_hopp;
			}
		    }
		} /* end of have next next_hopp */


		if ( next_hopp EQ NO_NEXTHOPP )
		{
		    targetp->error = error;
		    BUGSTOP (-,error,0x58007);

		    targetp->vlinkp = NO_VLINKP;
		    VLinkDel( vlinkp );

		    if ( IsTargState (targetp, TS_CONNECTED) )
			SetTargState (targetp, TS_FAILED); /* Run other side */
		    SetTargState (targetp, TS_RELEASED); /* No resources */
		    targetp->cp->needupdate++;
		    break;		/* out of routing case */
		}

		vlinkp->next_hopp = next_hopp;

#ifdef NOTYET
|		/*
|			    if uses existing interface
|				if mcast net,
|				    if (other) is not mcast,
|					if ok to use mcast,
|					    promote to mcast:
|						mcast addr (record resource)
|						mcast next hop
|						link to next hop
|					    mcast HID negotiation
|					    to state TS_WAIT_MCAST
|					(mesh) ...
|				    if (other) is (already) mcast,
|					link to next hop
|					mcast HID negotiation
|					to state TS_WAIT_MCAST
|				if bcast net,
|				    ... as for mcast ... TS_WAIT_MCAST
|				if ptp link,
|				    shouldn't be here - only one next hop
|				(mesh),
|				    create next hop
|				    to state TS_RSRCREQD
|			    (new interface)
|				if mcast net,
|				    if mcast preferred,
|					mcast addr (record resource)
|					create mcast next hop
|					to state TS_RSRCREQD
|				if bcast net,
|				    ... as for mcast ... TS_RSRCREQD
|				if ptp link,
|				    create next hop
|				    to state TS_RSRCREQD
|				(mesh),
|				    create next hop
|				    to state TS_RSRCREQD
|		 */
|			    break;
#endif NOTYET

		/* ??? Should copy flowspec to the next hop and
		 * then have reserve make suitable adjustment.
		 */

		next_hopp->flow_specp = targetp->flow_specp;

		/* ======================================================== */
		/* Start next hop machine */

		vlinkp->route = next_hopp->route; /* Structure copy */

		/* ??? both copies needed */

		/* Complete open wildcard field in the origin entry. */
		/* ??? use most popular?, per vlinkp */

		/* -------------------------------------------------------- */

		IfLog (DBG_UPDTARG)
		    log_NextHop( next_hopp, "update_Target/ready_NextHop" );

		if ( arcp->type EQ TO_APPL )
		{
		    SetState (NHP,next_hopp,nhop_state,RESERVED_RESOURCES);
		    SetTargState (targetp, TS_WANTOPEN);
		    targetp->cp->needupdate++;
		}
		else
		{
		    /* add downstream neighbor, just in case */

		    vlinkp->np = NbrCreate ( vlinkp->rmtIPAdr, FRF_NOSKT,
					     NO_PSCMP, targetp );
		    /* ??? TS_WAITNHOP */
		    SetState (NHP,next_hopp,nhop_state,AWAITING_RESOURCES);

		    (void) st2_RsrcAllocate( next_hopp /* multicast address
							* list; incremental*/);
		}
		/* Fall through */

	    case TS_RSRCREPLY:
		if (   (NOT IsTargState (targetp, TS_BUSY))
		    OR (targetp->vlinkp EQ NO_VLINKP)
		    OR (   (next_hopp = targetp->vlinkp->next_hopp)
			EQ NO_NEXTHOPP) )
		{
		    BUGSTOP (-,targetp->vlinkp,0x58008);
		    break;
		}

		/* ??? interposing error
		   if (targetp-error NE NoError )
		   {
		   }
		 */

		switch ( next_hopp->nhop_state )
		{
		case PENDING_RESOURCES:
		    break;

		case FREE_HOP:
		case NEW_HOP:
		case AWAITING_RESOURCES:
		case DELETING_HOP:
		default:
		    BUGSTOP (-,CantGetResrc,0x58009);
		    /* Fall through */

		case REJECTED_RESOURCES:
		    IfLog (DBG_UPDTARG)
			log_NextHop( next_hopp, "update_Target/rejected_NextHop" );

		    /* Restart target state machine in all
		     * affected targets.
		     */

		    ForAllTargets (targetp->cp, wtp)
		    {
			if ( NOT IsTargState (wtp, TS_BUSY) )
			    continue;

			if (   (wtp->vlinkp EQ NO_VLINKP)
			    OR (wtp->vlinkp->next_hopp NE next_hopp) )
			    continue;

			IfLog (DBG_UPDTARG)
			    log_Target( wtp, "update_Target-CantGetResc" );

			if ( wtp->curchoicep NE NO_RUTCHOICEP )
			{
			    wtp->curchoicep->flags |= RouteFailed | RouteNoRsrc;
			}
			else
			{
			    BUGSTOP (-,wtp,0x58010); /* paranoia */
			}

			oldpri = splnet ();
			{
			    vlinkp = wtp->vlinkp;
			    wtp->vlinkp = NO_VLINKP;
			}
			(void) splx ( oldpri );

			VLinkDel( vlinkp ); /* & NextHop & resources */

			if ( IsTargState (wtp, TS_WAITNHOP) )
			{
			    /* ??? check TS_CONNECTED redundant */
			    SetTargState (wtp, TS_ROUTING);
			    wtp->cp->needupdate++;
			}
			    
		    } /* end of for all pending targets loop */

		    /* do the right thing for this target, others ... ??? */
		    error = CantGetResrc;
		    goto reroute;

		case RESERVED_RESOURCES:

		    /* Restart target state machine in all
		     * affected targets.
		     */

		    ForAllTargets (targetp->cp, wtp)
		    {
			if ( NOT IsTargState (wtp, TS_BUSY) )
			    continue;

			if (   (wtp->vlinkp EQ NO_VLINKP)
			    OR (wtp->vlinkp->next_hopp NE next_hopp) )
			    continue;

			IfLog (DBG_UPDTARG)
			    log_Target( wtp, "update_Target-GotResec" );

			if ( IsTargState (wtp, TS_WAITNHOP) )
			{
			    SetTargState (wtp, TS_WANTOPEN);
			    wtp->cp->needupdate++;
			}
			    
		    } /* end of for all pending targets loop */

		    break;		/* out of reserved case */
		} /* end of switch on resource allocation result */

		break;	/* stop here to let other targets get resources for
			 * batched CONNECT */


	    case TS_WANTOPEN:

		/* -------------------------------------------------------- */

		/* Cannot put off selection of our IP address any longer ...
		 * but it's too late for local targets as the routing pased the
		 * info along.
		 */

		if (   (targetp->cp->soname.applent.IPAdr EQ INADDR_ANY)
		    OR (targetp->cp->detectorIPAdr EQ INADDR_ANY)
		    OR (    (targetp->cp->Origin.pcode NE (octet1) 0)
			AND (targetp->cp->Origin.IPAdr EQ INADDR_ANY)) )
		{
		    lclipadr = INADDR_ANY;

		    ForAllTargets (targetp->cp, wtp)
		    {
			if (   (NOT IsTargState (wtp, TS_BUSY))
			    OR (wtp->curchoicep EQ NO_RUTCHOICEP) )
			    continue;

			lclipadr = wtp->curchoicep->lclIPAdr;

			if ( wtp->curchoicep->type NE TO_APPL )
			    break;

		    } /* end of for all targets loop */

		    if ( lclipadr EQ INADDR_ANY )
			lclipadr = local_inetaddr;

		    if ( targetp->cp->soname.applent.IPAdr EQ INADDR_ANY )
			targetp->cp->soname.applent.IPAdr = lclipadr;

		    if ( targetp->cp->detectorIPAdr EQ INADDR_ANY )
			targetp->cp->detectorIPAdr = lclipadr;

		    if (    (targetp->cp->Origin.pcode NE (octet1) 0)
			AND (targetp->cp->Origin.IPAdr EQ INADDR_ANY) )
			targetp->cp->Origin.IPAdr = lclipadr;
		}


		if (    (targetp->cp->upstream_vlinkp NE NO_VLINKP)
		    AND (targetp->cp->upstream_vlinkp->type EQ APPL_VLINK) )
		{
		    if ( targetp->cp->upstream_vlinkp->lclIPAdr EQ INADDR_ANY )
			targetp->cp->upstream_vlinkp->lclIPAdr =
			    targetp->cp->detectorIPAdr;
		    if ( targetp->cp->upstream_vlinkp->rmtIPAdr EQ INADDR_ANY )
			targetp->cp->upstream_vlinkp->rmtIPAdr =
			    targetp->cp->detectorIPAdr;
		}

		/* -------------------------------------------------------- */


		reason = DoConnect( targetp ); /* ??? INLINE */
		break;


	    case TS_OPENING:
		break;			/* Awaiting timeout or ack */


	    case TS_OPEN:
		break;			/* Await disconnect/refuse */


	    case TS_WANTCLOSED:
		/* ??? G-Bit or individual */

		if ( targetp->vlinkp NE NO_VLINKP )
		{
		    reason = DoDisconnect( targetp,
				(targetp->error) ? targetp->error :
				(targetp->curchoicep EQ NO_RUTCHOICEP) ? ErrorUnknown :
				(targetp->curchoicep->error) ? targetp->curchoicep->error :
				ErrorUnknown );	/* ??? INLINE */
		}
		else
		{
		    SetTargState (targetp, TS_RELEASED);
		    goto released;
		}
		break;


	    case TS_CLOSING:
		break;			/* Awaiting timeout or ack/disconnect */


	    case TS_WANTSUSPENDED:
		/* ??? G-Bit (not individual) */
		reason = DoDisconnect( targetp,
				(targetp->error) ? targetp->error :
				(targetp->curchoicep EQ NO_RUTCHOICEP) ? ErrorUnknown :
				(targetp->curchoicep->error) ? targetp->curchoicep->error :
				ErrorUnknown );	/* ??? INLINE */
		break;

	    case TS_PENDSUSPEND:	/* Awaitting timeout or ack/disconnect */
		break;

	    case TS_SUSPENDED:
		/* ??? start timer, then to TS_CLOSED */
		SetTargState (targetp, TS_CLOSED);
		/* Fall into CLOSED */

	    case TS_CLOSED:

		oldpri = splnet ();
		{
		    vlinkp = targetp->vlinkp;
		    targetp->vlinkp = NO_VLINKP;
		}
		(void) splx ( oldpri );

		VLinkDel( vlinkp );	/* & NextHop & aNxtHop & rsrc */

		SetTargState (targetp, TS_RELEASED);
		/* Fall through */


	    case TS_RELEASED:
  released:
		if (    IsTargState (targetp, TS_CONNECTED)
		    AND ((targetp->cp->optapi & STOptSBit) EQ 0) )
		{
		    SetTargState (targetp, TS_ROUTING);
		    if ( (error = targetp->error) EQ NoError )
			error = targetp->curchoicep->error;
		    goto reroute;
		}

		if ( IsTargState (targetp, TS_DISCONNECTED) )
		{
		    if ( recurse++ NE 0 ) /* ??? does it recurse back here */
			BUGSTOP (-,recurse,0x58011);
		    /* ??? make sure TS_DEAD */
		    (void) TargDel( targetp );
		    recurse--;
		}
		else
		{
		    if ( IsTargState (targetp, TS_CONNECTED) )
			SetTargState (targetp, TS_FAILED); /* Run other side */
		    /* not need, FAILED processed below */
		    /* targetp->cp->needupdate++; */
		}
		break;

	    default:
		break;

	    } /* end of next hop switch */

	    /* ============================================================ */

	    if ( IsTargState (targetp, TS_DEAD) )
	    {
		IfLog (DBG_UPDTARG)
		    log_Target( targetp, "update_Target-completing" );
		continue;		/* on to next target */
	    }

	    /* ============================================================ */
	    /* Work the upstream state machine */

	    switch ( targetp->targstate[ (int) Prevhop ] )
	    {
	    case TS_CONNECTED:
		break;			/* Awaiting disconnect/refuse/failure */


	    case TS_REFUSING:
		break;			/* Awaiting acknowledgement/timeout */


	    case TS_FAILED:
		reason = DoRefuse( targetp, CompRefuse );

		if (    (reason EQ NoError)
		    AND (targetp->cp->upstream_vlinkp NE NO_VLINKP)
		    AND (targetp->cp->upstream_vlinkp->route.type NE TO_APPL) )
		{			/* Wait for ack */
		    SetTargState (targetp, TS_REFUSING);
		    break;
		}

		/* Don't get acks from socket layer, continue */
		SetTargState (targetp, TS_DISCONNECTED);
		/* Fall through */


	    case TS_DISCONNECTED:
		if ( IsTargState (targetp, TS_RELEASED) )
		{
		    if ( recurse++ NE 0 ) /* ??? does it recurse back here */
			BUGSTOP (-,recurse,0x58012);
		    (void) TargDel( targetp );
		    recurse--;
		}
		break;


	    case TS_WANTHOLD:
		reason = DoRefuse( targetp, CompRefuse );

		if (    (reason EQ NoError)
		    AND (targetp->cp->upstream_vlinkp NE NO_VLINKP)
		    AND (targetp->cp->upstream_vlinkp->route.type NE TO_APPL) )
		{			/* Don't get acks from socket layer */
		    SetTargState (targetp, TS_PENDHOLD);
		    break;
		}

		SetTargState (targetp, TS_DISCONNECTED);
		targetp->cp->needupdate++; /* to get to TS_RELEASED & TS_DEAD */
		break;


	    case TS_PENDHOLD:
		/* ??? if pendhold timeout, set onhold timeout & onhold state */
		break;			/* Awaiting acknowledgement/timeout */


	    case TS_ONHOLD:
		/* ??? on onhold timeout set disconnected state */
		break;			/* Awaiting timeout */

	    } /* end of next hop switch */

	    /* ============================================================ */

	    IfLog (DBG_UPDTARG)
		log_Target( targetp, "update_Target-completing" );

	} /* end of for selected targets loop */
    } /* end of while things to do */

    if ( IsState (FWD,cp,fwdr_state,FW_DISCONNECT) )
    {
	st2_PCBDel( cp );
	/* cp is be invalid */
    }

    return;
}

/*
S* st2_VLinkReset ( vlinkp, suggest_vlid )
s*
s*	Routine to prepare a virtual link structure for the next use by
s*	assigning a unique id to it.
s*
 */

  void					/* Next bugid 0x58100 */
st2_VLinkReset( vlinkp, suggest_vlid )	/* from VLinkDel st2_init */
  struct aVLink		*vlinkp;
  VLID			 suggest_vlid;
{
    PTRCHECK (vlinkp, struct aVLink *, return;);

    Bzero (vlinkp,sizeof (*vlinkp));
    vlinkp->type = FREE_VLINK;
    vlinkp->local_vlid = (suggest_vlid >= VLID_DYNAMIC) ? (suggest_vlid)
			 : suggest_vlid + ConfigParm (vlnk_mask) + 1;
					/* ??? use incarnation # */
    return;
}

/*
S* TargAdd ( cp, aTargetp, reference )
s*
s*	Routine to build a target descriptor and start its state machine.
s*
 */

  static enum ST2Errors			/* Next bugid 0x58204 */
TargAdd( cp, aTargetp, reference )	/* from HandleConnect st2_DoConnect */
  struct aST2pcb	*cp;		/* aka forwarder */
  struct aTarget	*aTargetp;	/* net-order */
  Reference		 reference;	/* may be zero  non-zero	0 */
{
    char		*padp;
    int			 targbytes,
			 parmbytes;
    long		*parmspace;	/* Points to PCode'd target */
    struct aSrcRut	*endp,
			*srp;
    struct Target	*targetp;


    /* Validate target format */

    endp = Mkp (struct aSrcRut *, aTargetp, ntoh1 (aTargetp->TargetBytes));
    padp = Mkp (char *, &( aTargetp->SAP[0] ), ntoh1 (aTargetp->SAPBytes));
    if ( padp > (char *) endp )
	return ( InvalidTarget );

    switch ( (int) padp & 3 )
    {
    case 1:	*padp++ = 0;		/* zero padding */
    case 2:	*padp++ = 0;		/* zero padding */
    case 3:	*padp++ = 0;		/* zero padding */
    case 0:
    default:	;
    } /* end of switch */

    srp = Mkp (struct aSrcRut *, padp, 0);

    while ( srp < endp )
    {
	switch ( ntoh1 (srp->pcode) )
	{
	case STpIPLSrcRut:		/* 16 */
	case STpIPSSrcRut:		/* 17 */
	case STpSTLSrcRut:		/* 18 */
	case STpSTSSrcRut:		/* 19 */
	    break;
	default:
	    return ( InvalidTarget );
	} /* end of switch */

	parmbytes = ntoh1 (srp->plen);
	if (   ((parmbytes & 3) NE 0)
	    OR (srp->fill2 NE 0) )
	    return ( InvalidTarget );

	srp = Mkp (struct aSrcRut *, srp, parmbytes);

	if ( srp > endp )
	    return ( InvalidTarget );

    } /* end of while more srcrut parms */

    /* -------------------------------------------------------------------- */
    /* If target isn't already on the list, add it */

    targetp = TargByParm( cp, UPSTREAM_VLINK, aTargetp );

    /* -------------------------------------------------------------------- */

    if ( targetp EQ NO_TRGP )
    {
	/* ??? Check to see if target is listed for any other
	 *     forwarder on this stream because if it is, that's
	 *     a routing error and we have some refusing to do.
	 * ??? extra work if multiple matches
	 */

	/* Complete the target entry and add it to list */
					/* new_target */
	/* First try to get a single chuck of space large enough to hold
	   the target structure followed by a copy of the target parameter
	   aTargetp. */
	targbytes = ROUND (sizeof (struct Target), octet4);
	parmbytes = ROUND (ntoh1 (aTargetp->TargetBytes), octet4);

	/* Leave room for FlowSpec, UserData */
	targetp = GetSpace (struct Target *, &( cp->fwdrfreespace ),
			    PKT_TARGET, PKT_MD_NEW, PKT_XA_PS,
			    0,200, targbytes, 0,0);
			/* ??? 200 > parmbytes + sizeof (struct aFlowSpec3) */
	if ( targetp EQ NO_TRGP )
	{
#ifdef  STIIDEBUG
	    (void) printf ( "Can't allocate new target\n" );
#endif  STIIDEBUG
	    Count (trg_fail,1);		/* Cannot do anything now */
	    BUGRETURN (-,CantGetResrc,0x58201,enum ST2Errors);
	}

	parmspace = AddSpace (long *, &( cp->fwdrfreespace ),
			      PKT_MD_PSOVHD, parmbytes);
	if ( parmspace EQ (long *) NULL )
	{
#ifdef  STIIDEBUG
	    (void) printf ( "No space in target for parameter\n" );
#endif  STIIDEBUG
	    Count (trg_fail,1);		/* Cannot do anything now */
	    BUGRETURN (-,CantGetResrc,0x58202,enum ST2Errors);
	}

	Bzero (targetp,sizeof (*targetp));
	targetp->cp = cp;
	targetp->error = NoError;
	SetTargState (targetp,TS_BUSY);
	SetTargState (targetp,TS_CONNECTED);
	SetTargState (targetp,TS_NEW);
	targetp->routes.allocated = DimensionOf (targetp->routes.RouteChoices);
	targetp->tparminp = (struct aTarget *) parmspace;

	Count (trg_get,1);
	uniqtime ( &( targetp->start ) );

	/*
	 * Cache copy of original parameter for use in quick comparisons.
	 * Precompute offsets to the key elements so that other code does
	 * not know the actual protocol representation.
	 */
	{   int tlen = ntoh1 (aTargetp->TargetBytes);
	    Bcopy (aTargetp,targetp->tparminp,tlen);
	    targetp->tparmin_bytes = tlen;
	}
	targetp->targetIPAdr = targetp->tparminp->IPAdr;

	targetp->tparmoutp = (struct aTarget *) ((long) targetp + targbytes);
	targetp->tparmoutp->TargetBytes = 0;

	kassert (NOT IsState (FWD,cp,fwdr_state,FW_FREE));

	targetp->target_index = ++(cp->target_index); /* diagnostic */

	targetp->cp->targets_active++;

	if ( cp->targetsp EQ NO_TRGP )
	    cp->targetsp = targetp;
	else
	    cp->targettailp->nextp = targetp;
	cp->targettailp = targetp;

	cp->needupdate++;

	IfLog (DBG_INI_DS_MSG)
	    log_Target( targetp, "new_Target" );

    } /* end of adding new target */
/*  else if ( old ref# EQ linked ) ignore duplicate ??? */

    /* -------------------------------------------------------------------- */

    /*??? Need to do something if it is in closing state: either
      send temporary refuse back so previous hop will retry
      or move to closed state, but then we need to be sure
      that resources have been properly released first and
      that we don't reuse next-hop vlids improperly. */

    switch ( targetp->targstate[ (int) Nexthop ] )
    {
    case TS_CLOSING:			/* Refuse connect on closing */
	targetp->error = CantGetResrc /*eStillClosing*/;

	/* ??? check TS_CONNECTED/TS_OPEN */

	SetTargState (targetp,TS_WANTCLOSED);
	SetTargState (targetp,TS_FAILED);
	targetp->cp->needupdate++;
	/* All our callers call st2_PCBUpdate */
	break;

    case TS_CLOSED:			/* Re-open closed connection */
	SetTargState (targetp,TS_NEW);
	targetp->cp->needupdate++;

	/* Falls through */

    case TS_NEW:
	targetp->linked_reference = reference;
	targetp->change_reference = reference;
	targetp->lnkrefflg = 1;		/* Refuse requires a LnkRef */
	break;

    default:
					/* ??? needs work for hold/suspend */
	/*?? Another open while in opening state, for now ignore
	  but ought to make sure linked reference, etc, is ok */
	BUGSTOP (-,PackTState (targetp,0),0x58203);
	break;
    } /* end of switch */

    /*?? else make sure flow spec is compatible? */

    return ( NoError );
}

/*
S* TargByParm ( cp, dir, aTargetp )
s*
s*	Routine to scan a connection to find a target matching the specified
s*	Target Parameter.  Note that a target parameter from upstream may be
s*	different than a target parameter from downstream if source routing
s*	is being used.  Try to find the "best match" if exact match fails.
**
 */

  static struct Target *		/* Next bugid 0x58304 */
TargByParm( cp, dir, aTargetp )		/* from HandleAccept HandleDisconnect HandleRefuse TargAdd */
					/* ??? MACRO/INLINE */
  struct aST2pcb	*cp;		/* 	non-0 */
  enum link_direction	 dir;		/* UPSTREAM_VLINK | DOWNSTREAM_VLINK */
  struct aTarget	*aTargetp;	/* 	non-0 */
{
    struct aTarget	*atp;
    struct Target	*base_targetp = NULL,
			*full_targetp = NULL,
			*wtargetp;
    int			 base_matches = 0,
			 full_matches = 0,
			 prefixlen,
			 tbytes = ntoh1 (aTargetp->TargetBytes);


    PTRCHECK (cp, struct aST2pcb *, return ( 0 ));
    PTRCHECK (aTargetp, struct aTarget *, return ( 0 ));

    prefixlen = ROUND ((OffsetOf (SAP[0],struct aTarget)
			+ ntoh1 (aTargetp->SAPBytes)), octet4);

    ForAllTargets (cp,wtargetp)
    {
	if ( NOT IsTargState (wtargetp, TS_BUSY) )
	    continue;			/* Uninteresting target */

	if ( prefixlen NE wtargetp->prefixlen )
	    continue;			/* Different SAPBytes */

	if ( aTargetp->IPAdr NE wtargetp->tparminp->IPAdr )
	    continue;

	if ( aTargetp->SAPBytes NE wtargetp->tparminp->SAPBytes )
	    continue;

	if ( Bcmp (&( aTargetp->SAP[0] ),&( wtargetp->tparminp->SAP[0] ),
		   wtargetp->tparminp->SAPBytes) NE 0 )
	    continue;			/* Different SAP */

	base_targetp = wtargetp;	/* Matched IPAdr & SAP */
	base_matches++;
	    
	switch ( dir )
	{
	default:		BUGSTOP (-,dir,0x58301); /* bad argument */
	case UPSTREAM_VLINK:	atp = wtargetp->tparminp;	break;
	case DOWNSTREAM_VLINK:	atp = wtargetp->tparmoutp;	break;
	} /* end of switch */

	if ( tbytes NE ntoh1 (atp->TargetBytes) )
	    continue;			/* Different srcrut lengths */

	if ( Bcmp (aTargetp + prefixlen,atp + prefixlen,
		   tbytes - prefixlen) NE 0 )
	    continue;			/* Different srcruts */

	full_targetp = wtargetp;	/* Full match */
	full_matches++;			/* Hope only one */

    } /* end of for all targets */

    if ( full_matches EQ 1 )
	return ( full_targetp );	/* Expected case */

    if (    (full_matches EQ 0)
	AND (base_matches EQ 1) )
    {
	BUGSTOP (-,(base_matches << 16),0x58302); /* ??? Interoperability */
	return ( base_targetp );
    }

    /* ??? Something unexpected is going on */
    if ( (full_matches NE 0) OR (base_matches NE 0) )
	BUGSTOP (-,((base_matches << 16) | full_matches),0x58303);

    return ( NO_TRGP );
}

/*
S* TargDel ( targetp )
s*
s*	Routine to delete a target structure.
s*
 */

  static enum ST2Errors			/* Next bugid 0x58404 */
TargDel( targetp )			/* from st2_PCBDel st2_PCBUpdate*2 TargDisc */
  struct Target		*targetp;
{
    int			 oldpri;
    struct aVLink	*vlinkp;


    PTRCHECK (targetp, struct Target *, return ( InvalidPointer ));

    if ( IsTargState (targetp, TS_DEAD) ) /* multiple calls */
	BUGRETURN (-,Inconsistency,0x58401,enum ST2Errors);

    if (   NOT IsTargState (targetp, TS_DISCONNECTED)
	OR NOT IsTargState (targetp, TS_RELEASED) )
	BUGSTOP (-,Inconsistency,0x58402);

    oldpri = splnet ();
    {
	SetTargState (targetp,TS_DEAD);

	if ( --(targetp->cp->targets_active) <= 0 )
	{
	    if ( (targetp->cp->flags & LocalOrigin) EQ 0 )
		SetState (FWD,targetp->cp,fwdr_state,FW_DISCONNECT);
	}

	vlinkp = targetp->vlinkp;
	targetp->vlinkp = NO_VLINKP;
    }
    (void) splx ( oldpri );

    Count (trg_rel,1);
    uniqtime ( &( targetp->stop ) );

#ifdef STIIDEBUG
#if 0
    BUGSTOP (-,targetp->cp->targets_active,0x58403);
#endif
#endif STIIDEBUG

    /* Delete virtual link/next hop/resources */
    if ( vlinkp NE NO_VLINKP )
    {
	VLinkDel( vlinkp );
    }


    /* ??? release resources per target (shared by all for next-hop) */

    return ( NoError );
}

/*
S* TargDisc ( targetp, reference, reason, userdatap )
s*
s*	Routine to start the specified target on the path to being
s*	disconnected for the given reason.
s*
 */

/*??? Should be TargDel to tear out data stuff.   Disconnect target
       should propagate disconnect to next hop.  And then call delete if
       target has been disconnected and refused.  Should take an optional
       message parameter that it uses for other parameters to propagate.
       [?? should we continue to have reference and reason or should
       caller be required to build partial message to provide them?] */


  static enum ST2Errors			/* Next bugid 0x58500 */
TargDisc( targetp, reference, reason, userdatap ) /* from HandleDisconnect */
  struct Target		*targetp;
  Reference		 reference;
  enum ST2Errors	 reason;
  struct aUserData	*userdatap;
{
    PTRCHECK (targetp, struct Target *, return ( -InvalidPointer ));
					/* cancel_pending_Target */
    targetp->pending_reference = (Reference) 0;
					/* ??? obsolete */
    targetp->discon_userdatap = userdatap;
    targetp->error = reason;

    /* ??? if nxthop, deactivate_aNxtHop (nxthop); */


    if ( NOT IsTargState (targetp, TS_DISCONNECTED) )
    {
	SetTargState (targetp,TS_DISCONNECTED);
    }


    switch ( targetp->targstate[ (int) Nexthop ] )
    {	
	/* These states indicate we have not opened a connection
	   to the next hop, so we need only tear down our data
	   structures. */
    case TS_NEW:
    case TS_ROUTEOPT:
    case TS_ROUTING:
					/* ??? to TS_RELEASED */
    case TS_RSRCREPLY:
    case TS_WAITNHOP:
    case TS_WANTOPEN:
    case TS_WANTSUSPENDED:
	SetTargState (targetp,TS_CLOSED);
	/* Fall through */

    case TS_CLOSED:
	(void) TargDel( targetp );
	break;				/* ??? why not reason arg */

	/* These states indicate we may have opened a connection
	   to the next hop, or at least initated one.  We need
	   to send a disconnect and wait for an ack to know
	   that the next hop has no residual connect to the
	   target */
    case TS_OPENING:
    case TS_OPEN:
    case TS_PENDSUSPEND:
    case TS_SUSPENDED:
    default:
	targetp->linked_reference = reference;

	/* We go straight to closed if link was never established. */
	if ( targetp->vlinkp NE NO_VLINKP )
	{
	    SetTargState (targetp,TS_WANTCLOSED);
	}
	else
	{
	    SetTargState (targetp,TS_CLOSED);
	}

	break;

	/* If a close has already been initiated, no more to do. */
    case TS_WANTCLOSED:
    case TS_CLOSING:
    case TS_RELEASED:
	break;
    } /* end of switch */

    return ( NoError );
}

/*
S* VLinkByVLId ( vlid )
s*
s*	Routine to map a given virtual link id onto ist virtual link
s*	structure.
s*
 */

  static struct aVLink *		/* Next bugid 0x58600 */
VLinkByVLId( vlid )			/* from st2_CMPInput */
  VLID			 vlid;
{
    struct aVLink	*vlinkp;


    /* We only use low order bits for index, but must check for match
     *  with confirming vlid value in the link structure.
     */

    vlinkp = &( st2_vlinks.table[ vlid & ConfigParm (vlnk_mask) ] );
    if ( (vlinkp->type NE FREE_VLINK)  AND  (vlinkp->local_vlid EQ vlid) )
	return ( vlinkp );

    /* Return null if given vlid is not known. */
    return ( NO_VLINKP );
}

/*
S* VLinkDel ( vlinkp )
s*
s*	Decrements the vlink_cnt for this control link.
s*	If this is the last use of the vlink, delete the control link.
s*
*/

  static void				/* Next bugid 0x58707 */
VLinkDel( vlinkp )			/* from st2_PCBDel st2_PCBUpdate*3 TargDel */
    struct aVLink	*vlinkp;
{
    int			 hash_index,
			 oldpri;
    VLID		 new_vlid;
    struct aVLink	*hash_scan;
    struct NextHop	*nexthopp;


    /* Don't make all callers check it */
    if ( vlinkp EQ NO_VLINKP )
    {
	BUGSTOP (-,1,0x58701);		/* But a NULL is probably a bug */
	return;
    }


    /* De enroll for failure detection */
    if (    (vlinkp->np NE NO_NBRP)
	AND (vlinkp->type EQ GATEWAY_VLINK)
	AND (vlinkp->vlink_cnt <= 1) )
	NbrRemove( vlinkp );


    /* Delete link if this is the last user */
    oldpri = splnet ();
    {
	/* One less user of vlink ... */
	if ( --(vlinkp->vlink_cnt) <= 0 )
	{
	    /* Time to delete the vlink */
	    if ( vlinkp->vlink_cnt < 0 )
		BUGSTOP (-,vlinkp->vlink_cnt,0x58702);

	    /* One less user of any nexthop ... */
	    if ( vlinkp->next_hopp NE NO_NEXTHOPP )
	    {
		/* for the meantime we will always delete */
		if ( --(vlinkp->next_hopp->next_cnt) <= 0 )
		{
		    if ( vlinkp->next_hopp->next_cnt < 0 )
		    {
			BUGSTOP (-,vlinkp->next_hopp->next_cnt,0x58703);
		    }

		    nexthopp = vlinkp->next_hopp;
		    vlinkp->next_hopp = NO_NEXTHOPP;

					/* ??? unlock next_hop */
		    (void) NextHopDel( nexthopp );
		    /* ??? error */
		}
	    }

	    IfLog (DBG_DTA_STR)
		printf ( "VLinkDel at 0x%x  cp 0x%x\n",
			 vlinkp, vlinkp->cp );

	    kassert (vlinkp->type NE FREE_VLINK);

	    if ( vlinkp->failurep NE NO_VLINKP )
	    {
		BUGSTOP (-,vlinkp,0x58704);
		BUGSTOP (-,vlinkp->type,0x58705);
		BUGSTOP (-,vlinkp->failurep,0x58706);
		/* ??? trouble here */
	    }

	    /* ------------------------------------------------------------ */
	    /* Remove from hash bucket, if any. */

	    /* Compute index to correct hash bucket. */
	    hash_index = VLinkHash (vlinkp->rmtIPAdr,
				    vlinkp->remote_vlid);

	    /* Remove link entry from the hash list */
	    hash_scan = st2_vlinks.remote_hash[ hash_index ];
	    if ( hash_scan EQ vlinkp ) 
		st2_vlinks.remote_hash[ hash_index ] = vlinkp->nextp;
	    else while ( hash_scan NE NO_VLINKP )
	    {
		if ( hash_scan->nextp EQ vlinkp )
		{
		    hash_scan->nextp = vlinkp->nextp;
		    hash_scan = NO_VLINKP; /* Found, leave loop. */
		}
		else
		{
		    /* traverse the hash bucket */
		    hash_scan = hash_scan->nextp;
		}
	    } /* end of else while */

	    /* ------------------------------------------------------------ */
	    /* Mark entry as idle and give it a new link number. */

	    vlinkp->type = FREE_VLINK;

	    new_vlid =   (vlinkp->local_vlid + ConfigParm (vlnk_mask) + 1)
		       & NBitMask (VLID_BITS);
	    st2_VLinkReset( vlinkp, new_vlid );
	    Count (vlinkrel,1);

	    /* Free entry goes to the end of the free list. */
	    if ( st2_vlinks.first_freep EQ NO_VLINKP )
	    {
		/* If empty, we're the start. */
		st2_vlinks.first_freep = vlinkp;
	    }
	    else
	    {
		/* Not empty, chain to end. */
		st2_vlinks.last_freep->nextp = vlinkp;
	    }
	    st2_vlinks.last_freep = vlinkp;
	}
    }
    (void) splx ( oldpri );

    return;
}

/*
S* VLinkNew ( cp, direction, type )
s*
s*	Allocate a new virtual link entry and return its id number.
s*	The stream argument must be provided to show which stream table
s*	entry this link is associated with.  This routine is called by
s*	the sender of a connect to be followed by set_remote_ids when
s*	the acknolwedgement returns.  The remote markings are left blank
s*	until a call to set them is made.  The link is usable for connect
s*	messages, but generally no others until both settings have been
s*	completed.
s*	Returns: virtual link identifier for the new link table entry
s*	Errors: no free link table entries available, returns zero
s*
 */

  static struct aVLink *		/* Next bugid 0x58800 */
VLinkNew( cp, direction, type )		/* from HandleConnect NbrCreate st2_DoConnect st2_PCBUpdate */
  struct aST2pcb	*cp;		/* NO_PCBP when NEIGHBOR_VLINK */
  enum link_direction	 direction;	/* upsteam | downstream */
  enum link_type	 type;
{
    int			 oldpri;
    VLID		 lclvlid;
    struct aVLink	*vlinkp = NO_VLINKP;


    /* Take first entry off the free list. */
    oldpri = splnet ();
    {
	if ( st2_vlinks.first_freep NE NO_VLINKP )
	{
	    vlinkp = st2_vlinks.first_freep;
	    st2_vlinks.first_freep = vlinkp->nextp;
	    if ( st2_vlinks.first_freep EQ NO_VLINKP )
		st2_vlinks.last_freep = NO_VLINKP;
	}
    }
    (void) splx ( oldpri );

    /* If allocation failed, return VLID 0 */
    if ( vlinkp EQ NO_VLINKP )
    {
	Count (vlinkfail,1);
	return ( NO_VLINKP ); /* ??? resource error */
    }

    lclvlid = vlinkp->local_vlid;
    Bzero (vlinkp,sizeof (struct aVLink));
    vlinkp->local_vlid = lclvlid;

    vlinkp->vlink_cnt = 1;
    vlinkp->direction = direction;
    vlinkp->type = type;
    vlinkp->cp = cp;
    Count (vlinkget,1);

    IfLog (DBG_DTA_STR)
	printf ( "VLinkNew at 0x%x for cp 0x%x  dir = %u  type = %u  VLId = %u\n",
		 vlinkp, cp, direction, type, vlinkp->local_vlid );

    return ( vlinkp );
}

/*
S* VLinkSetRemote ( vlinkp, remote_vlid, rmtIPAdr )
s*
s*	Set the remote markings for a virtual link.  This will happen
s*	immediately after creating the link if we are the receiver
s*	of a connect message, or when an acknowledgement is received
s*	if we are the sender of a connect message.
s*	Returns: true (1) is setting is successful.
s*	Errors: if remote markings have already been set to a conflicting
s*	value, this routine returns false (0).
s*
 */

  static boolean			/* Next bugid 0x58902 */
VLinkSetRemote( vlinkp, remote_vlid, rmtIPAdr ) /* from HandleConnect st2_CMPInput */
  struct aVLink		*vlinkp;
  VLID			 remote_vlid;
  INETADDR		 rmtIPAdr;
{
    int			 hash_index;


    PTRCHECK (vlinkp, struct aVLink *, return ( FALSE ));

    /* Don't allow remote vlid to be set to a reserved virtual link */
    /*?? Is this right, or should we just check for zero? */
    if ( remote_vlid < VLID_DYNAMIC )
    {
	BUGSTOP (-,remote_vlid,0x58901);
	return ( FALSE );
    }

    /* Make sure the we never get conflicting foreign link markings. */
    if ( vlinkp->remote_vlid NE VLID_NONE )
    {
	if (    (vlinkp->rmtIPAdr EQ rmtIPAdr)
	    AND (vlinkp->remote_vlid EQ remote_vlid) )

	    return ( TRUE );		/* Already set correctly */

	else
	    return ( FALSE );		/* Conflicts with original setting */
    }

    /* Set host field first because we check vlid to see if entry is set */
    vlinkp->rmtIPAdr = rmtIPAdr;
    vlinkp->remote_vlid = remote_vlid;

    /* Add entry to hash table for searching by remote identifiers. */
    hash_index = VLinkHash (rmtIPAdr,remote_vlid);
    {	/*lock st2_vlinks*/
	vlinkp->nextp = st2_vlinks.remote_hash[ hash_index ];
	st2_vlinks.remote_hash[ hash_index ] = vlinkp;
    }	/*unlock st2_vlinks*/

    return ( TRUE );
}

/*
 Local Variables:
 c-indent-level: 4
 c-brace-offset: -4
 c-brace-imaginary-offset: 0
 c-continued-statement-offset: 4
 comment-column: 40
 End:
*/
